from django.db import models
from django.utils.translation import gettext_lazy as _

from solawi_apps.db_lib.models import NanoIdPkMixin, CreatedMixin, LastModifiedMixin, EnabledMixin, OwnerMixin


class Ahg(NanoIdPkMixin, OwnerMixin, CreatedMixin, LastModifiedMixin, models.Model):

    class Meta:
        verbose_name = _("Abholgemeinschaft")
        verbose_name_plural = _("Abholgemeinschaften")
        default_permissions = ()
        unique_together = (
            ("owner_fk", "name"),
        )

    name = models.CharField(_("Name"), max_length=255, null=False, blank=False)
    note = models.TextField(_("Notiz"), null=True, blank=True, default=None, max_length=100_000)


class HarvestYear(NanoIdPkMixin, CreatedMixin, LastModifiedMixin, models.Model):

    class Meta:
        verbose_name = _("Erntejahr")
        verbose_name_plural = _("Erntejahre")
        default_permissions = ()
        unique_together = (
            ("year", "ahg_fk"),
        )

    ahg_fk = models.ForeignKey(
        Ahg, verbose_name=_("Abholgemeinschaft"), null=False, blank=False, on_delete=models.CASCADE
    )
    year = models.CharField(_("Jahr"), null=False, blank=False, max_length=255)
    note = models.TextField(_("Notiz"), null=True, blank=True, default=None, max_length=100_000)


class Date(NanoIdPkMixin, CreatedMixin, LastModifiedMixin, EnabledMixin, models.Model):

    class Meta:
        verbose_name = _("Datum")
        verbose_name_plural = _("Daten")
        default_permissions = ()
        unique_together = (
            ("harvest_year_fk", "date"),
        )

    harvest_year_fk = models.ForeignKey(
        HarvestYear, verbose_name=_("Erntejahr"), null=False, blank=False, on_delete=models.CASCADE
    )
    date = models.DateField(_("Datum"), null=False, blank=False)
    note = models.TextField(_("Notiz"), null=True, blank=True, default=None, max_length=100_000)
    needs_pickup = models.BooleanField(_("Benötigt Abholung"), null=False, blank=False, default=True)


class Member(NanoIdPkMixin, CreatedMixin, LastModifiedMixin, EnabledMixin, models.Model):

    class Meta:
        verbose_name = _("Mitglied")
        verbose_name_plural = _("Mitglieder")
        default_permissions = ()
        unique_together = (
            ("ahg_fk", "name"),
        )

    ahg_fk = models.ForeignKey(
        Ahg, verbose_name=_("Abholgemeinschaft"), null=False, blank=False, on_delete=models.CASCADE
    )
    name = models.CharField(_("Name"), max_length=255, null=False, blank=False)
    email = models.EmailField(_("E-Mail"), null=True, blank=False, default=None)
    requires_team = models.BooleanField(_("Benötigt Team"), null=False, blank=False, default=False)
    shares = models.DecimalField(_("Anteile"), decimal_places=2, max_digits=3, null=True, blank=True, default=1.0)
    note = models.TextField(_("Notiz"), null=True, blank=True, default=None, max_length=100_000)


class Team(NanoIdPkMixin, CreatedMixin, LastModifiedMixin, EnabledMixin, models.Model):

    class Meta:
        verbose_name = _("Abhol-Team")
        verbose_name_plural = _("Abhol-Teams")
        default_permissions = ()
        unique_together = (
            ("ahg_fk", "name"),
        )

    ahg_fk = models.ForeignKey(
        Ahg, verbose_name=_("Abholgemeinschaft"), null=False, blank=False, on_delete=models.CASCADE
    )
    name = models.CharField(_("Team-Name"), null=False, blank=False, max_length=255)
    note = models.TextField(_("Notiz"), null=True, blank=True, default=None, max_length=100_000)


class TeamMember(NanoIdPkMixin, CreatedMixin, models.Model):

    class Meta:
        verbose_name = _("Team-Mitglied")
        verbose_name_plural = _("Team-Mitglieder")
        default_permissions = ()
        unique_together = (
            ("team_fk", "member_fk"),
        )

    note = models.TextField(_("Notiz"), null=True, blank=True, default=None, max_length=100_000)
    team_fk = models.ForeignKey(Team, verbose_name=_("Abhol-Team"), null=False, blank=False, on_delete=models.CASCADE)
    member_fk = models.ForeignKey(Member, verbose_name=_("Mitglied"), null=False, blank=False, on_delete=models.CASCADE)


class PlanMemberLink(NanoIdPkMixin, CreatedMixin, LastModifiedMixin, models.Model):

    class Meta:
        verbose_name = _("Einplanung Mitglied")
        verbose_name_plural = _("Einplanungen Mitglied")
        default_permissions = ()
        unique_together = (
            ("date_fk", "member_fk"),
        )

    date_fk = models.ForeignKey(Date, verbose_name=_("Datum"), null=False, blank=False, on_delete=models.CASCADE)
    member_fk = models.ForeignKey(Member, verbose_name=_("Mitglied"), null=False, blank=False, on_delete=models.CASCADE)


class PlanTeamLink(NanoIdPkMixin, CreatedMixin, LastModifiedMixin, models.Model):

    class Meta:
        verbose_name = _("Einplanung Team")
        verbose_name_plural = _("Einplanungen Team")
        default_permissions = ()
        unique_together = (
            ("date_fk", "team_fk"),
        )

    date_fk = models.ForeignKey(Date, verbose_name=_("Datum"), null=False, blank=False, on_delete=models.CASCADE)
    team_fk = models.ForeignKey(Team, verbose_name=_("Team"), null=False, blank=False, on_delete=models.CASCADE)
