from django.contrib.messages.views import SuccessMessageMixin
from django.db import IntegrityError
from django.urls import reverse_lazy, reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import ListView, UpdateView, CreateView, DetailView, DeleteView

from solawi_apps.ahg.forms.ahgs import AhgForm
from solawi_apps.ahg.mixin import LoginRequiredMixin
from solawi_apps.ahg.models import Ahg


class AhgListView(LoginRequiredMixin, ListView):

    template_name = "solawi_apps/ahg/ahg-list.html"

    def get_queryset(self):
        return Ahg.objects.filter(owner_fk=self.request.user).order_by('name')


class AhgEditView(LoginRequiredMixin, SuccessMessageMixin, UpdateView):

    template_name = "solawi_apps/ahg/ahg-form.html"
    form_class = AhgForm
    success_message = _("Abholgemeinschaft angepasst.")

    def get_success_url(self):
        return reverse("solawi_apps.ahg:detail", kwargs={"pk": self.object.id})

    def get_queryset(self):
        return Ahg.objects.filter(owner_fk=self.request.user)


class AhgCreateView(LoginRequiredMixin, SuccessMessageMixin, CreateView):

    template_name = "solawi_apps/ahg/ahg-form.html"
    form_class = AhgForm
    success_message = _("Abholgemeinschaft erstellt.")

    def get_success_url(self):
        return reverse("solawi_apps.ahg:detail", kwargs={"pk": self.object.id})

    def form_valid(self, form):
        instance = form.instance
        instance.owner_fk = self.request.user
        try:
            return super().form_valid(form)
        except IntegrityError:
            form.add_error("name", _("Diese Abholgemeinschaft existiert bereits"))
        except Exception as err:
            form.add_error(None, str(err))
        return super().form_invalid(form)


class AhgDetailView(LoginRequiredMixin, DetailView):

    template_name = "solawi_apps/ahg/ahg.html"

    def get_queryset(self):
        return Ahg.objects.filter(owner_fk=self.request.user)


class AhgDeleteView(LoginRequiredMixin, SuccessMessageMixin, DeleteView):

    template_name = "solawi_apps/ahg/ahg-delete.html"
    success_url = reverse_lazy("solawi_apps.ahg:list")

    def get_success_message(self, cleaned_data):
        return _("Abholgemeinschaft %s gelöscht") % self.object.name

    def get_queryset(self):
        return Ahg.objects.filter(owner_fk=self.request.user)
