from csp.decorators import csp_update
from django.contrib.messages.views import SuccessMessageMixin
from django.core.signing import Signer, BadSignature
from django.http import Http404, HttpResponse, HttpRequest
from django.template.loader import get_template
from django.urls import reverse_lazy, reverse
from django.utils.http import urlsafe_base64_decode
from django.utils.translation import gettext_lazy as _
from django.views import View
from django.views.generic import TemplateView, FormView
from django_eventstream import send_event
from qrcode.constants import ERROR_CORRECT_L
from qrcode.image.svg import SvgFillImage
from qrcode.main import QRCode

from solawi_apps.beamer.channel import SCREEN_CODE_SESSION_KEY
from solawi_apps.beamer.config import CONNECTED_BEAMERS_SESSION_KEY
from solawi_apps.beamer.forms import CodeForm, MessageForm
from solawi_apps.beamer.mixin import LoginRequiredMixin
from solawi_apps.beamer.screen_code import get_screen_code


class BeamerView(TemplateView):

    template_name = "solawi_apps/beamer/beamer.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        if (screen_code := self.request.session.get(SCREEN_CODE_SESSION_KEY, None)) is None:
            screen_code = get_screen_code()
            self.request.session[SCREEN_CODE_SESSION_KEY] = screen_code
        context["screen_code"] = screen_code
        return context

    @csp_update(SCRIPT_SRC="'unsafe-inline'", STYLE_SRC="'unsafe-inline'")
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)


class ConnectionListView(LoginRequiredMixin, SuccessMessageMixin, FormView):

    template_name = "solawi_apps/beamer/connections/list.html"
    form_class = CodeForm
    success_message = _("Verbindung hinzugefügt.")
    success_url = reverse_lazy("solawi_apps.beamer:connections")

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        if (connected_beamers := self.request.session.get(CONNECTED_BEAMERS_SESSION_KEY)) is None:
            connected_beamers = []
            self.request.session[CONNECTED_BEAMERS_SESSION_KEY] = connected_beamers
        context["connected_beamers"] = connected_beamers
        context["abs_beamer_link"] = self.request.build_absolute_uri(
            reverse("solawi_apps.beamer:home"),
        )
        return context

    def form_valid(self, form):
        if (connected_beamers := self.request.session.get(CONNECTED_BEAMERS_SESSION_KEY)) is None:
            connected_beamers = []
        new_beamer = form.cleaned_data["screen_code"]
        if new_beamer not in connected_beamers:
            connected_beamers.append(new_beamer)
        self.request.session[CONNECTED_BEAMERS_SESSION_KEY] = connected_beamers
        return super().form_valid(form)


class DeleteConnectionView(LoginRequiredMixin, SuccessMessageMixin, FormView):

    form_class = CodeForm
    success_message = _("Verbindung entfernt.")
    success_url = reverse_lazy("solawi_apps.beamer:connections")

    def form_valid(self, form):
        if (connected_beamers := self.request.session.get(CONNECTED_BEAMERS_SESSION_KEY)) is None:
            connected_beamers = []
        self.request.session[CONNECTED_BEAMERS_SESSION_KEY] = list(
            filter(lambda x: x != form.cleaned_data["screen_code"], connected_beamers)
        )
        return super().form_valid(form)

    def form_invalid(self, form):
        raise Http404()

    def get(self, request, *args, **kwargs) -> HttpResponse:
        raise Http404()


class ResetConnectionView(LoginRequiredMixin, SuccessMessageMixin, FormView):

    form_class = CodeForm
    success_message = _("Verbindung zurückgesetzt.")
    success_url = reverse_lazy("solawi_apps.beamer:connections")

    def form_valid(self, form):
        send_event(
            f"beamer-{form.cleaned_data['screen_code']}",
            "message",
            "reset",
        )
        return super().form_valid(form)

    def form_invalid(self, form):
        raise Http404()

    def get(self, request, *args, **kwargs) -> HttpResponse:
        raise Http404()


class SendTestMessageView(LoginRequiredMixin, SuccessMessageMixin, FormView):

    form_class = MessageForm
    success_message = _("Test-Nachricht gesendet.")
    template_name = "solawi_apps/beamer/connections/test.html"

    def get_success_url(self):
        return reverse_lazy("solawi_apps.beamer:connections.test", kwargs={"beamer_id": self.kwargs["beamer_id"]})

    def form_valid(self, form):
        template = get_template("solawi_apps/beamer/slides/test.html")
        rendered = template.render({
            "test_message": form.cleaned_data["message"],
        })
        send_event(
            f"beamer-{self.kwargs['beamer_id']}",
            "message",
            rendered,
        )
        return super().form_valid(form)


class QrCode(View):

    http_method_names = ["get"]

    def get(self, request: HttpRequest, url_signed: str) -> HttpResponse:
        signer = Signer()
        try:
            url = signer.unsign(urlsafe_base64_decode(url_signed).decode("utf-8"))
        except BadSignature:
            raise Http404()
        qr = QRCode(
            error_correction=ERROR_CORRECT_L,
            box_size=24,
            border=4,
            image_factory=SvgFillImage,
        )
        qr.add_data(url)
        qr.make(fit=True)
        img = qr.make_image()
        svg_data = img.to_string(encoding="utf-8")
        return HttpResponse(svg_data, content_type="image/svg+xml", status=200)
