from django.conf import settings
from django.core.validators import MinValueValidator, MaxValueValidator
from django.db import models
from django.utils.translation import gettext_lazy as _

from solawi_apps.bidding.pin import make_event_pin, make_share_pin
from solawi_apps.db_lib.models import NanoIdPkMixin, CreatedMixin, EnabledMixin, ActiveFromMixin, ActiveToMixin, \
    ArchivedMixin, OwnerMixin, LastModifiedMixin


class Event(
    NanoIdPkMixin,
    EnabledMixin,
    ActiveFromMixin,
    ActiveToMixin,
    ArchivedMixin,
    OwnerMixin,
    CreatedMixin,
    LastModifiedMixin,
    models.Model,
):

    class Meta:
        verbose_name = _("Ereignis")
        verbose_name_plural = _("Ereignisse")
        default_permissions = ()
        unique_together = (
            ("owner_fk", "date", "name"),
        )

    name = models.CharField(verbose_name=_("Name"), null=False, blank=False, max_length=250)
    date = models.DateField(verbose_name=_("Datum"), null=False, blank=False)
    event_pin = models.CharField(
        verbose_name=_("Event-PIN"), null=False, blank=False, max_length=12, unique=True, default=make_event_pin
    )

    def __str__(self):
        return f"{self.name} ({self.date})"


class FiscalPlan(NanoIdPkMixin, CreatedMixin, models.Model):

    class Meta:
        verbose_name = _("Finanzplan")
        verbose_name_plural = _("Finanzpläne")
        default_permissions = ()

    event_fk = models.ForeignKey(Event, verbose_name=_("Ereignis"), null=False, blank=False, on_delete=models.CASCADE)
    number_of_months = models.PositiveSmallIntegerField(
        verbose_name=_("Anzahl Monate"), default=12, null=False, blank=False, validators=[
            MinValueValidator(limit_value=1),
            MaxValueValidator(limit_value=24),
        ])
    required_amount = models.DecimalField(
        verbose_name=_("Benötigter Betrag"), null=False, blank=False, decimal_places=2, max_digits=16,
        validators=[
            MinValueValidator(limit_value=0),
            MaxValueValidator(limit_value=99999999999999.99),
        ],
    )
    currency = models.CharField(verbose_name=_("Währung"), null=False, blank=False, default="EUR", max_length=5)
    minimum_bid_percentage_of_soli_average = models.DecimalField(
        verbose_name=_("Minimalgebot in Prozent des mittleren Solidarbeitrags"),
        decimal_places=3,
        null=False,
        blank=False,
        default=0.0,
        max_digits=10,
        validators=[
            MinValueValidator(limit_value=0),
            MaxValueValidator(limit_value=9999999.999),
        ],
    )
    default_bid_percentage_of_soli_average = models.DecimalField(
        verbose_name=_("Standardgebot in Prozent des mittleren Solidarbeitrags"),
        decimal_places=3,
        default=0.0,
        blank=False,
        null=False,
        max_digits=10,
        validators=[
            MinValueValidator(limit_value=0),
            MaxValueValidator(limit_value=9999999.999),
        ],
    )
    default_bid_percentage_of_bids_average = models.DecimalField(
        verbose_name=_("Standardgebot in Prozent des Durchschnitts aller abgegebenen Gebote"),
        decimal_places=3,
        default=0.0,
        blank=False,
        null=False,
        max_digits=10,
        validators=[
            MinValueValidator(limit_value=0),
            MaxValueValidator(limit_value=9999999.999),
        ],
    )
    display_share_divider_1 = models.BooleanField(
        verbose_name=_("Zeige 1/1 Anteil"), null=False, blank=False, default=True
    )
    display_share_divider_2 = models.BooleanField(
        verbose_name=_("Zeige 1/2 Anteil"), null=False, blank=False, default=True
    )
    display_share_divider_3 = models.BooleanField(
        verbose_name=_("Zeige 1/3 Anteil"), null=False, blank=False, default=False
    )
    display_share_divider_4 = models.BooleanField(
        verbose_name=_("Zeige 1/4 Anteil"), null=False, blank=False, default=False
    )
    display_share_divider_5 = models.BooleanField(
        verbose_name=_("Zeige 1/5 Anteil"), null=False, blank=False, default=False
    )
    display_share_divider_6 = models.BooleanField(
        verbose_name=_("Zeige 1/6 Anteil"), null=False, blank=False, default=False
    )
    display_share_divider_7 = models.BooleanField(
        verbose_name=_("Zeige 1/7 Anteil"), null=False, blank=False, default=False
    )
    display_share_divider_8 = models.BooleanField(
        verbose_name=_("Zeige 1/8 Anteil"), null=False, blank=False, default=False
    )
    display_share_divider_9 = models.BooleanField(
        verbose_name=_("Zeige 1/9 Anteil"), null=False, blank=False, default=False
    )
    display_share_divider_10 = models.BooleanField(
        verbose_name=_("Zeige 1/10 Anteil"), null=False, blank=False, default=False
    )

    def __str__(self):
        return _("Version von %s") % self.created


class Share(NanoIdPkMixin, CreatedMixin, LastModifiedMixin, models.Model):

    class Meta:
        verbose_name = _("Anteil")
        verbose_name_plural = _("Anteile")
        default_permissions = ()
        unique_together = (
            ("event_fk", "share_id"),
        )

    event_fk = models.ForeignKey(Event, verbose_name=_("Ereignis"), null=False, blank=False, on_delete=models.CASCADE)
    divider = models.PositiveIntegerField(verbose_name=_("Teiler"), null=False, blank=False, default=1, validators=[
        MinValueValidator(limit_value=1),
        MaxValueValidator(limit_value=10),
    ])
    share_id = models.CharField(verbose_name=_("Anteils-ID"), null=False, blank=False, max_length=32)
    share_pin = models.CharField(
        verbose_name=_("Anteils-PIN"), null=False, blank=False, max_length=6, default=make_share_pin
    )
    owner_name = models.CharField(verbose_name=_("Anteilseigner"), null=True, blank=True, default=None, max_length=200)
    owner_email = models.EmailField(verbose_name=_("Anteilseigner E-Mail"), null=True, blank=True, default=None)


class PreBid(NanoIdPkMixin, CreatedMixin, LastModifiedMixin, models.Model):

    class Meta:
        verbose_name = _("Vorab-Gebot")
        verbose_name_plural = _("Vorab-Gebote")
        default_permissions = ()
        unique_together = (
            ("share_fk", "from_round"),
        )

    share_fk = models.ForeignKey(Share, verbose_name=_("Anteil"), null=False, blank=False, on_delete=models.CASCADE)
    from_round = models.PositiveIntegerField(
        verbose_name=_("Gültig ab Runde"),
        default=1,
        null=False,
        blank=False,
        validators=[
            MinValueValidator(limit_value=1),
        ],
    )
    fixed_value = models.DecimalField(
        verbose_name=_("Fester Gebotsbetrag"), decimal_places=2, default=0.00, null=False, blank=False, max_digits=16,
        validators=[
            MinValueValidator(limit_value=0),
        ],
    )
    percentage_of_average_soli = models.DecimalField(
        verbose_name=_("Prozentsatz vom mittleren Solidarbeitrag"),
        decimal_places=3,
        default=0.000,
        null=False,
        blank=False,
        max_digits=10,
        validators=[
            MinValueValidator(limit_value=0),
        ],
    )
    percentage_of_average_bid = models.DecimalField(
        verbose_name=_("Prozentsatz vom Mittelwert aller abgegebenen Gebote"),
        decimal_places=3,
        default=0.000,
        null=False,
        blank=False,
        max_digits=10,
        validators=[
            MinValueValidator(limit_value=0),
        ],
    )


class Round(NanoIdPkMixin, CreatedMixin, LastModifiedMixin, models.Model):

    class Meta:
        verbose_name = _("Gebots-Runde")
        verbose_name_plural = _("Gebots-Runden")
        default_permissions = ()

    fiscal_plan_fk = models.ForeignKey(
        FiscalPlan, verbose_name=_("Finanzplan"), null=False, blank=False, on_delete=models.CASCADE
    )
    name = models.CharField(verbose_name=_("Runden-Name"), max_length=200, null=False, blank=False)
    closed = models.BooleanField(verbose_name=_("Runde geschlossen"), default=False, null=False, blank=False)
    result = models.TextField(verbose_name=_("Ergebnis"), null=True, blank=False, default=None, max_length=10_000_000)
    traces = models.TextField(
        verbose_name=_("Entstehungen"), null=True, blank=False, default=None, max_length=100_000_000
    )


class Bid(NanoIdPkMixin, CreatedMixin, LastModifiedMixin, models.Model):

    class Meta:
        verbose_name = _("Gebot")
        verbose_name_plural = _("Gebote")
        unique_together = (
            ("round_fk", "share_fk"),
        )
        default_permissions = ()

    round_fk = models.ForeignKey(Round, verbose_name=_("Runde"), null=False, blank=False, on_delete=models.CASCADE)
    share_fk = models.ForeignKey(Share, verbose_name=_("Anteil"), null=False, blank=False, on_delete=models.CASCADE)
    value = models.DecimalField(verbose_name=_("Wert"), decimal_places=2, null=False, blank=False, max_digits=16)
    origin = models.TextField(verbose_name=_("Gebotsherkunft"), null=False, blank=False, max_length=100_000)


class TeamMember(
    NanoIdPkMixin, CreatedMixin, LastModifiedMixin, EnabledMixin, ActiveFromMixin, ActiveToMixin, models.Model
):

    class Meta:
        verbose_name = _("Teammitglied")
        verbose_name_plural = _("Teammitglieder")
        default_permissions = ()
        unique_together = (
            ("event_fk", "user_fk"),
        )

    user_fk = models.ForeignKey(
        settings.AUTH_USER_MODEL, verbose_name=_("Benutzer"), null=False, blank=False, on_delete=models.CASCADE,
    )
    event_fk = models.ForeignKey(Event, verbose_name=_("Ereignis"), null=False, blank=False, on_delete=models.CASCADE)
    perm_start_round = models.BooleanField(
        verbose_name=_("Berechtigung: Runde starten"), null=False, blank=False, default=False,
    )
    perm_end_round = models.BooleanField(
        verbose_name=_("Berechtigung: Runde beenden"), null=False, blank=False, default=False,
    )
    perm_get_report = models.BooleanField(
        verbose_name=_("Berechtigung: Auswertung abrufen"), null=False, blank=False, default=False,
    )
    perm_add_bid = models.BooleanField(
        verbose_name=_("Berechtigung: Gebot erfassen"), null=False, blank=False, default=False,
    )
    perm_add_prebid = models.BooleanField(
        verbose_name=_("Berechtigung: Vorabgebot erfassen"), null=False, blank=False, default=False,
    )
