from datetime import datetime
from typing import Literal, Optional, Annotated, List

import annotated_types
from django.utils.timezone import now
from pydantic import BaseModel, Field, StringConstraints


OriginType = Literal["WEB_BID", "PRE_BID", "DEFAULT_BID", "TEAM_BID"]
IpType = Annotated[str, StringConstraints(
    max_length=255,
    strict=True,
    strip_whitespace=True,
)]
IdentifierType = Annotated[str, StringConstraints(
    max_length=1023,
    strict=True,
    strip_whitespace=True,
)]
RoundNumberType = Annotated[int, annotated_types.Ge(1)]
BidValueType = Annotated[float, annotated_types.Ge(0.0)]
NameType = Annotated[str, StringConstraints(
    strip_whitespace=True,
    strict=True,
    min_length=1,
    max_length=255,
)]
EMailType = Annotated[str, StringConstraints(
    strip_whitespace=True,
    strict=True,
    min_length=3,
    max_length=255,
)]
ShareIdType = Annotated[str, StringConstraints(
    strip_whitespace=True,
    strict=True,
    min_length=1,
    max_length=255,
)]
DescriptionType = Annotated[str, StringConstraints(
    strip_whitespace=True,
    strict=True,
    min_length=1,
    max_length=50_000,
)]


class BidOrigin(BaseModel):

    origin: OriginType = Field()
    identifier: Optional[IdentifierType] = Field(default=None)
    timestamp: datetime = Field(default_factory=now)
    ip_address: Optional[IpType] = Field(default=None)


class Owner(BaseModel):

    name: Optional[NameType] = Field(default=None)
    email: Optional[EMailType] = Field(default=None)


class Share(BaseModel):

    share_id: ShareIdType = Field()
    share_owner: Owner = Field()


class Bid(BaseModel):

    share: Share = Field()
    origin: BidOrigin = Field()
    value: BidValueType = Field()


class RoundResult(BaseModel):

    round_number: RoundNumberType = Field()
    bids: List[Bid] = Field(default_factory=list)
    total_amount: BidValueType = Field()
    closed_at: datetime = Field(default_factory=now)


class BidComponent(BaseModel):

    name: NameType = Field()
    description: Optional[DescriptionType] = Field(default=None)
    value: BidValueType = Field(default=0.0)


class BidTrace(BaseModel):

    share_id: ShareIdType = Field()
    value: BidValueType = Field()
    components: List[BidComponent] = Field(default_factory=list)


class BidTraces(BaseModel):

    traces: List[BidTrace] = Field(default_factory=list)
