from json import loads
from typing import List, Tuple, Dict, Optional

from django.db import transaction
from django.http import Http404
from django.template.loader import get_template
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from solawi_apps.bidding.models import Event, FiscalPlan, Round, Share, PreBid, Bid, TeamMember
from solawi_apps.bidding.object_model import BidOrigin, RoundResult, Bid as MBid, Share as MShare, Owner, BidTrace, \
    BidComponent, BidTraces
from solawi_apps.bidding.pdf import render_pdf
from solawi_apps.bidding.shares import get_sum_of_shares


def can_start_round(event: Event) -> bool:
    t_now = now()
    if not FiscalPlan.objects.filter(
        event_fk=event, event_fk__enabled=True, event_fk__active_from__lte=t_now, event_fk__active_to__gte=t_now,
    ).exists():
        return False
    if Round.objects.filter(
        fiscal_plan_fk__event_fk=event,
        closed=False,
        fiscal_plan_fk__event_fk__enabled=True,
        fiscal_plan_fk__event_fk__active_from__lte=t_now,
        fiscal_plan_fk__event_fk__active_to__gte=t_now,
    ).exists():
        return False
    if not Share.objects.filter(
        event_fk=event, event_fk__enabled=True, event_fk__active_from__lte=t_now, event_fk__active_to__gte=t_now,
    ).exists():
        return False
    return True


def team_can_start_round(tm: TeamMember) -> bool:
    if not tm.perm_start_round:
        return False
    return can_start_round(tm.event_fk)


def can_stop_round(event: Event) -> bool:
    return Round.objects.filter(fiscal_plan_fk__event_fk=event, closed=False).exists()


def team_can_stop_round(tm: TeamMember) -> bool:
    if not tm.perm_end_round:
        return False
    return can_stop_round(tm.event_fk)


class Tracing:

    def __init__(self):
        self.__traces: Dict[str, List[Tuple[str, str, float]]] = {}

    def add_trace(self, share_id: str, name: str, text: str, value: float):
        if share_id not in self.__traces:
            self.__traces[share_id] = []
        self.__traces[share_id].append((name, text, value))

    def get(self, share_id: str) -> Optional[List[Tuple[str, str, float]]]:
        return self.__traces.get(share_id, None)

    def get_traced_shares(self) -> List[str]:
        return [*self.__traces.keys()]


def close_round(event: Event) -> Round:
    sum_of_shares = get_sum_of_shares(event)
    traces = Tracing()
    with (transaction.atomic()):
        tx_id = transaction.savepoint()
        current_round_number = Round.objects.filter(fiscal_plan_fk__event_fk=event).count()
        try:
            bid_round = Round.objects.filter(
                fiscal_plan_fk__event_fk=event
            ).select_related("fiscal_plan_fk").get(closed=False)
        except (Round.DoesNotExist, Round.MultipleObjectsReturned):
            transaction.savepoint_rollback(tx_id)
            raise Http404()
        required_amount = bid_round.fiscal_plan_fk.required_amount
        currency = bid_round.fiscal_plan_fk.currency
        required_amount_per_full_share = float(required_amount) / sum_of_shares
        required_amount_per_full_share_and_month = (
            required_amount_per_full_share / bid_round.fiscal_plan_fk.number_of_months
        )
        # Check all provided bids
        given_bids_count = 0
        given_bids_total_based_on_full_share = 0.0
        for bid in Bid.objects.filter(round_fk=bid_round).select_related("share_fk").all():
            traces.add_trace(
                share_id=bid.share_fk.share_id,
                name=str(_("Gebotsrunde Nr. %d") % current_round_number),
                text=str(_("Es wurde ein Gebot von %(val).2f %(cur)s für 1/%(div)d Anteil abgegeben") % {
                    "val": bid.value,
                    "cur": currency,
                    "div": bid.share_fk.divider,
                }),
                value=float(bid.value),
            )
            given_bids_count += 1
            given_bids_total_based_on_full_share += float(bid.value) * bid.share_fk.divider
        if given_bids_count > 0:
            given_bids_average = given_bids_total_based_on_full_share / float(given_bids_count)
        else:
            given_bids_average = required_amount_per_full_share_and_month
        # Check all Pre-Bids
        pre_bids_done_shares = []
        for pre_bid in PreBid.objects.filter(
            share_fk__event_fk=event, from_round__lte=current_round_number
        ).order_by("share_fk__share_id", "-from_round").select_related("share_fk").all():
            if pre_bid.share_fk.share_id in pre_bids_done_shares:
                continue
            pre_bids_done_shares.append(pre_bid.share_fk.share_id)
            pre_bid_value = 0.0
            if pre_bid.fixed_value > 0:
                pre_bid_value += float(pre_bid.fixed_value)
                traces.add_trace(
                    share_id=pre_bid.share_fk.share_id,
                    name=str(_("Vorabgebot, gültig ab Runde %d") % pre_bid.from_round),
                    text=str(_("Fester Gebotswert")),
                    value=pre_bid.fixed_value,
                )
            if pre_bid.percentage_of_average_soli > 0:
                share_avg_soli = required_amount_per_full_share_and_month / float(pre_bid.share_fk.divider)
                value = round(share_avg_soli * (float(pre_bid.percentage_of_average_soli) / 100.0), 2)
                pre_bid_value += value
                traces.add_trace(
                    share_id=pre_bid.share_fk.share_id,
                    name=str(_("Vorabgebot, gültig ab Runde %d") % pre_bid.from_round),
                    text=str(_("%(prc).3f%% des durchschnittlichen monatlichen Solidarbeitrags für 1/%(div)d Anteil "
                               "von %(soli).2f %(cur)s") % {
                        "prc": pre_bid.percentage_of_average_soli,
                        "div": pre_bid.share_fk.divider,
                        "soli": share_avg_soli,
                        "cur": currency,
                    }),
                    value=value,
                )
            if pre_bid.percentage_of_average_bid > 0:
                share_avg_bid = given_bids_average / float(pre_bid.share_fk.divider)
                value = round(share_avg_bid * (float(pre_bid.percentage_of_average_bid) / 100.0), 2)
                pre_bid_value += value
                traces.add_trace(
                    share_id=pre_bid.share_fk.share_id,
                    name=str(_("Vorabgebot, gültig ab Runde %d") % pre_bid.from_round),
                    text=str(_("%(prc).3f%% des durchschnittlichen Gebots für 1/%(div)d Anteil "
                               "von %(avg_bid).2f %(cur)s") % {
                        "prc": pre_bid.percentage_of_average_bid,
                        "div": pre_bid.share_fk.divider,
                        "avg_bid": share_avg_bid,
                        "cur": currency,
                    }),
                    value=value,
                )
            if (minimum_percentage := float(bid_round.fiscal_plan_fk.minimum_bid_percentage_of_soli_average)) > 0:
                minimum_bid_for_share = round(
                    (required_amount_per_full_share_and_month / float(pre_bid.share_fk.divider)) *
                    (minimum_percentage / 100.0),
                    2,
                )
                if pre_bid_value < minimum_bid_for_share:
                    missing_part = round(minimum_bid_for_share - pre_bid_value, 2)
                    traces.add_trace(
                        share_id=pre_bid.share_fk.share_id,
                        name=str(_("Vorabgebot unterschreitet Mindestgebot")),
                        text=str(_("Anhebung des aktuellen Gebots von %(now).2f %(cur)s um %(add).2f %(cur)s zur "
                                   "Erreichung des Mindestgebots von %(min).2f %(cur)s für einen 1/%(div)d Anteil") % {
                            "now": pre_bid_value,
                            "cur": currency,
                            "add": missing_part,
                            "min": minimum_bid_for_share,
                            "div": pre_bid.share_fk.divider,
                        }),
                        value=missing_part,
                    )
                    pre_bid_value += missing_part
            Bid.objects.create(
                round_fk=bid_round,
                share_fk=pre_bid.share_fk,
                value=pre_bid_value,
                origin=BidOrigin(
                    origin="PRE_BID",
                    identifier=pre_bid.id,
                ).model_dump_json(),
            )
        # All the missing bids...
        for share in Share.objects.filter(event_fk=event).exclude(share_id__in=traces.get_traced_shares()).all():
            share_bid_value = 0.0
            if (percentage := bid_round.fiscal_plan_fk.default_bid_percentage_of_soli_average) > 0:
                share_avg_soli = required_amount_per_full_share_and_month / float(share.divider)
                value = round(share_avg_soli * (float(percentage) / 100.0), 2)
                share_bid_value += value
                traces.add_trace(
                    share_id=share.share_id,
                    name=str(_("Nicht abgegebenes Gebot in Runde %d") % current_round_number),
                    text=str(_("%(prc).3f%% des durchschnittlichen monatlichen Solidarbeitrags für 1/%(div)d Anteil "
                               "von %(soli).2f %(cur)s") % {
                        "prc": percentage,
                        "div": share.divider,
                        "soli": share_avg_soli,
                        "cur": currency,
                    }),
                    value=value,
                )
            if (percentage := bid_round.fiscal_plan_fk.default_bid_percentage_of_bids_average) > 0:
                share_avg_bid = given_bids_average / float(share.divider)
                value = round(share_avg_bid * (float(percentage) / 100.0), 2)
                share_bid_value += value
                traces.add_trace(
                    share_id=share.share_id,
                    name=str(_("Nicht abgegebenes Gebot in Runde %d") % current_round_number),
                    text=str(_("%(prc).3f%% des durchschnittlichen Gebots für 1/%(div)d Anteil "
                               "von %(avg_bid).2f %(cur)s") % {
                        "prc": percentage,
                        "div": share.divider,
                        "avg_bid": share_avg_bid,
                        "cur": currency,
                    }),
                    value=value,
                )
            if (percentage := float(bid_round.fiscal_plan_fk.minimum_bid_percentage_of_soli_average)) > 0:
                minimum_bid_for_share = round(
                    (required_amount_per_full_share_and_month / float(share.divider)) *
                    (percentage / 100.0),
                    2,
                )
                if share_bid_value < minimum_bid_for_share:
                    missing_part = round(minimum_bid_for_share - share_bid_value, 2)
                    traces.add_trace(
                        share_id=share.share_id,
                        name=str(
                            _("Nicht abgegebenes Gebot in Runde %d unterschreitet Mindestgebot") % current_round_number
                        ),
                        text=str(_("Anhebung des aktuellen Gebots von %(now).2f %(cur)s um %(add).2f %(cur)s zur "
                                   "Erreichung des Mindestgebots von %(min).2f %(cur)s für einen 1/%(div)d Anteil") % {
                            "now": share_bid_value,
                            "cur": currency,
                            "add": missing_part,
                            "min": minimum_bid_for_share,
                            "div": share.divider,
                        }),
                        value=missing_part,
                    )
                    share_bid_value += missing_part
            Bid.objects.create(
                round_fk=bid_round,
                share_fk=share,
                value=share_bid_value,
                origin=BidOrigin(
                    origin="DEFAULT_BID",
                ).model_dump_json(),
            )
        # Close the bid round
        bid_round.closed = True
        # Calculate the results
        total_amount_achieved = 0.0
        bids_for_result: List[MBid] = []
        bid_traces: List[BidTrace] = []
        for bid in Bid.objects.filter(round_fk=bid_round).select_related("share_fk").order_by(
            "share_fk__share_id"
        ).all():
            total_amount_achieved += float(bid.value) * float(bid_round.fiscal_plan_fk.number_of_months)
            bids_for_result.append(
                MBid(
                    origin=BidOrigin(**loads(bid.origin)),
                    value=bid.value,
                    share=MShare(
                        share_id=bid.share_fk.share_id,
                        share_owner=Owner(
                            name=bid.share_fk.owner_name,
                            email=bid.share_fk.owner_email,
                        ),
                    ),
                ),
            )
            tracing = traces.get(bid.share_fk.share_id)
            if tracing is not None:
                bid_traces.append(
                    BidTrace(
                        share_id=bid.share_fk.share_id,
                        value=bid.value,
                        components=[
                            BidComponent(
                                name=name,
                                description=text,
                                value=value,
                            ) for name, text, value in tracing
                        ],
                    ),
                )
        bid_round.result = RoundResult(
            round_number=current_round_number,
            bids=bids_for_result,
            total_amount=total_amount_achieved,
        ).model_dump_json()
        bid_round.traces = BidTraces(traces=bid_traces).model_dump_json()
        bid_round.save()
        transaction.savepoint_commit(tx_id)
    return bid_round


def get_round_result_pdf(bid_round: Round) -> bytes:
    printable_template = get_template("solawi_apps/bidding/rounds/printable_sheet.html")
    context = {
        "round": bid_round,
    }
    if bid_round.result is not None:
        context["result"] = RoundResult(**loads(bid_round.result))
    if bid_round.traces is not None:
        context["tracing"] = BidTraces(**loads(bid_round.traces))
    rendered_template = printable_template.render(context)
    return render_pdf(rendered_template)
