from typing import TypeVar, List, Optional, Iterable

from django import template

from solawi_apps.bidding.models import Event, Share, FiscalPlan
from solawi_apps.bidding.shares import get_sum_of_shares

register = template.Library()

T = TypeVar("T")


@register.simple_tag(name="sum_of_shares")
def get_sum_of_shares_tag(event: Event):
    return get_sum_of_shares(event)


@register.inclusion_tag("solawi_apps/bidding/shares/event.tag.html", name="shares_summary")
def get_shares_summary(event: Event):
    sum_shares = get_sum_of_shares(event)
    context = {
        "num_shares": Share.objects.filter(event_fk=event).count(),
        "sum_shares": sum_shares,
    }
    fiscal_plan = FiscalPlan.objects.filter(event_fk=event).order_by("-created").first()
    if fiscal_plan is not None:
        context["currency"] = fiscal_plan.currency
        share_values = []
        if sum_shares > 0:
            for d in range(10):
                divider = d + 1
                d_key = f"display_share_divider_{divider:d}"
                if getattr(fiscal_plan, d_key):
                    share_values.append(
                        (divider, round(float(fiscal_plan.required_amount) / sum_shares / divider / 12.0, 2))
                    )
        context["share_values"] = share_values
    return context


@register.filter(name="in_rows")
def in_rows(items: Iterable[T], number_of_elements_per_row: int = 3) -> List[List[Optional[T]]]:
    rows = []
    row_items = []
    for item in items:
        row_items.append(item)
        if len(row_items) >= number_of_elements_per_row:
            rows.append([*row_items])
            row_items = []
    if len(row_items) > 0:
        while len(row_items) < number_of_elements_per_row:
            row_items.append(None)
        rows.append([*row_items])
    return rows
