from json import loads
from typing import Optional, Dict, Any

from csp.decorators import csp_update
from django.contrib.messages.views import SuccessMessageMixin
from django.http import HttpRequest, Http404
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _
from django.views.generic import ListView, DetailView, DeleteView

from solawi_apps.bidding.mixin import LoginRequiredMixin
from solawi_apps.bidding.models import Event, Round
from solawi_apps.bidding.object_model import RoundResult


class GetEventMixin:

    event = None

    def get_event(
        self,
        request: Optional[HttpRequest] = None,
        event_pk: Optional[str] = None,
        kwargs: Optional[Dict[str, Any]] = None,
    ) -> Event:
        if self.event is not None:
            return self.event
        if request is None:
            request = self.request
        if kwargs is None:
            kwargs = self.kwargs
        if event_pk is not None:
            kwargs['event_pk'] = event_pk
        try:
            self.event = Event.objects.filter(
                owner_fk=request.user, archived=True,
            ).get(id=kwargs.get("event_pk"))
        except (Event.DoesNotExist, Event.MultipleObjectsReturned):
            raise Http404()
        return self.event


class EventArchiveListView(LoginRequiredMixin, ListView):

    paginate_by = 100
    template_name = "solawi_apps/bidding/archive/events.html"

    def get_queryset(self):
        return Event.objects.filter(owner_fk=self.request.user, archived=True).order_by("-date", "-created")


class ArchivedEventView(LoginRequiredMixin, DetailView):

    template_name = "solawi_apps/bidding/archive/event.html"

    def get_queryset(self):
        return Event.objects.filter(owner_fk=self.request.user, archived=True)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({
            "rounds": [{
                "obj": r,
                "res": RoundResult(**loads(r.result))
            } for r in Round.objects.filter(
                fiscal_plan_fk__event_fk=self.object, closed=True,
            ).select_related("fiscal_plan_fk").order_by("created").all()],
        })
        return context


class ArchivedEventDeleteView(LoginRequiredMixin, SuccessMessageMixin, DeleteView):

    success_message = _("Ereignis gelöscht.")
    success_url = reverse_lazy("solawi_apps.bidding:archive")
    template_name = "solawi_apps/bidding/archive/event.delete.html"

    def get_queryset(self):
        return Event.objects.filter(owner_fk=self.request.user, archived=True)


class ArchivedEventRoundResultView(LoginRequiredMixin, GetEventMixin, DetailView):

    template_name = "solawi_apps/bidding/archive/result.html"

    def get_queryset(self):
        return Round.objects.filter(fiscal_plan_fk__event_fk=self.get_event(), result__isnull=False)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["event"] = self.get_event()
        context["result"] = RoundResult(**loads(self.object.result))
        return context


class ArchivedRoundResultsPdfPageView(LoginRequiredMixin, GetEventMixin, DetailView):

    template_name = "solawi_apps/bidding/archive/result.pdf.html"

    def get_queryset(self):
        return Round.objects.filter(fiscal_plan_fk__event_fk=self.get_event(), result__isnull=False)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["event"] = self.get_event()
        return context

    @csp_update(FRAME_SRC="blob:")
    def dispatch(self, request: HttpRequest, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)
