from json import loads

from django.contrib import messages
from django.utils.timezone import now

from solawi_apps.bidding.mixin import LoginRequiredMixin
from django.http import HttpRequest, Http404
from django.shortcuts import redirect
from django.template.loader import get_template
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views import View

from solawi_apps.beamer.qr import make_qr_code_link
from solawi_apps.beamer.sender import send_rendered_slide
from solawi_apps.bidding.models import Round, Event, TeamMember
from solawi_apps.bidding.object_model import RoundResult
from solawi_apps.bidding.views.events import GetEventMixin


class ShowEventView(LoginRequiredMixin, GetEventMixin, View):

    http_method_names = ["post"]
    success_message = _("Ereignis an angeschlossene Beamer gesendet.")

    def post(self, request: HttpRequest, pk: str, *args, **kwargs):
        event = self.get_event(request, event_pk=pk)
        template = get_template("solawi_apps/bidding/slides/event-overview.html")
        go_url = request.build_absolute_uri(
            reverse("solawi_apps.bidding:access.pin"),
        )
        rendered_event = template.render({
            "event": event,
            "go_url": go_url,
            "qr_code": make_qr_code_link(go_url),
        })
        send_rendered_slide(request, rendered_event)
        messages.success(request, self.success_message)
        return redirect(reverse("solawi_apps.bidding:events.manage", kwargs={
            "pk": pk,
        }), permanent=False)


class ShowResultView(LoginRequiredMixin, GetEventMixin, View):

    http_method_names = ["post"]
    success_message = _("Runden-Ergebnis an angeschlossene Beamer gesendet.")

    def post(self, request: HttpRequest, event_pk: str, pk: str, *args, **kwargs):
        event = self.get_event(request, event_pk=event_pk)
        try:
            bid_round = Round.objects.filter(fiscal_plan_fk__event_fk=event).get(id=pk)
        except (Round.DoesNotExist, Round.MultipleObjectsReturned):
            raise Http404()
        result_context = RoundResult(**loads(bid_round.result))
        template = get_template("solawi_apps/bidding/slides/round-result.html")
        rendered = template.render({
            "result": result_context,
            "fiscal_plan": bid_round.fiscal_plan_fk,
        })
        send_rendered_slide(request, rendered)
        messages.success(request, self.success_message)
        return redirect(reverse("solawi_apps.bidding:events.round.result", kwargs={
            "event_pk": event_pk,
            "pk": pk,
        }), permanent=False)


class ShowTeamEventView(LoginRequiredMixin, View):

    http_method_names = ["post"]
    success_message = _("Ereignis an angeschlossene Beamer gesendet.")

    def post(self, request: HttpRequest, event_pk: str, pk: str, *args, **kwargs):
        t_now = now()
        try:
            team_member = TeamMember.objects.filter(
                user_fk=self.request.user,
                enabled=True,
                active_from__lte=t_now,
                active_to__gte=t_now,
                event_fk__id=event_pk,
                event_fk__archived=False,
                event_fk__enabled=True,
                event_fk__active_from__lte=t_now,
                event_fk__active_to__gte=t_now,
            ).select_related("event_fk").get(id=pk)
        except (TeamMember.DoesNotExist, TeamMember.MultipleObjectsReturned):
            raise Http404()
        template = get_template("solawi_apps/bidding/slides/event-overview.html")
        go_url = request.build_absolute_uri(
            reverse("solawi_apps.bidding:access.pin"),
        )
        rendered_event = template.render({
            "event": team_member.event_fk,
            "go_url": go_url,
            "qr_code": make_qr_code_link(go_url),
        })
        send_rendered_slide(request, rendered_event)
        messages.success(request, self.success_message)
        return redirect(reverse("solawi_apps.bidding:team.event", kwargs={
            "event_pk": event_pk,
            "pk": pk,
        }), permanent=False)


class ShowTeamResultView(LoginRequiredMixin, View):

    http_method_names = ["post"]
    success_message = _("Runden-Ergebnis an angeschlossene Beamer gesendet.")

    def post(self, request: HttpRequest, event_pk: str, team_member_pk: str, pk: str, *args, **kwargs):
        t_now = now()
        try:
            tm = TeamMember.objects.filter(
                user_fk=self.request.user,
                enabled=True,
                active_from__lte=t_now,
                active_to__gte=t_now,
                event_fk__id=event_pk,
                event_fk__enabled=True,
                event_fk__archived=False,
                event_fk__active_from__lte=t_now,
                event_fk__active_to__gte=t_now,
            ).only("event_fk__id").get(id=team_member_pk)
            bid_round = Round.objects.filter(
                fiscal_plan_fk__event_fk__id=tm.event_fk.id,
                closed=True,
                result__isnull=False,
            ).select_related("fiscal_plan_fk").get(id=pk)
        except (
            TeamMember.DoesNotExist, TeamMember.MultipleObjectsReturned,
            Round.DoesNotExist, Round.MultipleObjectsReturned,
        ):
            raise Http404()
        result_context = RoundResult(**loads(bid_round.result))
        template = get_template("solawi_apps/bidding/slides/round-result.html")
        rendered = template.render({
            "result": result_context,
            "fiscal_plan": bid_round.fiscal_plan_fk,
        })
        send_rendered_slide(request, rendered)
        messages.success(request, self.success_message)
        return redirect(reverse("solawi_apps.bidding:team.event.round.result", kwargs={
            "event_pk": event_pk,
            "team_member_pk": team_member_pk,
            "pk": pk,
        }), permanent=False)
