from datetime import timedelta
from typing import Optional, Any, Dict

from django.contrib.auth import get_user_model
from django.db.models import Q
from django.utils.timezone import now

from solawi_apps.bidding.forms.team import AddTeamMemberForm, TeamMemberForm, TransferEventForm
from solawi_apps.bidding.mixin import LoginRequiredMixin
from django.contrib.messages.views import SuccessMessageMixin
from django.db import IntegrityError, transaction
from django.http import Http404, HttpRequest, HttpResponse
from django.urls import reverse, reverse_lazy
from django.utils.translation import gettext_lazy as _
from django.views import View
from django.views.generic import ListView, DetailView, FormView, UpdateView, DeleteView, CreateView
from qrcode.constants import ERROR_CORRECT_M
from qrcode.image.svg import SvgFillImage
from qrcode.main import QRCode

from solawi_apps.bidding.forms.events import EventForm
from solawi_apps.bidding.models import Event, TeamMember
from solawi_apps.bidding.pin import make_event_pin
from solawi_apps.ui.forms.form import ErrorSensitiveFormViewMixin
from solawi_apps.ui.next_url import get_from_url, get_next_url
from solawi_apps.ui.views.generic import ConfirmationView
from solawi_apps.url.params import encrypt_str_parameter, decrypt_str_parameter, EncryptedParameter


class GetEventMixin:

    event = None

    def get_event(
        self,
        request: Optional[HttpRequest] = None,
        event_pk: Optional[str] = None,
        kwargs: Optional[Dict[str, Any]] = None,
    ) -> Event:
        if self.event is not None:
            return self.event
        if request is None:
            request = self.request
        if kwargs is None:
            kwargs = self.kwargs
        if event_pk is not None:
            kwargs['event_pk'] = event_pk
        try:
            self.event = Event.objects.filter(
                owner_fk=request.user, archived=False,
            ).get(id=kwargs.get("event_pk"))
        except (Event.DoesNotExist, Event.MultipleObjectsReturned):
            raise Http404()
        return self.event


class BiddingEventsView(LoginRequiredMixin, ListView):

    paginate_by = 100
    template_name = "solawi_apps/bidding/events.html"

    def get_queryset(self):
        return Event.objects.filter(owner_fk=self.request.user, archived=False).order_by("-date", "-created")


class NewBiddingEventView(LoginRequiredMixin, SuccessMessageMixin, ErrorSensitiveFormViewMixin, FormView):

    template_name = "solawi_apps/bidding/event.form.html"
    form_class = EventForm
    success_message = _("Ereignis erfolgreich angelegt.")

    extra_context = {
        "title": _("Neues Ereignis anlegen"),
        "icon": "fa-solid fa-circle-plus",
        "button": _("Neues Ereignis anlegen"),
    }

    def form_valid(self, form):
        try:
            event = Event.objects.create(
                owner_fk=self.request.user,
                **form.cleaned_data,
            )
        except IntegrityError:
            form.add_error(None, _("Es gibt bereits ein Ereignis mit diesem Namen zum angegebenen Zeitpunkt."))
            form.add_error("name", _("Bitte prüfen"))
            form.add_error("date", _("Bitte prüfen"))
            return self.form_invalid(form)
        setattr(self, "event_obj", event)
        return super().form_valid(form)

    def get_success_url(self):
        return reverse("solawi_apps.bidding:events.manage", kwargs={
            "pk": getattr(self, "event_obj").id,
        })


class EditEventView(LoginRequiredMixin, SuccessMessageMixin, ErrorSensitiveFormViewMixin, UpdateView):

    template_name = "solawi_apps/bidding/event.form.html"
    form_class = EventForm
    success_message = _("Ereignis erfolgreich gespeichert.")

    extra_context = {
        "title": _("Ereignis bearbeiten"),
        "icon": "fa-solid fa-pen-to-square",
        "button": _("Änderungen speichern"),
    }

    def get_queryset(self):
        return Event.objects.filter(owner_fk=self.request.user, archived=False)

    def get_success_url(self):
        return reverse("solawi_apps.bidding:events.manage", kwargs={
            "pk": self.object.id,
        })


class ManageEventView(LoginRequiredMixin, DetailView):

    template_name = "solawi_apps/bidding/event.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["title"] = self.object.name
        return context

    def get_queryset(self):
        return Event.objects.filter(owner_fk=self.request.user, archived=False)


class DeleteEventView(LoginRequiredMixin, SuccessMessageMixin, DeleteView):

    success_message = _("Ereignis gelöscht.")
    success_url = reverse_lazy("solawi_apps.bidding:events")
    template_name = "solawi_apps/bidding/event.delete.html"

    def get_queryset(self):
        return Event.objects.filter(owner_fk=self.request.user, archived=False)


class ArchiveEventView(LoginRequiredMixin, SuccessMessageMixin, ConfirmationView):

    template_name = "solawi_apps/bidding/event.archive.html"
    success_message = _("Ereignis archiviert.")
    success_url = reverse_lazy("solawi_apps.bidding:events")

    def get_queryset(self):
        return Event.objects.filter(owner_fk=self.request.user, archived=False)

    def action(self):
        obj = self.get_object()
        obj.archived = True
        obj.save()
        return super().action()


class RenewEventPinView(LoginRequiredMixin, SuccessMessageMixin, ConfirmationView):

    template_name = "solawi_apps/bidding/event.new-pin.html"
    success_message = _("Neue PIN erfolgreich erstellt.")

    def get_success_url(self):
        obj = self.get_object()
        return reverse("solawi_apps.bidding:events.manage", kwargs={"pk": obj.pk})

    def get_queryset(self):
        return Event.objects.filter(owner_fk=self.request.user, archived=False)

    def action(self):
        obj = self.get_object()
        obj.event_pin = make_event_pin()
        obj.save()
        return super().action()


class AccessLinksView(LoginRequiredMixin, DetailView):

    template_name = "solawi_apps/bidding/event.access-link.html"

    def get_queryset(self):
        return Event.objects.filter(owner_fk=self.request.user, archived=False)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        pin_link = self.request.build_absolute_uri(
            reverse("solawi_apps.bidding:access.pin"),
        )
        context["pin_link"] = pin_link
        access_code_obj = encrypt_str_parameter(
            purpose="public-access-bidding",
            payload=self.object.event_pin,
        )
        direct_link = self.request.build_absolute_uri(
            reverse("solawi_apps.bidding:access.link", kwargs={
                "salt": access_code_obj.public_salt,
                "access_code": access_code_obj.payload,
            })
        )
        context["direct_link"] = direct_link
        encrypted_direct_link = encrypt_str_parameter(
            purpose="qr-access-bidding",
            payload=direct_link,
        )
        context["direct_img_url"] = reverse("solawi_apps.bidding:access.qr", kwargs={
            "salt": encrypted_direct_link.public_salt,
            "payload": encrypted_direct_link.payload,
        })
        encrypted_pin_link = encrypt_str_parameter(
            purpose="qr-access-bidding",
            payload=pin_link,
        )
        context["pin_img_url"] = reverse("solawi_apps.bidding:access.qr", kwargs={
            "salt": encrypted_pin_link.public_salt,
            "payload": encrypted_pin_link.payload,
        })
        return context


class AccessLinkQrCodeView(LoginRequiredMixin, View):

    http_method_names = ["get"]

    def get(self, request: HttpRequest, salt: str, payload: str) -> HttpResponse:
        url = decrypt_str_parameter(EncryptedParameter(
            payload=payload,
            public_salt=salt,
            purpose="qr-access-bidding",
        ), ttl=timedelta(hours=1))
        if url is None:
            raise Http404()
        qr = QRCode(
            error_correction=ERROR_CORRECT_M,
            box_size=64,
            border=4,
            image_factory=SvgFillImage,
        )
        qr.add_data(url)
        qr.make(fit=True)
        img = qr.make_image()
        svg_data = img.to_string(encoding="utf-8")
        return HttpResponse(svg_data, content_type="image/svg+xml", status=200)


class TeamMemberView(LoginRequiredMixin, GetEventMixin, ListView):

    template_name = "solawi_apps/bidding/event.team.html"
    paginate_by = 100

    def get_queryset(self):
        return TeamMember.objects.filter(
            event_fk=self.get_event(),
        ).select_related("user_fk").order_by("user_fk__email", "user_fk__last_name", "user_fk__first_name")

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["event"] = self.get_event()
        return context


class AddTeamMemberView(
    LoginRequiredMixin, GetEventMixin, SuccessMessageMixin, ErrorSensitiveFormViewMixin, CreateView
):

    form_class = AddTeamMemberForm
    template_name = "solawi_apps/bidding/event.team.form.html"
    success_message = _("Team erfolgreich erweitert.")
    email_error = _("Wir können kein Benutzerkonto mit dieser E-Mail-Adresse finden.")

    def get_success_url(self):
        return reverse("solawi_apps.bidding:events.team.members", kwargs={
            "event_pk": self.get_event().id,
        })

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["event"] = self.get_event()
        return context

    def form_valid(self, form):
        t_now = now()
        UserModel = get_user_model()
        try:
            user = UserModel.objects.filter(is_active=True).exclude(id=self.request.user.id).filter(
                Q(
                    accesspermission__app_name="solawi_apps.bidding",
                    accesspermission__enabled=True,
                    accesspermission__active_from__lte=t_now,
                    accesspermission__active_to__gte=t_now,
                ) | Q(
                    is_superuser=True,
                )
            ).get(
                email__iexact=form.cleaned_data["email"],
            )
        except (UserModel.DoesNotExist, UserModel.MultipleObjectsReturned):
            form.add_error("email", self.email_error)
            return self.form_invalid(form)
        form.instance.user_fk = user
        form.instance.event_fk = self.get_event()
        try:
            return super().form_valid(form)
        except IntegrityError:
            form.add_error("email", _("Scheinbar gibt es bereits eine Zuordnung für dieses Benutzerkonto."))
            return self.form_invalid(form)


class EditTeamMemberView(
    LoginRequiredMixin, GetEventMixin, SuccessMessageMixin, ErrorSensitiveFormViewMixin, UpdateView
):

    form_class = TeamMemberForm
    template_name = "solawi_apps/bidding/event.team.form.html"
    success_message = _("Teammitglied erfolgreich geändert.")

    def get_queryset(self):
        return TeamMember.objects.filter(
            event_fk=self.get_event(),
        ).select_related("user_fk")

    def get_success_url(self):
        return reverse("solawi_apps.bidding:events.team.members", kwargs={
            "event_pk": self.get_event().id,
        })

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["event"] = self.get_event()
        return context


class DeleteTeamMemberView(LoginRequiredMixin, GetEventMixin, SuccessMessageMixin, DeleteView):

    success_message = _("Teammitglied gelöscht.")
    template_name = "solawi_apps/bidding/event.team.delete.html"

    def get_queryset(self):
        return TeamMember.objects.filter(
            event_fk=self.get_event(),
        ).select_related("user_fk")

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["event"] = self.get_event()
        return context

    def get_success_url(self):
        return reverse("solawi_apps.bidding:events.team.members", kwargs={
            "event_pk": self.get_event().id,
        })


class TransferEventView(LoginRequiredMixin, GetEventMixin, SuccessMessageMixin, ErrorSensitiveFormViewMixin, FormView):

    template_name = "solawi_apps/bidding/event.transfer.html"
    success_message = _("Ereignis erfolgreich transferiert.")
    success_url = reverse_lazy("solawi_apps.bidding:events")
    form_class = TransferEventForm
    email_error = _("Wir können kein Benutzerkonto mit dieser E-Mail-Adresse finden.")

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["event"] = self.get_event()
        return context

    def form_valid(self, form):
        t_now = now()
        UserModel = get_user_model()
        try:
            user = UserModel.objects.filter(is_active=True).filter(
                Q(
                    accesspermission__app_name="solawi_apps.bidding",
                    accesspermission__enabled=True,
                    accesspermission__active_from__lte=t_now,
                    accesspermission__active_to__gte=t_now,
                ) | Q(
                    is_superuser=True,
                )
            ).exclude(id=self.request.user.id).get(
                email__iexact=form.cleaned_data["email"],
            )
        except (UserModel.DoesNotExist, UserModel.MultipleObjectsReturned):
            form.add_error("email", self.email_error)
            return self.form_invalid(form)
        event = self.get_event()
        with transaction.atomic():
            tx_id = transaction.savepoint()
            event.owner_fk = user
            event.save()
            TeamMember.objects.filter(event_fk=event, user_fk=user).delete()
            if form.cleaned_data["stay_in_team"]:
                form.instance.event_fk = event
                form.instance.user_fk = self.request.user
                form.instance.save()
            transaction.savepoint_commit(tx_id)
        return super().form_valid(form)
