from solawi_apps.bidding.mixin import LoginRequiredMixin
from django.contrib.messages.views import SuccessMessageMixin
from django.db import IntegrityError
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import ListView, FormView, UpdateView, DeleteView

from solawi_apps.bidding.forms.prebid import PreBidForm
from solawi_apps.bidding.models import PreBid, FiscalPlan, Share
from solawi_apps.bidding.round import can_stop_round
from solawi_apps.bidding.views.events import GetEventMixin


class GetCurrencyMixin:

    def get_currency(self):
        fiscal_plan = FiscalPlan.objects.filter(event_fk=self.get_event()).order_by("-created").only("currency").first()
        if fiscal_plan is not None:
            return fiscal_plan.currency
        return None


class PrebidListView(LoginRequiredMixin, GetEventMixin, GetCurrencyMixin, ListView):

    template_name = "solawi_apps/bidding/rounds/prebids.list.html"

    def get_context_data(self, *, object_list=None, **kwargs):
        context = super().get_context_data(object_list=object_list, **kwargs)
        context["currency"] = self.get_currency()
        context['event'] = self.get_event()
        return context

    def get_queryset(self):
        return PreBid.objects.filter(share_fk__event_fk=self.get_event()).order_by("share_fk__share_id", "from_round")


class AddPreBidView(LoginRequiredMixin, SuccessMessageMixin, GetEventMixin, GetCurrencyMixin, FormView):

    template_name = "solawi_apps/bidding/rounds/prebids.form.html"
    form_class = PreBidForm
    success_message = _("Vorabgebot erfolgreich angelegt.")

    def get_success_url(self):
        return reverse("solawi_apps.bidding:events.pre-bids", kwargs={"event_pk": self.get_event().id})

    def get_context_data(self, *, object_list=None, **kwargs):
        context = super().get_context_data(object_list=object_list, **kwargs)
        context['event'] = self.get_event()
        context["currency"] = self.get_currency()
        return context

    def form_valid(self, form):
        if can_stop_round(self.get_event()):
            form.add_error(None, _("Aktuell läuft eine Gebotsrunde. Eine Änderung ist nicht möglich."))
            return self.form_invalid(form)
        try:
            share = Share.objects.filter(event_fk=self.get_event()).get(share_id=form.cleaned_data['share_id'])
        except (Share.DoesNotExist, Share.MultipleObjectsReturned):
            form.add_error("share_id", _("Unbekannte Anteils-ID"))
            return self.form_invalid(form)
        try:
            PreBid.objects.create(
                share_fk=share,
                from_round=form.cleaned_data.get('from_round'),
                fixed_value=form.cleaned_data.get('fixed_value'),
                percentage_of_average_soli=form.cleaned_data.get('percentage_of_average_soli'),
                percentage_of_average_bid=form.cleaned_data.get('percentage_of_average_bid'),
            )
        except IntegrityError:
            form.add_error("from_round", _("Es gibt bereits ein Vorabgebot ab der angegebenen Runde"))
            return self.form_invalid(form)
        return super().form_valid(form)


class EditPreBidView(LoginRequiredMixin, SuccessMessageMixin, GetEventMixin, GetCurrencyMixin, UpdateView):

    template_name = "solawi_apps/bidding/rounds/prebids.form.html"
    form_class = PreBidForm
    success_message = _("Vorabgebot angepasst.")

    def get_success_url(self):
        return reverse("solawi_apps.bidding:events.pre-bids", kwargs={"event_pk": self.get_event().id})

    def get_context_data(self, *, object_list=None, **kwargs):
        context = super().get_context_data(object_list=object_list, **kwargs)
        context['event'] = self.get_event()
        context["currency"] = self.get_currency()
        return context

    def get_queryset(self):
        return PreBid.objects.filter(share_fk__event_fk=self.get_event())

    def get_form(self, form_class=None):
        form = super().get_form(form_class=form_class)
        form.fields["share_id"].disabled = True
        return form

    def get_initial(self):
        initial = super().get_initial()
        initial["share_id"] = self.object.share_fk.share_id
        return initial
    
    def form_valid(self, form):
        if can_stop_round(self.get_event()):
            form.add_error(None, _("Aktuell läuft eine Gebotsrunde. Eine Änderung ist nicht möglich."))
            return self.form_invalid(form)
        return super().form_valid(form)


class DeletePreBitView(LoginRequiredMixin, SuccessMessageMixin, GetEventMixin, DeleteView):

    template_name = "solawi_apps/bidding/rounds/prebids.delete.html"
    success_message = _("Vorabgebot gelöscht.")

    def get_success_url(self):
        return reverse("solawi_apps.bidding:events.pre-bids", kwargs={"event_pk": self.get_event().id})

    def get_context_data(self, *, object_list=None, **kwargs):
        context = super().get_context_data(object_list=object_list, **kwargs)
        context['event'] = self.get_event()
        return context

    def get_queryset(self):
        return PreBid.objects.filter(share_fk__event_fk=self.get_event()).select_related("share_fk")
    
    def form_valid(self, form):
        if can_stop_round(self.get_event()):
            form.add_error(None, _("Aktuell läuft eine Gebotsrunde. Eine Änderung ist nicht möglich."))
            return self.form_invalid(form)
        return super().form_valid(form)
