from json import loads

from csp.decorators import csp_update

from solawi_apps.bidding.mixin import LoginRequiredMixin
from django.contrib.messages.views import SuccessMessageMixin
from django.http import HttpRequest, HttpResponse, Http404
from django.urls import reverse
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from django.views import View
from django.views.generic import DetailView

from solawi_apps.bidding.models import Event, FiscalPlan, Round
from solawi_apps.bidding.object_model import RoundResult
from solawi_apps.bidding.round import can_start_round, can_stop_round, close_round, get_round_result_pdf
from solawi_apps.bidding.views.events import GetEventMixin
from solawi_apps.ui.forms.form import ErrorSensitiveFormViewMixin
from solawi_apps.ui.views.generic import ConfirmationView


class StartRoundView(
    LoginRequiredMixin, SuccessMessageMixin, ErrorSensitiveFormViewMixin, ConfirmationView
):

    template_name = "solawi_apps/bidding/rounds/start.html"
    success_message = _("Runde gestartet.")
    pk_url_kwarg = "event_pk"

    def get_queryset(self):
        t_now = now()
        return Event.objects.filter(
            owner_fk=self.request.user, enabled=True, active_from__lte=t_now, active_to__gte=t_now
        )

    def get_success_url(self):
        return reverse("solawi_apps.bidding:events.manage", kwargs={
            "pk": self.kwargs.get("event_pk"),
        })

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["back_link"] = self.get_success_url()
        return context

    def form_valid(self, form):
        if not can_start_round(self.get_object()):
            form.add_error(None, _("Derzeit kann keine Runde gestartet werden."))
            return self.form_invalid(form)
        fiscal_plan = FiscalPlan.objects.filter(event_fk=self.get_object()).order_by("-created").first()
        Round.objects.create(
            fiscal_plan_fk=fiscal_plan,
            closed=False,
        )
        return super().form_valid(form)


class StopRoundView(
    LoginRequiredMixin, SuccessMessageMixin, ErrorSensitiveFormViewMixin, ConfirmationView
):

    template_name = "solawi_apps/bidding/rounds/stop.html"
    success_message = _("Runde erfolgreich beendet.")
    pk_url_kwarg = "event_pk"

    def get_queryset(self):
        t_now = now()
        return Event.objects.filter(
            owner_fk=self.request.user, enabled=True, active_from__lte=t_now, active_to__gte=t_now
        )

    def get_success_url(self):
        return reverse("solawi_apps.bidding:events.round.result", kwargs={
            "event_pk": self.kwargs.get("event_pk"),
            "pk": self.kwargs.get("result_pk"),
        })

    def get_back_url(self):
        return reverse("solawi_apps.bidding:events.manage", kwargs={
            "pk": self.kwargs.get("event_pk"),
        })

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["back_link"] = self.get_back_url()
        return context

    def form_valid(self, form):
        if not can_stop_round(self.get_object()):
            form.add_error(None, _("Aktuell kann keine Runde beendet werden."))
            return self.form_invalid(form)
        try:
            result = close_round(self.get_object())
        except ValueError as err:
            form.add_error(
                None, _("Die Runde nicht abgeschlossen werden: %s") % ' / '.join([str(e) for e in err.args])
            )
            return self.form_invalid(form)
        self.kwargs["result_pk"] = result.id
        return super().form_valid(form)


class RoundResultView(LoginRequiredMixin, GetEventMixin, DetailView):

    template_name = "solawi_apps/bidding/rounds/result.html"

    def get_queryset(self):
        return Round.objects.filter(fiscal_plan_fk__event_fk=self.get_event(), result__isnull=False)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["event"] = self.get_event()
        context["result"] = RoundResult(**loads(self.object.result))
        return context


class RoundResultsPdfPageView(LoginRequiredMixin, GetEventMixin, DetailView):

    template_name = "solawi_apps/bidding/rounds/result.pdf.html"

    def get_queryset(self):
        return Round.objects.filter(fiscal_plan_fk__event_fk=self.get_event(), result__isnull=False)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["event"] = self.get_event()
        return context

    @csp_update(FRAME_SRC="blob:")
    def dispatch(self, request: HttpRequest, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)


class RoundResultPdfView(LoginRequiredMixin, View):

    http_method_names = ["get"]
    raise_exception = True

    def get(self, request: HttpRequest, event_pk: str, pk: str) -> HttpResponse:
        try:
            bid_round = Round.objects.filter(
                fiscal_plan_fk__event_fk__id=event_pk,
                fiscal_plan_fk__event_fk__owner_fk=request.user,
                result__isnull=False,
            ).get(id=pk)
        except (RoundResult.DoesNotExist, RoundResult.MultipleObjectsReturned):
            raise Http404()
        return HttpResponse(
            status=200,
            content_type="application/pdf",
            content=get_round_result_pdf(bid_round),
        )
