import csv
from io import BytesIO
from typing import List, Tuple

from csp.decorators import csp_update
from django.contrib import messages
from solawi_apps.bidding.mixin import LoginRequiredMixin
from django.contrib.messages.views import SuccessMessageMixin
from django.core.exceptions import ValidationError
from django.core.validators import validate_email
from django.db import DatabaseError, transaction, IntegrityError
from django.db.models.functions import Lower
from django.http import HttpRequest, HttpResponse
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views import View
from django.views.generic import ListView, FormView, UpdateView, DeleteView, TemplateView

from solawi_apps.bidding.csv import SolawiCsvDialect
from solawi_apps.bidding.forms.shares import ShareForm, ImportForm
from solawi_apps.bidding.models import Share, Event
from solawi_apps.bidding.round import can_stop_round
from solawi_apps.bidding.shares import get_share_sheet_pdf
from solawi_apps.bidding.views.events import GetEventMixin
from solawi_apps.ui.forms.form import ErrorSensitiveFormViewMixin


class ShareListView(LoginRequiredMixin, ListView):

    template_name = "solawi_apps/bidding/shares/list.html"
    paginate_by = 100

    def get_context_data(self, *, object_list=None, **kwargs):
        context = super().get_context_data(object_list=object_list, **kwargs)
        context["event"] = Event.objects.filter(
            owner_fk=self.request.user, archived=False, id=self.kwargs.get("event_pk"),
        ).first()
        return context

    def get_queryset(self):
        return Share.objects.filter(
            event_fk__id=self.kwargs.get("event_pk"),
            event_fk__owner_fk=self.request.user,
            event_fk__archived=False,
        ).annotate(share_id_lower=Lower("share_id")).order_by("share_id_lower")


class AddShareView(LoginRequiredMixin, SuccessMessageMixin, ErrorSensitiveFormViewMixin, GetEventMixin, FormView):

    template_name = "solawi_apps/bidding/shares/form.html"
    form_class = ShareForm
    success_message = _("Anteil hinzugefügt")

    event = None

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["event"] = self.get_event()
        return context

    def get_success_url(self):
        return reverse("solawi_apps.bidding:events.shares", kwargs={"event_pk": self.get_event().id})

    def form_valid(self, form):
        if can_stop_round(self.get_event()):
            form.add_error(None, _("Aktuell läuft eine Gebotsrunde. Eine Änderung ist nicht möglich."))
            return self.form_invalid(form)
        try:
            Share.objects.create(
                event_fk=self.get_event(),
                **form.cleaned_data,
            )
        except DatabaseError as err:
            form.add_error(
                None, _("Der Anteil kann nicht angelegt werden: %s") % " / ".join(err.args)
            )
            return self.form_invalid(form)
        return super().form_valid(form)


class EditShareView(LoginRequiredMixin, SuccessMessageMixin, ErrorSensitiveFormViewMixin, GetEventMixin, UpdateView):

    template_name = "solawi_apps/bidding/shares/form.html"
    form_class = ShareForm
    success_message = _("Anteil geändert.")

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["event"] = self.get_event()
        return context

    def get_success_url(self):
        return reverse("solawi_apps.bidding:events.shares", kwargs={"event_pk": self.object.event_fk.id})

    def get_queryset(self):
        return Share.objects.filter(event_fk=self.get_event())

    def form_valid(self, form):
        if can_stop_round(self.get_event()):
            form.add_error(None, _("Aktuell läuft eine Gebotsrunde. Eine Änderung ist nicht möglich."))
            return self.form_invalid(form)
        return super().form_valid(form)


class DeleteShareView(LoginRequiredMixin, SuccessMessageMixin, GetEventMixin, DeleteView):

    template_name = "solawi_apps/bidding/shares/delete.html"
    success_message = _("Anteil gelöscht.")

    def get_queryset(self):
        return Share.objects.filter(
            event_fk=self.get_event(),
        )

    def get_success_url(self):
        return reverse("solawi_apps.bidding:events.shares", kwargs={"event_pk": self.object.event_fk.id})

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["event"] = self.get_event()
        return context

    def form_valid(self, form):
        if can_stop_round(self.get_event()):
            form.add_error(None, _("Aktuell läuft eine Gebotsrunde. Eine Änderung ist nicht möglich."))
            return self.form_invalid(form)
        return super().form_valid(form)


class ImportSharesView(LoginRequiredMixin, SuccessMessageMixin, GetEventMixin, FormView):

    template_name = "solawi_apps/bidding/shares/import.html"
    form_class = ImportForm
    success_message = _("Import abgeschlossen.")

    def get_success_url(self):
        return reverse("solawi_apps.bidding:events.shares", kwargs={"event_pk": self.get_event().id})

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["event"] = self.get_event()
        return context

    @staticmethod
    def errors_to_form(form: ImportForm, errors: List[Tuple[int, str, str]]):
        for e in errors:
            form.add_error(None, _("Zeile %(line)d: '%(raw)s' - %(msg)s") % {
                "line": e[0],
                "raw": str(e[1]),
                "msg": str(e[2]),
            })

    def warnings_to_request(self, errors: List[Tuple[int, str, str]]):
        for e in errors:
            messages.warning(self.request, _("Zeile %(line)d: '%(raw)s' - %(msg)s") % {
                "line": e[0],
                "raw": str(e[1]),
                "msg": str(e[2]),
            })

    def import_file(self, file, form: ImportForm):
        csv_data = BytesIO()
        for chunk in file.chunks():
            csv_data.write(chunk)
        data_received = csv_data.getvalue()
        data_decoded = None
        for encoding in ("utf-8-sig", "utf-8", "utf-16"):
            try:
                data_decoded = data_received.decode(encoding)
                break
            except UnicodeDecodeError:
                continue
        if data_decoded is None:
            data_decoded = data_received.decode("latin1")
        reader = csv.reader(
            [d.strip() for d in data_decoded.split("\n")],
            dialect=SolawiCsvDialect,
        )
        errors = []
        overwrite = form.cleaned_data["overwrite_with_same_id"]
        row_num = 0
        event = self.get_event()
        with transaction.atomic():
            tx_id = transaction.savepoint()
            for row in reader:
                row_num += 1
                data = {}
                if len(row) >= 2:
                    data["share_id"] = row[0].strip()
                    try:
                        data["divider"] = int(row[1])
                        if data["divider"] < 1 or data["divider"] > 10:
                            raise ValueError("divider out of range")
                    except ValueError:
                        errors.append((row_num, SolawiCsvDialect.delimiter.join(row), _("Ungültiger Teiler")))
                        continue
                else:
                    errors.append((row_num, SolawiCsvDialect.delimiter.join(row), _("Pflichtfelder fehlen")))
                    continue
                if len(row) >= 3:
                    email = row[2].strip()
                    if len(email) > 0:
                        try:
                            validate_email(email)
                            data["owner_email"] = email
                        except ValidationError:
                            errors.append(
                                (row_num, SolawiCsvDialect.delimiter.join(row), _("Ungültige E-Mail-Adresse")),
                            )
                if len(row) >= 4:
                    name = row[3].strip()
                    if len(name) > 0:
                        data["owner_name"] = name
                try:
                    data["event_fk"] = event
                    try:
                        with transaction.atomic():
                            Share.objects.create(**data)
                    except IntegrityError:
                        if overwrite:
                            with transaction.atomic():
                                Share.objects.filter(event_fk=event, share_id=data["share_id"]).update(**data)
                        else:
                            raise
                except DatabaseError as err:
                    errors.append((row_num, SolawiCsvDialect.delimiter.join(row), _("Datenfehler: %s") % str(err)))
            if len(errors) > 0:
                if form.cleaned_data["ignore_errors"]:
                    transaction.savepoint_commit(tx_id)
                    self.warnings_to_request(errors)
                    return super().form_valid(form)
                else:
                    transaction.savepoint_rollback(tx_id)
                    ImportSharesView.errors_to_form(form, errors)
                    return self.form_invalid(form)
            else:
                transaction.savepoint_commit(tx_id)
                return super().form_valid(form)

    def form_valid(self, form):
        if can_stop_round(self.get_event()):
            form.add_error(None, _("Aktuell läuft eine Gebotsrunde. Eine Änderung ist nicht möglich."))
            return self.form_invalid(form)
        return self.import_file(self.request.FILES["file"], form)


class ShareSheetView(LoginRequiredMixin, GetEventMixin, TemplateView):

    template_name = "solawi_apps/bidding/shares/sheet.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        event = self.get_event()
        context["event"] = event
        return context

    @csp_update(FRAME_SRC="blob:")
    def dispatch(self, request: HttpRequest, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)


class ShareSheetPdfView(LoginRequiredMixin, View):

    http_method_names = ["get"]
    raise_exception = True

    def get(self, request: HttpRequest, event_pk: str) -> HttpResponse:
        shares = Share.objects.filter(
            event_fk__id=event_pk,
            event_fk__owner_fk=request.user,
        ).order_by("share_id").all()
        return HttpResponse(
            status=200,
            content_type="application/pdf",
            content=get_share_sheet_pdf(shares),
        )
