from json import loads

from csp.decorators import csp_update
from django.contrib.messages.views import SuccessMessageMixin
from django.db import transaction, IntegrityError
from django.db.models import Value
from django.http import Http404, HttpRequest, HttpResponse
from django.urls import reverse
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from django.views import View
from django.views.generic import ListView, DetailView, FormView, UpdateView, DeleteView

from solawi_apps.bidding.forms.prebid import PreBidForm
from solawi_apps.bidding.forms.team import TeamAddBidForm
from solawi_apps.bidding.mixin import LoginRequiredMixin
from solawi_apps.bidding.models import TeamMember, FiscalPlan, Round, Bid, Share, PreBid
from solawi_apps.bidding.object_model import RoundResult, BidOrigin
from solawi_apps.bidding.round import team_can_start_round, team_can_stop_round, close_round, get_round_result_pdf, \
    can_stop_round
from solawi_apps.bidding.shares import get_sum_of_shares
from solawi_apps.ui.forms.form import ErrorSensitiveFormViewMixin
from solawi_apps.ui.views.generic import ConfirmationView
from solawi_apps.user.ip import get_ip


class TeamBiddingEventsView(LoginRequiredMixin, ListView):

    paginate_by = 100
    template_name = "solawi_apps/bidding/team/events.html"

    def get_queryset(self):
        t_now = now()
        return TeamMember.objects.filter(user_fk=self.request.user).filter(
            enabled=True,
            active_from__lte=t_now,
            active_to__gte=t_now,
        ).filter(
            event_fk__enabled=True,
            event_fk__archived=False,
            event_fk__active_to__gte=t_now,
            event_fk__active_from__lte=t_now,
        ).select_related("event_fk").order_by("-event_fk__date", "-event_fk__created")


class TeamBiddingEventView(LoginRequiredMixin, DetailView):

    template_name = "solawi_apps/bidding/team/event.html"

    def get_queryset(self):
        t_now = now()
        return TeamMember.objects.filter(
            user_fk=self.request.user,
        ).filter(
            enabled=True,
            active_from__lte=t_now,
            active_to__gte=t_now,
        ).filter(
            event_fk__id=self.kwargs.get("event_pk"),
        ).filter(
            event_fk__enabled=True,
            event_fk__archived=False,
            event_fk__active_to__gte=t_now,
            event_fk__active_from__lte=t_now,
        ).select_related("event_fk")


class TeamStartRoundView(LoginRequiredMixin, SuccessMessageMixin, ErrorSensitiveFormViewMixin, ConfirmationView):

    template_name = "solawi_apps/bidding/team/round.start.html"
    success_message = _("Runde gestartet.")

    def get_queryset(self):
        t_now = now()
        return TeamMember.objects.filter(
            user_fk=self.request.user,
        ).filter(
            enabled=True,
            active_from__lte=t_now,
            active_to__gte=t_now,
            perm_start_round=True,
        ).filter(
            event_fk__id=self.kwargs.get("event_pk"),
        ).filter(
            event_fk__enabled=True,
            event_fk__archived=False,
            event_fk__active_to__gte=t_now,
            event_fk__active_from__lte=t_now,
        ).select_related("event_fk")

    def get_success_url(self):
        return reverse("solawi_apps.bidding:team.event", kwargs={
            "event_pk": self.kwargs.get("event_pk"),
            "pk": self.kwargs.get("pk"),
        })

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["tm"] = self.get_object()
        context["back_link"] = self.get_success_url()
        return context

    def form_valid(self, form):
        if not team_can_start_round(self.get_object()):
            form.add_error(None, _("Derzeit kann keine Runde gestartet werden."))
            return self.form_invalid(form)
        fiscal_plan = FiscalPlan.objects.filter(event_fk=self.get_object().event_fk).order_by("-created").first()
        Round.objects.create(
            fiscal_plan_fk=fiscal_plan,
            closed=False,
        )
        return super().form_valid(form)


class TeamEndRoundView(LoginRequiredMixin, SuccessMessageMixin, ErrorSensitiveFormViewMixin, ConfirmationView):

    template_name = "solawi_apps/bidding/team/round.stop.html"
    success_message = _("Runde erfolgreich beendet.")

    def get_queryset(self):
        t_now = now()
        return TeamMember.objects.filter(
            user_fk=self.request.user,
        ).filter(
            enabled=True,
            active_from__lte=t_now,
            active_to__gte=t_now,
            perm_end_round=True,
        ).filter(
            event_fk__id=self.kwargs.get("event_pk"),
        ).filter(
            event_fk__enabled=True,
            event_fk__archived=False,
            event_fk__active_to__gte=t_now,
            event_fk__active_from__lte=t_now,
        ).select_related("event_fk")

    def get_success_url(self):
        if self.kwargs.get("result_pk") is not None:
            return reverse("solawi_apps.bidding:team.event.round.result", kwargs={
                "event_pk": self.kwargs.get("event_pk"),
                "team_member_pk": self.kwargs.get("pk"),
                "pk": self.kwargs.get("result_pk"),
            })
        return reverse("solawi_apps.bidding:team.event", kwargs={
            "event_pk": self.kwargs.get("event_pk"),
            "pk": self.kwargs.get("pk"),
        })

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["tm"] = self.get_object()
        context["back_link"] = self.get_success_url()
        return context

    def form_valid(self, form):
        if not team_can_stop_round(self.get_object()):
            form.add_error(None, _("Aktuell kann keine Runde beendet werden."))
            return self.form_invalid(form)
        try:
            result = close_round(self.get_object().event_fk)
        except ValueError as err:
            form.add_error(
                None, _("Die Runde nicht abgeschlossen werden: %s") % ' / '.join([str(e) for e in err.args])
            )
            return self.form_invalid(form)
        if self.get_object().perm_get_report:
            self.kwargs["result_pk"] = result.id
        return super().form_valid(form)


class TeamRoundResultView(LoginRequiredMixin, DetailView):

    template_name = "solawi_apps/bidding/team/round.result.html"

    def get_queryset(self):
        t_now = now()
        try:
            tm = TeamMember.objects.filter(
                perm_get_report=True,
                user_fk=self.request.user,
                enabled=True,
                active_from__lte=t_now,
                active_to__gte=t_now,
                event_fk__id=self.kwargs.get("event_pk"),
                event_fk__enabled=True,
                event_fk__archived=False,
                event_fk__active_from__lte=t_now,
                event_fk__active_to__gte=t_now,
            ).only("event_fk__id").get(id=self.kwargs.get("team_member_pk"))
        except (TeamMember.DoesNotExist, TeamMember.MultipleObjectsReturned):
            raise Http404()
        return Round.objects.filter(
            fiscal_plan_fk__event_fk__id=tm.event_fk.id,
            closed=True,
            result__isnull=False,
        ).annotate(tm=Value(tm.id)).select_related("fiscal_plan_fk").select_related("fiscal_plan_fk__event_fk")

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["result"] = RoundResult(**loads(self.object.result))
        return context


class TeamRoundResultsPdfPageView(LoginRequiredMixin, DetailView):

    template_name = "solawi_apps/bidding/team/round.result.pdf.html"

    def get_queryset(self):
        t_now = now()
        try:
            tm = TeamMember.objects.filter(
                perm_get_report=True,
                user_fk=self.request.user,
                enabled=True,
                active_from__lte=t_now,
                active_to__gte=t_now,
                event_fk__id=self.kwargs.get("event_pk"),
                event_fk__enabled=True,
                event_fk__archived=False,
                event_fk__active_from__lte=t_now,
                event_fk__active_to__gte=t_now,
            ).only("event_fk__id").get(id=self.kwargs.get("team_member_pk"))
        except (TeamMember.DoesNotExist, TeamMember.MultipleObjectsReturned):
            raise Http404()
        return Round.objects.filter(
            fiscal_plan_fk__event_fk__id=tm.event_fk.id,
            closed=True,
            result__isnull=False,
        ).annotate(tm=Value(tm.id)).select_related("fiscal_plan_fk").select_related("fiscal_plan_fk__event_fk")

    @csp_update(FRAME_SRC="blob:")
    def dispatch(self, request: HttpRequest, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)


class TeamRoundResultPdfView(LoginRequiredMixin, View):

    http_method_names = ["get"]
    raise_exception = True

    def get(self, request: HttpRequest, event_pk: str, team_member_pk: str, pk: str) -> HttpResponse:
        t_now = now()
        try:
            tm = TeamMember.objects.filter(
                perm_get_report=True,
                user_fk=self.request.user,
                enabled=True,
                active_from__lte=t_now,
                active_to__gte=t_now,
                event_fk__id=event_pk,
                event_fk__enabled=True,
                event_fk__archived=False,
                event_fk__active_from__lte=t_now,
                event_fk__active_to__gte=t_now,
            ).only("event_fk__id").get(id=team_member_pk)
            bid_round = Round.objects.filter(
                fiscal_plan_fk__event_fk__id=tm.event_fk.id,
                result__isnull=False,
            ).get(id=pk)
        except (
            TeamMember.DoesNotExist, TeamMember.MultipleObjectsReturned,
            RoundResult.DoesNotExist, RoundResult.MultipleObjectsReturned,
        ):
            raise Http404()
        return HttpResponse(
            status=200,
            content_type="application/pdf",
            content=get_round_result_pdf(bid_round),
        )


class TeamAddBidView(LoginRequiredMixin, SuccessMessageMixin, ErrorSensitiveFormViewMixin, FormView):

    success_message = _("Gebot erfolgreich erfasst.")
    form_class = TeamAddBidForm
    template_name = "solawi_apps/bidding/team/round.add-bid-form.html"
    object = None

    def get_success_url(self):
        return reverse("solawi_apps.bidding:team.event", kwargs={
            "event_pk": self.kwargs.get("event_pk"),
            "pk": self.kwargs.get("pk"),
        })

    def get_object(self):
        if self.object is None:
            try:
                t_now = now()
                tm = TeamMember.objects.filter(
                    perm_add_bid=True,
                    user_fk=self.request.user,
                    enabled=True,
                    active_from__lte=t_now,
                    active_to__gte=t_now,
                    event_fk__id=self.kwargs.get("event_pk"),
                    event_fk__enabled=True,
                    event_fk__archived=False,
                    event_fk__active_from__lte=t_now,
                    event_fk__active_to__gte=t_now,
                ).select_related("event_fk").get(id=self.kwargs.get("pk"))
                fp = FiscalPlan.objects.filter(event_fk=tm.event_fk).order_by("-created").first()
                if fp is None:
                    raise Http404()
                self.object = {
                    "tm": tm,
                    "fp": fp,
                }
            except (
                TeamMember.DoesNotExist, TeamMember.MultipleObjectsReturned,
            ):
                raise Http404()
        return self.object

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        for k, v in self.get_object().items():
            context[k] = v
        return context

    def form_valid(self, form):
        p_obj = self.get_object()
        if not can_stop_round(p_obj["tm"].event_fk):
            form.add_error(None, _("Aktuell kann kein Gebot erfasst werden, da keine Runde aktiv ist."))
            return self.form_invalid(form)
        with transaction.atomic():
            tx_id = transaction.savepoint()
            round_obj = Round.objects.filter(
                fiscal_plan_fk__event_fk=p_obj["tm"].event_fk,
                fiscal_plan_fk=p_obj["fp"],
                closed=False,
            ).order_by("-created").first()
            if round_obj is None:
                transaction.savepoint_rollback(tx_id)
                form.add_error(None, _("Aktuell kann kein Gebot erfasst werden, da keine Runde aktiv ist."))
                return self.form_invalid(form)
            try:
                share = Share.objects.filter(event_fk=p_obj["tm"].event_fk).get(
                    share_id__iexact=form.cleaned_data["share"]
                )
            except (Share.DoesNotExist, Share.MultipleObjectsReturned):
                transaction.savepoint_rollback(tx_id)
                form.add_error("share", _("Wir können diesen Anteil nicht finden."))
                return self.form_invalid(form)
            if PreBid.objects.filter(share_fk=share).exists():
                transaction.savepoint_rollback(tx_id)
                form.add_error("share", _("Für diesen Anteil liegt bereits ein Vorabgebot vor."))
                return self.form_invalid(form)
            if p_obj["fp"].minimum_bid_percentage_of_soli_average > 0:
                shares = get_sum_of_shares(p_obj["tm"].event_fk)
                average_soli = (
                    (float(p_obj["fp"].required_amount) / shares / float(p_obj["fp"].number_of_months))
                    / float(share.divider)
                )
                minium_bid = round(average_soli * (float(p_obj["fp"].minimum_bid_percentage_of_soli_average) / 100.0), 2)
                if form.cleaned_data["bid"] < minium_bid:
                    transaction.savepoint_rollback(tx_id)
                    form.add_error("bid", _("Das Mindestgebot beträgt %(min).2f %(cur)s.") % {
                        "min": minium_bid,
                        "cur": p_obj["fp"].currency,
                    })
                    return self.form_invalid(form)
            try:
                Bid.objects.create(
                    round_fk=round_obj,
                    share_fk=share,
                    value=form.cleaned_data["bid"],
                    origin=BidOrigin(origin="TEAM_BID", ip_address=get_ip(self.request)).model_dump_json(),
                )
            except IntegrityError:
                transaction.savepoint_rollback(tx_id)
                form.add_error(
                    "share", _("Für diesen Anteil kann kein weiteres Gebot in dieser Tunde abgegeben werden."),
                )
                return self.form_invalid(form)
            transaction.savepoint_commit(tx_id)
        return super().form_valid(form)


class TeamPreBidView(LoginRequiredMixin, ListView):

    template_name = "solawi_apps/bidding/team/prebid.list.html"
    object = None

    def get_object(self):
        if self.object is None:
            t_now = now()
            tm = TeamMember.objects.filter(
                perm_add_prebid=True,
                user_fk=self.request.user,
                enabled=True,
                active_from__lte=t_now,
                active_to__gte=t_now,
                event_fk__id=self.kwargs.get("event_pk"),
                event_fk__enabled=True,
                event_fk__archived=False,
                event_fk__active_from__lte=t_now,
                event_fk__active_to__gte=t_now,
            ).select_related("event_fk").get(id=self.kwargs.get("pk"))
            if (fp := FiscalPlan.objects.filter(event_fk=tm.event_fk).order_by("-created").first()) is None:
                raise Http404()
            self.object = (tm, fp)
        return self.object

    def get_queryset(self):
        try:
            tm, fp = self.get_object()
        except (TeamMember.DoesNotExist, TeamMember.MultipleObjectsReturned):
            raise Http404()
        return PreBid.objects.filter(
            share_fk__event_fk=tm.event_fk,
        ).select_related("share_fk").annotate(curr=Value(fp.currency)).order_by(
            "share_fk__share_id", "from_round"
        )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["tm"], context["fp"] = self.get_object()
        return context


class TeamAddPreBidView(LoginRequiredMixin, SuccessMessageMixin, ErrorSensitiveFormViewMixin, FormView):

    form_class = PreBidForm
    template_name = "solawi_apps/bidding/team/prebid.form.html"
    success_message = _("Vorabgebot erfolgreich angelegt.")

    object = None

    def get_success_url(self):
        return reverse("solawi_apps.bidding:team.event.pre-bids", kwargs={
            "event_pk": self.kwargs["event_pk"],
            "pk": self.kwargs["pk"],
        })

    def get_object(self):
        if self.object is None:
            t_now = now()
            try:
                tm = TeamMember.objects.filter(
                    perm_add_prebid=True,
                    user_fk=self.request.user,
                    enabled=True,
                    active_from__lte=t_now,
                    active_to__gte=t_now,
                    event_fk__id=self.kwargs.get("event_pk"),
                    event_fk__enabled=True,
                    event_fk__archived=False,
                    event_fk__active_from__lte=t_now,
                    event_fk__active_to__gte=t_now,
                ).select_related("event_fk").get(id=self.kwargs.get("pk"))
            except (TeamMember.DoesNotExist, TeamMember.MultipleObjectsReturned):
                raise Http404()
            if (fp := FiscalPlan.objects.filter(event_fk=tm.event_fk).order_by("-created").first()) is None:
                raise Http404()
            self.object = (tm, fp)
        return self.object

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["tm"], context["fp"] = self.get_object()
        return context

    def form_valid(self, form):
        tm, fp = self.get_object()
        if can_stop_round(tm.event_fk):
            form.add_error(None, _("Aktuell läuft eine Gebotsrunde. Eine Änderung ist nicht möglich."))
            return self.form_invalid(form)
        try:
            share = Share.objects.filter(event_fk=tm.event_fk).get(share_id=form.cleaned_data['share_id'])
        except (Share.DoesNotExist, Share.MultipleObjectsReturned):
            form.add_error("share_id", _("Unbekannte Anteils-ID"))
            return self.form_invalid(form)
        try:
            PreBid.objects.create(
                share_fk=share,
                from_round=form.cleaned_data.get('from_round'),
                fixed_value=form.cleaned_data.get('fixed_value'),
                percentage_of_average_soli=form.cleaned_data.get('percentage_of_average_soli'),
                percentage_of_average_bid=form.cleaned_data.get('percentage_of_average_bid'),
            )
        except IntegrityError:
            form.add_error("from_round", _("Es gibt bereits ein Vorabgebot ab der angegebenen Runde"))
            return self.form_invalid(form)
        return super().form_valid(form)


class TeamEditPreBidView(LoginRequiredMixin, SuccessMessageMixin, ErrorSensitiveFormViewMixin, UpdateView):

    form_class = PreBidForm
    template_name = "solawi_apps/bidding/team/prebid.form.html"
    success_message = _("Vorabgebot angepasst.")

    team_object = None

    def get_success_url(self):
        return reverse("solawi_apps.bidding:team.event.pre-bids", kwargs={
            "event_pk": self.kwargs["event_pk"],
            "pk": self.kwargs["team_member_pk"],
        })

    def get_team_object(self):
        if self.team_object is None:
            t_now = now()
            tm = TeamMember.objects.filter(
                perm_add_prebid=True,
                user_fk=self.request.user,
                enabled=True,
                active_from__lte=t_now,
                active_to__gte=t_now,
                event_fk__id=self.kwargs.get("event_pk"),
                event_fk__enabled=True,
                event_fk__archived=False,
                event_fk__active_from__lte=t_now,
                event_fk__active_to__gte=t_now,
            ).select_related("event_fk").get(id=self.kwargs.get("team_member_pk"))
            if (fp := FiscalPlan.objects.filter(event_fk=tm.event_fk).order_by("-created").first()) is None:
                raise Http404()
            self.team_object = (tm, fp)
        return self.team_object

    def get_queryset(self):
        try:
            tm, fp = self.get_team_object()
        except (TeamMember.DoesNotExist, TeamMember.MultipleObjectsReturned):
            raise Http404()
        return PreBid.objects.filter(
            share_fk__event_fk=tm.event_fk,
        ).select_related("share_fk").annotate(curr=Value(fp.currency))

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["tm"], context["fp"] = self.get_team_object()
        return context

    def get_form(self, form_class=None):
        form = super().get_form(form_class=form_class)
        form.fields["share_id"].disabled = True
        return form

    def get_initial(self):
        initial = super().get_initial()
        initial["share_id"] = self.object.share_fk.share_id
        return initial


class TeamDeletePreBidView(LoginRequiredMixin, SuccessMessageMixin, DeleteView):

    template_name = "solawi_apps/bidding/team/prebid.delete.html"
    success_message = _("Vorabgebot gelöscht.")

    team_object = None

    def get_success_url(self):
        return reverse("solawi_apps.bidding:team.event.pre-bids", kwargs={
            "event_pk": self.kwargs["event_pk"],
            "pk": self.kwargs["team_member_pk"],
        })

    def get_team_object(self):
        if self.team_object is None:
            t_now = now()
            tm = TeamMember.objects.filter(
                perm_add_prebid=True,
                user_fk=self.request.user,
                enabled=True,
                active_from__lte=t_now,
                active_to__gte=t_now,
                event_fk__id=self.kwargs.get("event_pk"),
                event_fk__enabled=True,
                event_fk__archived=False,
                event_fk__active_from__lte=t_now,
                event_fk__active_to__gte=t_now,
            ).select_related("event_fk").get(id=self.kwargs.get("team_member_pk"))
            if (fp := FiscalPlan.objects.filter(event_fk=tm.event_fk).order_by("-created").first()) is None:
                raise Http404()
            self.team_object = (tm, fp)
        return self.team_object

    def get_context_data(self, *, object_list=None, **kwargs):
        context = super().get_context_data(object_list=object_list, **kwargs)
        context['tm'], context['fp'] = self.get_team_object()
        return context

    def get_queryset(self):
        try:
            tm, fp = self.get_team_object()
        except (TeamMember.DoesNotExist, TeamMember.MultipleObjectsReturned):
            raise Http404()
        return PreBid.objects.filter(share_fk__event_fk=tm.event_fk).select_related("share_fk")

    def form_valid(self, form):
        tm, fp = self.get_team_object()
        if can_stop_round(tm.event_fk):
            form.add_error(None, _("Aktuell läuft eine Gebotsrunde. Eine Änderung ist nicht möglich."))
            return self.form_invalid(form)
        return super().form_valid(form)
