import secrets

from django.contrib.auth.mixins import UserPassesTestMixin
from django.contrib.messages.views import SuccessMessageMixin
from django.db import IntegrityError
from django.http import HttpRequest, HttpResponse, Http404
from django.shortcuts import redirect
from django.urls import reverse_lazy, reverse
from django.utils.decorators import method_decorator
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from django.views import View
from django.views.decorators.cache import never_cache
from django.views.generic import FormView

from solawi_apps.bidding.forms.access import SelectEventForm, BidForm
from solawi_apps.bidding.models import Event, Round, FiscalPlan, Share, Bid, PreBid
from solawi_apps.bidding.object_model import BidOrigin
from solawi_apps.bidding.shares import get_sum_of_shares
from solawi_apps.ui.forms.dummy import DummyForm
from solawi_apps.ui.forms.form import ErrorSensitiveFormViewMixin
from solawi_apps.ui.next_url import get_from_url
from solawi_apps.url.params import EncryptedParameter, decrypt_str_parameter
from solawi_apps.user.ip import get_ip

EVENT_PIN_SESSION_KEY = "__bidding__event_pin__"
EVENT_ID_SESSION_KEY = "__bidding__event_id__"


def register_for_event(request: HttpRequest, event: Event):
    request.session[EVENT_PIN_SESSION_KEY] = event.event_pin
    request.session[EVENT_ID_SESSION_KEY] = event.id


def check_event(request: HttpRequest) -> bool:
    if (event_id := request.session.get(EVENT_ID_SESSION_KEY, None)) is None:
        return False
    if (event_pin := request.session.get(EVENT_PIN_SESSION_KEY, None)) is None:
        return False
    t_now = now()
    return Event.objects.filter(
        enabled=True, active_from__lte=t_now, active_to__gte=t_now, event_pin=event_pin, id=event_id
    ).exists()


class AccessByPinView(SuccessMessageMixin, ErrorSensitiveFormViewMixin, FormView):

    form_class = SelectEventForm
    template_name = "solawi_apps/bidding/user_access/access-by-pin.html"
    success_url = reverse_lazy("solawi_apps.bidding:if.event")
    success_message = _("Schön, dass du dabei bist!")

    def form_valid(self, form):
        event_pin = form.cleaned_data["event_pin"]
        event_pin = event_pin.replace(" ", "").strip()
        t_now = now()
        try:
            event = Event.objects.filter(
                enabled=True, active_from__lte=t_now, active_to__gte=t_now
            ).get(event_pin=event_pin)
        except (Event.DoesNotExist, Event.MultipleObjectsReturned):
            form.add_error("event_pin", _("Dieses Ereignis ist uns nicht bekannt. Bitte prüfe die PIN."))
            return self.form_invalid(form)
        register_for_event(self.request, event)
        return super().form_valid(form)


class AccessByLinkView(View):

    http_method_names = ["get"]

    def get(self, request: HttpRequest, salt: str, access_code: str) -> HttpResponse:
        encrypted_parameter = EncryptedParameter(
            public_salt=salt,
            payload=access_code,
            purpose="public-access-bidding",
        )
        if (event_pin := decrypt_str_parameter(encrypted_parameter)) is None:
            raise Http404()
        t_now = now()
        try:
            event = Event.objects.filter(
                enabled=True, active_from__lte=t_now, active_to__gte=t_now
            ).get(event_pin=event_pin)
        except (Event.DoesNotExist, Event.MultipleObjectsReturned):
            raise Http404()
        register_for_event(request, event)
        return redirect("solawi_apps.bidding:if.event", permanent=False)


class EventView(UserPassesTestMixin, SuccessMessageMixin, ErrorSensitiveFormViewMixin, FormView):

    raise_exception = False
    login_url = reverse_lazy("solawi_apps.bidding:access.pin")
    template_name = "solawi_apps/bidding/user_access/event.html"
    form_class = BidForm
    success_message = _("Gebot erfolgreich abgegeben.")

    request_context_key = "__bidding_context__"

    def test_func(self):
        return check_event(self.request)

    def get_success_url(self):
        return f"{reverse('solawi_apps.bidding:if.event')}?success=true"

    @method_decorator(never_cache)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    def get_round_context(self) -> dict:
        if (context := getattr(self.request_context_key, self.request_context_key, None)) is not None:
            return context
        context = {}
        t_now = now()
        try:
            event = Event.objects.filter(
                enabled=True, active_from__lte=t_now, active_to__gte=t_now
            ).get(id=self.request.session.get(EVENT_ID_SESSION_KEY))
            context["event"] = event
            fiscal_plan = FiscalPlan.objects.filter(event_fk=event).order_by("-created").first()
            if fiscal_plan is not None:
                context["fiscal_plan"] = fiscal_plan
                bid_round = Round.objects.filter(fiscal_plan_fk=fiscal_plan, closed=False).order_by("-created").first()
                if bid_round is not None:
                    context["round"] = bid_round
        except (Event.DoesNotExist, Event.MultipleObjectsReturned):
            raise Http404()
        setattr(self.request, self.request_context_key, context)
        return context

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(self.get_round_context())
        context["rand"] = secrets.token_hex(nbytes=64)
        if self.request.GET.get("success", None) is not None:
            context["success"] = True
        return context

    def form_valid(self, form):
        r_context = self.get_round_context()
        if r_context.get("round", None) is None:
            form.add_error(None, _("Aktuell läuft keine Gebotsrunde. Bitte versuche es später noch einmal."))
            return self.form_invalid(form)
        has_errors = False
        try:
            share = Share.objects.filter(event_fk=r_context["event"]).get(share_id=form.cleaned_data["share"])
            if share.share_pin != form.cleaned_data["pin"].replace(" ", ""):
                has_errors = True
                form.add_error("pin", _("Bitte überprüfe deine Anteils-PIN."))
        except (Share.DoesNotExist, Share.MultipleObjectsReturned):
            form.add_error("share", _("Dieser Anteil ist uns nicht bekannt."))
            return self.form_invalid(form)
        if PreBid.objects.filter(share_fk=share).exists():
            form.add_error("share", _("Für diesen Anteil liegt bereits ein Vorabgebot vor."))
            return self.form_invalid(form)
        f_plan = r_context["fiscal_plan"]  # type: FiscalPlan
        shares = get_sum_of_shares(r_context["event"])
        average_soli = (float(f_plan.required_amount) / shares / float(f_plan.number_of_months)) / float(share.divider)
        if f_plan.minimum_bid_percentage_of_soli_average > 0:
            minium_bid = round(average_soli * (float(f_plan.minimum_bid_percentage_of_soli_average) / 100.0), 2)
            if form.cleaned_data["bid"] < minium_bid:
                has_errors = True
                form.add_error("bid", _("Das Mindestgebot beträgt %(min).2f %(cur)s.") % {
                    "min": minium_bid,
                    "cur": f_plan.currency,
                })
        if not has_errors:
            try:
                Bid.objects.create(
                    share_fk=share,
                    round_fk=r_context["round"],
                    value=form.cleaned_data["bid"],
                    origin=BidOrigin(origin="WEB_BID", ip_address=get_ip(self.request)).model_dump_json(),
                )
            except IntegrityError:
                form.add_error(None, _("Du hast bereits in dieser Runde für diesen Anteil ein Gebot abgegeben."))
                has_errors = True
        return self.form_invalid(form) if has_errors else super().form_valid(form)


class LeaveEventView(SuccessMessageMixin, FormView):

    form_class = DummyForm
    template_name = "solawi_apps/bidding/user_access/leave.html"
    success_message = _("Bis bald!")
    success_url = reverse_lazy("solawi_apps.bidding:access.pin")

    def form_valid(self, form):
        self.request.session[EVENT_ID_SESSION_KEY] = None
        del self.request.session[EVENT_ID_SESSION_KEY]
        self.request.session[EVENT_PIN_SESSION_KEY] = None
        del self.request.session[EVENT_PIN_SESSION_KEY]
        return super().form_valid(form)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["back_link"] = get_from_url(self.request)
        return context
