import re

from argparse import ArgumentParser
from datetime import datetime
from pathlib import Path
from typing import Tuple, List

from django.conf import settings
from django.core.management import BaseCommand
from django.template.loader import get_template
from django.utils.safestring import mark_safe

from solawi_apps.changelog.object_model import Version as VersionModel, Item as ItemModel, Ticket, \
    ChangelogDataMigration


MIG_MATCH = re.compile(r'^(?P<migration>(?P<version>\d{4})_.*)\.py$')


class ContentElement:

    def __init__(self):
        self._sub_content = []

    def append_sub_content(self, line: str):
        stripped_line = line.strip()
        if stripped_line == "" and len(self._sub_content) == 0:
            return
        self._sub_content.append(stripped_line)

    def parse(self):
        while len(self._sub_content) > 0 and self._sub_content[len(self._sub_content) - 1] == "":
            del self._sub_content[len(self._sub_content) - 1]


class Item(ContentElement):

    @staticmethod
    def __parse_header(header: str) -> Tuple[str, str, List[str], bool, bool]:
        tickets = []
        item_type, rest = header[2:].split(":", maxsplit=1)
        descriptive_parts = rest.rsplit("/", maxsplit=1)
        breaking = False
        db = False
        if len(descriptive_parts) == 2:
            db = "db" in descriptive_parts[1].lower()
            breaking = "break" in descriptive_parts[1].lower()
        w_parts = descriptive_parts[0].split("#")
        if len(w_parts) > 1:
            for p in w_parts[1:]:
                ticket_id = p.strip()
                if len(ticket_id) > 0:
                    tickets.append(ticket_id)
        return item_type.strip().lower(), w_parts[0].strip(), tickets, breaking, db

    def __init__(self, header: str):
        super().__init__()
        self.__type, self.__title, self.__tickets, self.__breaking, self.__db = Item.__parse_header(header)
        self.__note = None

    def parse(self):
        super().parse()
        md = ("\n".join(self._sub_content)).strip()
        self.__note = md if len(md) > 0 else None

    def as_object(self) -> ItemModel:
        return ItemModel(
            is_breaking_change=self.__breaking,
            requires_db_update=self.__db,
            title=self.__title,
            item_type=self.__type,
            tickets=[Ticket(
                ticket_id=t,
                ticket_url=f"https://git.codebau.dev/solawi-suite/platform/issues/{t}",
            ) for t in self.__tickets],
            note=self.__note,
        )


class Version(ContentElement):

    @staticmethod
    def __parse_header(header: str) -> Tuple[str, str, datetime]:
        version_part, rest = header[1:].split(":", maxsplit=1)
        version = version_part.strip()
        rest, date_part = rest.strip().rsplit("/", maxsplit=1)
        date_str = date_part.strip()
        version_name = rest.strip()
        version_date = datetime.strptime(date_str, "%d.%m.%Y")
        version_date = version_date.replace(tzinfo=datetime.now().astimezone().tzinfo)
        return version, version_name, version_date

    def __init__(self, header: str):
        super().__init__()
        self.__version_id, self.__version_name, self.__version_date = Version.__parse_header(header)
        self.__version_note = []
        self.__items = []

    def parse(self):
        super().parse()
        current_item = None
        for line in self._sub_content:
            line = line.strip()
            if line.startswith("## "):
                current_item = Item(line)
                self.__items.append(current_item)
            else:
                if current_item is None:
                    self.__version_note.append(line)
                else:
                    current_item.append_sub_content(line)
        for item in self.__items:
            item.parse()

    def as_object(self) -> VersionModel:
        md = ("\n".join(self.__version_note)).strip()
        return VersionModel(
            items=[item.as_object() for item in self.__items],
            version_identifier=self.__version_id,
            release_date=self.__version_date,
            version_name=self.__version_name,
            note=md if len(md) > 0 else None,
        )


class Command(BaseCommand):

    help = 'Create changelog migration'

    def add_arguments(self, parser: ArgumentParser):
        parser.add_argument("--dry-run", action="store_true", help="parse and display result")
        parser.add_argument("--clear", action="store_true", help="clear CHANGES.md")

    def handle(self, *args, **options):
        changes_file = settings.BASE_DIR / "CHANGES.md"
        with changes_file.open(mode="rt", encoding="utf-8") as changelog_fd:
            changelog = changelog_fd.read()
        changelog = changelog.replace("\r\n", "\n")
        versions = []
        current_version = None
        lines = changelog.strip().split("\n")
        for line in lines:
            if line.startswith("# "):
                current_version = Version(line)
                versions.append(current_version)
                continue
            if current_version is not None:
                current_version.append_sub_content(line)
        for version in versions:
            version.parse()
        data = ChangelogDataMigration(
            versions=[v.as_object() for v in versions],
        )
        if options["dry_run"]:
            self.stdout.write(data.model_dump_json(indent=2))
            return
        migration_path = Path(__file__).parent.parent.parent / "migrations"
        highest_number = 0
        last_migration = None
        for migration in migration_path.glob("*.py"):
            if (match := MIG_MATCH.match(migration.name)) is not None:
                mig_number = int(match.group("version"))
                if mig_number > highest_number:
                    highest_number = mig_number
                    last_migration = match.group("migration")
        migration_filename_base = "%(version)04d_changelog_append" % {
            "version": highest_number + 1,
        }
        json_file = f"{migration_filename_base}.json"
        python_file = f"{migration_filename_base}.py"
        context = {
            "dependency": last_migration,
            "news_file": json_file,
        }
        tpl = get_template("solawi_apps/changelog/code/migration.py.j2")
        migration_content = tpl.render(context)
        migration_file = migration_path / python_file
        with open(migration_file, "wt", encoding="utf-8") as migration_fd:
            migration_fd.write(migration_content)
        migration_json = migration_path / json_file
        with open(migration_json, "wt", encoding="utf-8") as migration_fd:
            migration_fd.write(data.model_dump_json())
        if options["clear"]:
            with open(changes_file, "wt", encoding="utf-8") as changes_fd:
                changes_fd.write("")
        self.stdout.write(self.style.SUCCESS("Done"))
