from datetime import datetime
from typing import List, Annotated, Optional

from pydantic import BaseModel, Field, StringConstraints, AfterValidator

from solawi_apps.changelog.enhancement import ENHANCEMENT_TYPE


def check_item_type(item_type: str) -> str:
    if item_type not in [t[0] for t in ENHANCEMENT_TYPE]:
        raise AssertionError(f"Item type '{item_type}' is not valid")
    return item_type


MarkdownString = Annotated[str, StringConstraints(
    strict=True,
    strip_whitespace=True,
    max_length=100_000,
    min_length=0,
)]

IdentifierString = Annotated[str, StringConstraints(
    strict=True,
    strip_whitespace=True,
    max_length=100,
    min_length=1,
)]

TitleString = Annotated[str, StringConstraints(
    strict=True,
    strip_whitespace=True,
    max_length=10_000,
    min_length=1,
)]

UrlString = Annotated[str, StringConstraints(
    strict=True,
    strip_whitespace=True,
    max_length=255,
    min_length=1,
)]

ItemType = Annotated[str, AfterValidator(check_item_type)]


class Ticket(BaseModel):

    ticket_id: IdentifierString = Field()
    ticket_url: Optional[UrlString] = Field(default=None)


class Item(BaseModel):

    is_breaking_change: bool = Field(default=False)
    requires_db_update: bool = Field(default=False)
    title: TitleString = Field()
    note: Optional[MarkdownString] = Field(default=None)
    item_type: ItemType = Field()

    tickets: List[Ticket] = Field(default_factory=list, max_length=1_000)


class Version(BaseModel):

    version_identifier: IdentifierString = Field()
    release_date: datetime = Field()
    note: Optional[MarkdownString] = Field(default=None)
    version_name: Optional[TitleString] = Field(default=None)

    items: List[Item] = Field(default_factory=list, max_length=100_000)


class ChangelogDataMigration(BaseModel):

    versions: List[Version] = Field(default_factory=list, min_length=1)
