import re
from string import ascii_lowercase, digits
from typing import Optional

from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from nanoid import generate

CHAR_RESERVE = ascii_lowercase + digits
ID_LENGTH = 32
NANO_ID_PATTERN = fr'[{CHAR_RESERVE}]{{{ID_LENGTH:d}}}'
NANO_ID_RE = re.compile(f"^{NANO_ID_PATTERN}$")


def generate_id(length: int = ID_LENGTH, char_reserve: str = CHAR_RESERVE) -> str:
    """
    Generates a random nanoID

    :param length: Length of the generated ID. Defaults to :attr:`ID_LENGTH`.
    :param char_reserve: Characters to use to build an ID. Defaults to :attr:`CHAR_RESERVE`.
    :return: A generated unique ID.
    """
    return generate(char_reserve, length)


def validate_nanoid(nanoid: str, message: Optional[str] = None) -> None:
    """
    Validates a nanoID. Returns `None` if the ID is not valid.

    :param nanoid: The ID to check.
    :param message: The message when the ID is invalid. Defaults to a standard message.
    :return: `None` if the ID is valid.
    :raises ValidationError: If the ID is not valid.
    """
    if nanoid is None or NANO_ID_RE.match(nanoid) is None:
        if message is not None:
            message = _("ungültige nano id")
        raise ValidationError(message)


class NanoIdConverter:
    regex = NANO_ID_PATTERN

    def to_python(self, value) -> str:
        return str(value)

    def to_url(self, value) -> str:
        return str(value)
