"""
Abstract Model Helpers.
"""
from django.conf import settings
from django.db import models
from django.db.models import ForeignKey
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from solawi_apps.db_lib.dbid import generate_id, ID_LENGTH, validate_nanoid
from solawi_apps.db_lib.dt import MIN_DATE, MAX_DATE
from solawi_apps.encrypted_settings.crypt import encrypt_item, decrypt_item


class NanoIdPkMixin(models.Model):
    """
    Add a nanoID Primary Key field.
    """

    class Meta:
        abstract = True

    id = models.CharField(
        verbose_name=_("ID"),
        primary_key=True,
        editable=False,
        default=generate_id,
        max_length=ID_LENGTH,
        validators=[validate_nanoid],
        null=False,
        blank=False,
    )


class CreatedMixin(models.Model):
    """
    Adds a timestamp when this record was created.
    """

    class Meta:
        abstract = True

    created = models.DateTimeField(
        verbose_name=_("Erstellt"),
        auto_now_add=True,
        null=False,
        blank=False,
        editable=False
    )


class LastModifiedMixin(models.Model):
    """
    Adds a timestamp when this record was last modified.
    """

    class Meta:
        abstract = True

    last_modified = models.DateTimeField(
        verbose_name=_("Zuletzt geändert"),
        auto_now=True,
        null=False,
        blank=False,
        editable=False,
    )


class EnabledMixin(models.Model):

    class Meta:
        abstract = True

    enabled = models.BooleanField(verbose_name=_("Aktiviert"), default=False, blank=False, null=False)

    def is_active(self) -> bool:
        if self.enabled:
            if (super_active := getattr(super(), "is_active", None)) is not None:
                if callable(super_active):
                    return super_active()
                if isinstance(super_active, bool):
                    return super_active
            return True
        return False


class ActiveFromMixin(models.Model):

    class Meta:
        abstract = True

    active_from = models.DateTimeField(verbose_name=_("Aktiv ab"), default=MIN_DATE, blank=False, null=False)

    def is_active(self) -> bool:
        if self.active_from <= now():
            if (super_active := getattr(super(), "is_active", None)) is not None:
                if callable(super_active):
                    return super_active()
                if isinstance(super_active, bool):
                    return super_active
            return True
        return False


class ActiveToMixin(models.Model):

    class Meta:
        abstract = True

    active_to = models.DateTimeField(verbose_name=_("Aktiv bis"), default=MAX_DATE, blank=False, null=False)

    def is_active(self) -> bool:
        if now() <= self.active_to:
            if (super_active := getattr(super(), "is_active", None)) is not None:
                if callable(super_active):
                    return super_active()
                if isinstance(super_active, bool):
                    return super_active
            return True
        return False


class OwnerMixin(models.Model):

    class Meta:
        abstract = True

    owner_fk = ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_("Besitzer"),
        null=False,
        blank=False,
        on_delete=models.RESTRICT,
        editable=True
    )


class ArchivedMixin(models.Model):

    class Meta:
        abstract = True

    archived = models.BooleanField(verbose_name=_("Archiviert"), null=False, blank=False, default=False, editable=True)


class EncryptedTextField(models.TextField):
    """
    Transparent encryption for text fields.

    Attention: You cannot search or filter in those fields!
    Attention: Those fields must be very long!
    """

    def get_db_prep_save(self, value, connection):
        """
        Encrypt
        """
        value = super().get_db_prep_save(value, connection)
        if value is not None:
            return encrypt_item(value)

    def from_db_value(self, value, expression, connection, *args):
        """
        Decrypt
        """
        if value is not None:
            return self.to_python(decrypt_item(value))
