import binascii
import os
from base64 import urlsafe_b64encode, urlsafe_b64decode
from typing import Optional, Union

from coincurve import PublicKey
from ecies import encrypt, decrypt
from ecies.utils import hex2sk


class SecretKeyKeeper:

    @classmethod
    def get_secret_key(cls) -> str:
        """
        Get the secret key

        :return: Hex secret key
        :raises ValueError: If no secret key available
        """
        if (secret_key := getattr(cls, "secret_key", None)) is not None:
            return secret_key
        if (key_setup := os.environ.get("SOLAWI_SUITE_SECRET_KEY", None)) is None:
            raise ValueError("no secret key available")
        key_setup = key_setup.strip()
        if key_setup.startswith("0x"):
            setattr(cls, "secret_key", key_setup)
            return key_setup
        if key_setup.startswith("file://"):
            key_file = key_setup[len("file://"):]
            with open(key_file, "rt") as key_fo:
                key_data = key_fo.read().strip()
            setattr(cls, "secret_key", key_data)
            return key_data
        raise ValueError("no secret key available")


def encrypt_item(item: str, public_key: Optional[Union[str, PublicKey]] = None) -> str:
    """
    Encrypt an item with a public key.

    If no public key is provided, it trys to use a derived public key from the secret key.

    :param item: The item to encrypt.
    :param public_key: The public key.
    :return: The encrypted item, already urlsafe-base64-encoded.
    """
    if public_key is None:
        secret_key_hex = SecretKeyKeeper.get_secret_key()
        secret_key = hex2sk(secret_key_hex)
        public_key = secret_key.public_key.format()
    if public_key is None:
        raise ValueError("no public key available")
    return urlsafe_b64encode(encrypt(public_key, item.encode("utf-8"))).decode("utf-8")


def decrypt_item(item: str, secret_key_hex: Optional[str] = None, no_raise: bool = False) -> Optional[str]:
    """
    Decrypt an encrypted item.

    :param item: URL-safe base64-encoded encrypted item.
    :param secret_key_hex: The secret key to use for decryption.
                           If not set, the key is looked up in the :envvar:`SOLAWI_SUITE_SECRET_KEY`.
                           If there's also no secret key, an exception is raised.
    :param no_raise: If no error should be raised (return None instead)
    :return: The decrypted item.
    :raises RuntimeError: If no secret key can be found.
    :raises ValueError: If base64 decoding fails.
    """
    if secret_key_hex is None:
        secret_key_hex = SecretKeyKeeper.get_secret_key()
    try:
        encrypted_data = urlsafe_b64decode(item)
    except binascii.Error as err:
        raise RuntimeError(*err.args) from err
    try:
        return decrypt(secret_key_hex, encrypted_data).decode("utf-8")
    except ValueError:
        if no_raise:
            return None
        else:
            raise
