from argparse import ArgumentParser

from django.conf import settings
from django.core.management import BaseCommand, CommandError
from django.utils.translation import gettext_lazy as _
from ecies.utils import generate_eth_key

from solawi_apps.encrypted_settings.files import PUBLIC_KEY_FILE_NAME_TEMPLATE
from solawi_apps.encrypted_settings.validators import validate_environment_name


class Command(BaseCommand):

    help = str(_("Erzeugt ein Schlüsselpaar für die Verschlüsselung von Einstellungen"))

    def add_arguments(self, parser: ArgumentParser):
        parser.add_argument(
            "--env",
            required=False,
            type=str,
            default=None,
            help=str(_("Name der Zielumgebung, wenn der Public Key direkt gespeichert werden soll.")),
        )
        parser.add_argument(
            "--force", action="store_true", default=False,
            help=str(_("Erzwinge das Überrschreiben eines vorhandenen Public Keys.")),
        )

    def handle(self, *args, **options):
        if (env := options.get("env")) is not None:
            if not validate_environment_name(env, raise_on_validation_error=False):
                raise CommandError(_("ungültiger Umgebungsname"))
        key = generate_eth_key()
        hex_public_key = key.public_key.to_hex()
        hex_secret_key = key.to_hex()
        self.stdout.write(f"Public Key: {hex_public_key}")
        self.stdout.write(f"Secret Key: {hex_secret_key}")
        if env is not None:
            target_file = settings.PUBLIC_KEY_DIR / (PUBLIC_KEY_FILE_NAME_TEMPLATE % env)
            if target_file.exists() and not options["force"]:
                raise CommandError(
                    str(_("Public Key '%s' existiert bereits; mit --force überschreiben.") % str(target_file))
                )
            with open(target_file, "wt", encoding="utf-8") as pk_fd:
                pk_fd.write(hex_public_key)
            self.stdout.write(str(_("Public Key '%s' gespeichert.") % str(target_file)))
