from typing import Optional

from django.conf import settings
from django.contrib.auth.hashers import verify_password
from django.http import HttpRequest, HttpResponse, JsonResponse
from django.utils.timezone import now

from solawi_apps.db_lib.dbid import NANO_ID_RE
from solawi_apps.health.models import AccessToken


class HealthCheckApiAuthMiddleware:

    def __init__(self, get_response):
        self.get_response = get_response
        self.__disable_access_allow_origin_request_key = "__request__disable_access_allow_origin"
        self.access_token_header = settings.HEALTH_CHECK_ACCESS_TOKEN_HEADER
        self.access_denied_response = JsonResponse({
            "status": "access denied",
        }, status=403)

    def __call__(self, request: HttpRequest) -> HttpResponse:
        response = self.get_response(request)
        if getattr(request, self.__disable_access_allow_origin_request_key, False):
            response["Access-Control-Allow-Origin"] = "*"
        return response

    @staticmethod
    def verify_token(token: str) -> bool:
        t_split = token.split(":", 1)
        if len(t_split) != 2:
            return False
        token_id, token_passwd = t_split
        if NANO_ID_RE.match(token_id) is None:
            return False
        t_now = now()
        try:
            token_obj = AccessToken.objects.filter(
                enabled=True,
                active_from__lte=t_now,
                active_to__gte=t_now,
            ).get(id=token_id)
        except (AccessToken.DoesNotExist, AccessToken.MultipleObjectsReturned):
            return False
        verified, _ = verify_password(token_passwd, token_obj.verification_code)
        return verified

    def handle_health_api_auth(self, request: HttpRequest) -> Optional[HttpResponse]:
        if (token := request.headers.get(self.access_token_header, None)) is None:
            return self.access_denied_response
        if not HealthCheckApiAuthMiddleware.verify_token(token):
            return self.access_denied_response
        return None

    def process_view(self, request: HttpRequest, *args, **kwargs) -> Optional[HttpResponse]:
        if (request.resolver_match is not None
                and request.resolver_match.namespaces is not None and 'health' in request.resolver_match.namespaces):
            setattr(request, self.__disable_access_allow_origin_request_key, True)
            return self.handle_health_api_auth(request)
        return None
