from json import loads

from django.db import connection
from django.test import TestCase
from django.urls import reverse

from solawi_apps.health.management.commands.create_health_api_access_token import Command


class HealthCheckApiTestCase(TestCase):

    failed_response = {
        "status": "access denied",
    }

    good_response = {
        "Cache backend: default": "working",
        "DatabaseBackend": "working",
        "DefaultFileStorageHealthCheck": "working",
        "MigrationsHealthCheck": "working"
    }

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        token = Command.get_new_token()
        cls.header_name, cls.header_value = [t.strip() for t in token.split(':', 1)]
        cls.endpoint = reverse("health:health_check:health_check_home")

    def test_without_header(self):
        response = self.client.get(self.endpoint)
        self.assertEqual(response.status_code, 403)
        self.assertDictEqual(self.failed_response, loads(response.content))

    def test_with_bad_headers(self):
        values = (
            'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa',
            '0' * 128,
            ':1:2:3:4:5:6:7:8:9:10',
            '',
        )
        for value in values:
            with self.subTest(value=value):
                response = self.client.get(self.endpoint, headers={
                    self.header_name: value,
                })
                self.assertEqual(response.status_code, 403)
                self.assertDictEqual(self.failed_response, loads(response.content))

    def test_good_header(self):
        if '?mode=memory' in str(connection.settings_dict.get('NAME')):
            self.skipTest("in-memory db not supported")
        response = self.client.get(self.endpoint, headers={
            self.header_name: self.header_value,
        })
        self.assertEqual(response.status_code, 200)
        self.assertDictEqual(self.good_response, loads(response.content))

    def test_bad_mutations(self):
        values = (
            self.header_value[1:],
            self.header_value[:-1],
            self.header_value + "-a",
            "a-" + self.header_value,
            self.header_value.replace(":", "-"),
        )
        for value in values:
            with self.subTest(value=value):
                response = self.client.get(self.endpoint, headers={
                    self.header_name: value,
                })
                self.assertEqual(response.status_code, 403)
                self.assertDictEqual(self.failed_response, loads(response.content))
