from django.contrib import messages
from django.http import HttpRequest, HttpResponse
from django.urls import reverse, reverse_lazy
from django.utils.html import escape
from django.utils.safestring import mark_safe
from django.utils.timezone import now
from django.utils.translation import gettext as _

from solawi_apps.announce.context import add_announcement_to_context
from solawi_apps.passkeys.models import Passkey


class NoPasskeyAvailableMiddleware:

    passkey_space = reverse_lazy("solawi_apps.passkeys:passkeys")
    mark_marked = "__no_passkey__warning_issued__"
    announce_id = "__no_passkey__announce_id__"

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request: HttpRequest) -> HttpResponse:
        if getattr(request, self.mark_marked, None) is None:
            passkey_space = str(self.passkey_space)
            if request.user.is_authenticated:
                t_now = now()
                if not Passkey.objects.filter(
                    owner_fk=request.user, enabled=True, active_to__gte=t_now, active_from__lte=t_now,
                ).exists():
                    setattr(request, self.mark_marked, True)
                    add_announcement_to_context(request, messages.WARNING, mark_safe(
                        '<p><strong>%(message)s</strong><br>%(explanation)s</p>'
                        '<a class="btn btn-sm btn-secondary" href="%(url)s">'
                        '<i class="fa fa-fw fa-solid fa-unlock-keyhole"></i> %(button)s'
                        '</a>' % {
                            "message": escape(_("Du hast aktuell keinen aktiven Passkey.")),
                            "explanation": escape(
                                _("Passkeys ermöglichen dir einen unkomplizierten und sicheren Zugang zur "
                                  "[solawi-suite], ganz ohne Benutzername, Passwort und Einmalpasswort.")
                            ),
                            "url": escape(reverse("solawi_apps.passkeys:passkeys")),
                            "button": escape(_("Passkeys verwalten")),
                        }
                    ), self.announce_id, [passkey_space])
        return self.get_response(request)
