import json
from base64 import b85encode, b85decode
from json import JSONDecodeError

from django.conf import settings
from django.contrib import messages
from django.contrib.auth import login
from django.contrib.auth.mixins import UserPassesTestMixin
from django.http import HttpRequest, HttpResponse, JsonResponse
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from django.views import View
from django.views.generic import TemplateView
from webauthn import generate_authentication_options, options_to_json, verify_authentication_response, \
    base64url_to_bytes
from webauthn.helpers.exceptions import InvalidAuthenticationResponse

from solawi_apps.encrypted_settings.crypt import encrypt_item, decrypt_item
from solawi_apps.passkeys.models import Passkey
from solawi_apps.passkeys.passkey import create_challenge
from solawi_apps.ui.next_url import get_next_url

CHALLENGE_SESSION_KEY = "__passkey_authentication_challenge__"


class PasskeyLoginView(UserPassesTestMixin, TemplateView):

    raise_exception = True
    template_name = "solawi_apps/passkeys/auth/login.html"

    def test_func(self):
        return not self.request.user.is_authenticated

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["next_url"] = get_next_url(self.request)
        context["default_welcome_url"] = settings.LOGIN_REDIRECT_URL
        return context


class PasskeyAuthOptionsView(UserPassesTestMixin, View):

    raise_exception = True
    http_method_names = ["post"]

    def test_func(self):
        return not self.request.user.is_authenticated

    def post(self, request: HttpRequest) -> HttpResponse:
        challenge = create_challenge()
        request.session[CHALLENGE_SESSION_KEY] = encrypt_item(challenge)
        options = generate_authentication_options(
            rp_id=settings.PASSKEYS_RELYING_PARTY.get("id"),
            challenge=challenge.encode("utf-8"),
        )
        return HttpResponse(options_to_json(options), content_type="application/json", status=200)


class PasskeyAuthView(UserPassesTestMixin, View):

    raise_exception = True
    http_method_names = ["post"]

    def test_func(self):
        return not self.request.user.is_authenticated

    def post(self, request: HttpRequest) -> HttpResponse:
        if (expected_challenge := request.session.get(CHALLENGE_SESSION_KEY, None)) is None:
            messages.error(request, _("Deine Passkey-Challenge wurde ungültig beantwortet."))
            return JsonResponse(
                {
                    "error": "challenge not defined",
                },
                status=400,
            )
        del request.session[CHALLENGE_SESSION_KEY]
        try:
            credential = json.loads(request.body.decode("utf-8"))
        except JSONDecodeError:
            messages.error(request, _("Passkey: Ungültiges JSON"))
            return JsonResponse({"error": "invalid JSON"}, status=400)
        if (cred_id := credential.get("id", None)) is None:
            messages.error(request, _("Passkey: Ungültiges JSON"))
            return JsonResponse({"error": "invalid JSON"}, status=400)
        t_now = now()
        try:
            passkey = Passkey.objects.filter(
                enabled=True, active_from__lte=t_now, active_to__gte=t_now, owner_fk__is_active=True
            ).select_related("owner_fk").get(credential_id=b85encode(base64url_to_bytes(cred_id)).decode("utf-8"))
        except (Passkey.DoesNotExist, Passkey.MultipleObjectsReturned):
            messages.error(request, _("Passkey: Unbekannter Schlüssel"))
            return JsonResponse({"error": "unknown key"}, status=401)
        try:
            verification = verify_authentication_response(
                credential=credential,
                expected_challenge=decrypt_item(expected_challenge).encode("utf-8"),
                expected_rp_id=settings.PASSKEYS_RELYING_PARTY.get("id"),
                expected_origin=settings.PASSKEYS_EXPECTED_ORIGINS,
                credential_public_key=b85decode(passkey.public_key),
                credential_current_sign_count=passkey.sign_count,
            )
        except InvalidAuthenticationResponse:
            messages.error(request, _("Passkey: Ungültiger Schlüssel"))
            return JsonResponse({"error": "invalid key"}, status=401)
        passkey.sign_count = verification.new_sign_count
        passkey.save()
        request.session.clear()
        login(request, passkey.owner_fk)
        messages.success(request, _("Herzlich willkommen, %s!") % passkey.owner_fk.first_name)
        return JsonResponse({"status": "ok"}, status=200)
