import json
from base64 import b85encode, b85decode
from json import JSONDecodeError
from uuid import uuid4

from django.conf import settings
from django.contrib import messages
from django.contrib.messages.views import SuccessMessageMixin
from django.db import IntegrityError
from django.db.models.functions import Lower
from django.http import HttpRequest, HttpResponse, JsonResponse
from django.urls import reverse, reverse_lazy
from django.utils.translation import gettext_lazy as _
from django.views import View
from django.views.generic import ListView, TemplateView, UpdateView, DeleteView
from webauthn import generate_registration_options, options_to_json, verify_registration_response
from webauthn.helpers.exceptions import InvalidRegistrationResponse
from webauthn.helpers.structs import PublicKeyCredentialDescriptor

from solawi_apps.encrypted_settings.crypt import encrypt_item, decrypt_item
from solawi_apps.passkeys.forms import PasskeyForm
from solawi_apps.passkeys.mixin import LoginRequiredMixin
from solawi_apps.passkeys.models import Passkey
from solawi_apps.passkeys.passkey import create_challenge


CHALLENGE_SESSION_KEY = "__passkey_generation_challenge__"


class ListPasskeysView(LoginRequiredMixin, ListView):

    template_name = "solawi_apps/passkeys/user/list.html"
    paginate_by = 50

    def get_queryset(self):
        return Passkey.objects.filter(
            owner_fk=self.request.user,
        ).annotate(lower_name=Lower("name")).order_by('lower_name', 'created')


class AddPasskeyView(LoginRequiredMixin, TemplateView):

    template_name = "solawi_apps/passkeys/user/add.html"


class ManagePasskeyView(LoginRequiredMixin, SuccessMessageMixin, UpdateView):

    template_name = "solawi_apps/passkeys/user/edit.html"
    form_class = PasskeyForm
    success_url = reverse_lazy("solawi_apps.passkeys:passkeys")
    success_message = _("Änderungen am Passkey gespeichert.")

    def get_queryset(self):
        return Passkey.objects.filter(owner_fk=self.request.user)

    def form_valid(self, form):
        try:
            return super().form_valid(form)
        except IntegrityError:
            form.add_error("name", _("Dieser Name ist bereits vergeben."))
            return self.form_invalid(form)


class DeletePasskeyView(LoginRequiredMixin, SuccessMessageMixin, DeleteView):

    template_name = "solawi_apps/passkeys/user/delete.html"
    success_url = reverse_lazy("solawi_apps.passkeys:passkeys")
    success_message = _("Passkey gelöscht.")

    def get_queryset(self):
        return Passkey.objects.filter(owner_fk=self.request.user)


class InitCreatePasskeyView(LoginRequiredMixin, View):

    http_method_names = ["post"]

    def post(self, request: HttpRequest) -> HttpResponse:
        passkey_challenge = create_challenge()
        request.session[CHALLENGE_SESSION_KEY] = encrypt_item(passkey_challenge)
        options = generate_registration_options(
            rp_id=settings.PASSKEYS_RELYING_PARTY.get("id"),
            rp_name=settings.PASSKEYS_RELYING_PARTY.get("name"),
            user_id=self.request.user.username.encode("utf-8"),
            user_name=self.request.user.email,
            user_display_name=self.request.user.email,
            challenge=passkey_challenge.encode("utf-8"),
            exclude_credentials=[
                PublicKeyCredentialDescriptor(
                    id=b85decode(p.credential_id),
                ) for p in Passkey.objects.filter(owner_fk=self.request.user).only("credential_id").all()
            ],
        )
        return HttpResponse(options_to_json(options), content_type="application/json", status=200)


class VerifyCreatePasskeyView(LoginRequiredMixin, View):

    http_method_names = ["post"]

    def post(self, request: HttpRequest) -> HttpResponse:
        if (expected_challenge := request.session.get(CHALLENGE_SESSION_KEY, None)) is None:
            return JsonResponse(
                {
                    "error": "challenge not defined",
                },
                status=400,
            )
        del request.session[CHALLENGE_SESSION_KEY]
        try:
            credential = json.loads(request.body.decode("utf-8"))
        except JSONDecodeError:
            return JsonResponse({"error": "invalid JSON"}, status=400)
        try:
            registration = verify_registration_response(
                credential=credential,
                expected_challenge=decrypt_item(expected_challenge).encode("utf-8"),
                expected_rp_id=settings.PASSKEYS_RELYING_PARTY.get("id"),
                expected_origin=settings.PASSKEYS_EXPECTED_ORIGINS,
            )
        except InvalidRegistrationResponse as ex:
            return JsonResponse(
                {
                    "error": str(ex),
                },
                status=400,
            )
        try:
            passkey = Passkey.objects.create(
                owner_fk=request.user,
                enabled=True,
                credential_id=b85encode(registration.credential_id).decode("utf-8"),
                public_key=b85encode(registration.credential_public_key).decode("utf-8"),
                name=str(uuid4()),
                backup_state=registration.credential_backed_up,
                sign_count=registration.sign_count,
            )
        except IntegrityError as ex:
            return JsonResponse(
                {
                    "error": "key already exists",
                },
                status=400,
            )
        messages.success(request, _("Passkey erfolgreich angelegt."))
        return JsonResponse(
            {
                "status": "ok",
                "next": reverse("solawi_apps.passkeys:passkeys.manage", kwargs={"pk": passkey.id}),
            },
            status=201,
        )
