from datetime import timedelta

from django.contrib import admin
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from solawi_apps.user.models import AllowedEmailDomain, Registration


class AllowedEmailDomainAdmin(admin.ModelAdmin):

    ordering = ("domain",)

    search_fields = ("domain",)

    list_filter = ("enabled",)

    @admin.display(description=_("Ist aktuell aktiv"), boolean=True)
    def get_currently_active(self, obj: AllowedEmailDomain) -> bool:
        return obj.is_active()

    list_display = (
        "domain", "get_currently_active", "enabled", "active_from", "active_to",
    )

    readonly_fields = ("get_currently_active", "id", "created", "last_modified")

    fieldsets = (
        (
            _("Domain"),
            {
                "fields": ("domain",),
            },
        ),
        (
            _("Gültigkeit"),
            {
                "fields": ("enabled", "active_from", "active_to",),
            },
        ),
        (
            _("Status"),
            {
                "fields": ("get_currently_active",),
            },
        ),
        (
            _("Meta"),
            {
                "fields": ("id", "created", "last_modified",),
            },
        )
    )


class RegistrationAdmin(admin.ModelAdmin):

    @admin.display(description=_("Kann sich noch registrieren"), boolean=True)
    def get_can_register(self, obj: Registration) -> bool:
        return obj.is_active() and not obj.registration_finished and obj.created >= now() - timedelta(days=2)

    @admin.display(description=_("Ist aktuell aktiv"), boolean=True)
    def get_currently_active(self, obj: Registration) -> bool:
        return obj.is_active()

    @admin.display(description=_("E-Mail"), ordering="user_fk__email")
    def get_email(self, obj: Registration) -> str:
        return obj.user_fk.email

    search_fields = ("user_fk__email",)

    list_display = (
        "get_email",
        "get_can_register",
        "get_currently_active",
        "created",
        "enabled",
        "as_superuser",
        "registration_finished",
    )

    readonly_fields = (
        "created", "last_modified", "id", "get_currently_active", "get_can_register",
        "registration_id", "registration_finished", "registration_check", "get_email",
        "user_fk",
    )

    fieldsets = (
        (_("Registrierung"), {
            "fields": ("get_email", "user_fk")
        }),
        (_("Gültigkeit"), {
            "fields": ("enabled", "active_from", "active_to",),
        }),
        (_("Status"), {
            "fields": ("get_can_register", "get_currently_active",),
        }),
        (_("Meta"), {
            "fields": ("id", "created", "last_modified",),
        })
    )

    def has_add_permission(self, request, obj=None):
        return False


class DomainBasedAppAccessAdmin(admin.ModelAdmin):

    readonly_fields = ("id", "created", "last_modified")

    fieldsets = (
        (_("Berechtigung"), {
            "fields": ("allowed_domain_fk", "app_name")
        }),
        (_("Gültigkeit"), {
            "fields": ("enabled", "active_from", "active_to",),
        }),
        (_("Meta"), {
            "fields": ("id", "created", "last_modified",),
        })
    )

    list_display = ("allowed_domain_fk", "app_name", "enabled", "active_from", "active_to")

    list_filter = ("app_name",)

    ordering = ("allowed_domain_fk", "app_name")
