from typing import Union

from django.contrib import admin
from django.utils.translation import gettext_lazy as _

from solawi_apps.user.ip import get_hostname
from solawi_apps.user.models import TotpSetup, OtpAttempt, LoginAttempt, BurnedOtp, UserSession, ForgotPassword


class TotpSetupAdmin(admin.ModelAdmin):

    @admin.display(description=_("Für Benutzer"), ordering="user_fk__email")
    def get_owner(self, obj: TotpSetup):
        return obj.user_fk.email

    search_fields = ("user_fk__email",)

    ordering = ("user_fk__email",)

    list_display = ("get_owner", "created")

    fieldsets = (
        (_("Besitzer"), {
            "fields": ("get_owner", "user_fk",),
        }),
        (_("TOTP"), {
            "fields": ("totp_provisioning",),
        }),
        (_("Meta"), {
            "fields": ("id", "created", "last_modified",),
        }),
    )

    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False


class AuthAttemptAdmin(admin.ModelAdmin):

    @admin.display(description=_("Für Benutzer"), ordering="user_fk__email")
    def get_owner(self, obj: Union[OtpAttempt, LoginAttempt]):
        return obj.user_fk.email

    @admin.display(description=_("Hostname"))
    def get_ip_hostname(self, obj: Union[OtpAttempt, LoginAttempt]):
        return get_hostname(obj.ip)

    search_fields = ("user_fk__email", "ip",)

    list_display = ("get_owner", "created", "success", "ip",)
    list_filter = ("user_fk__email",)

    ordering = ("-created", "user_fk__email",)

    fieldsets = (
        (_("Besitzer"), {
            "fields": ("get_owner", "user_fk"),
        }),
        (_("Versuch"), {
            "fields": ("ip", "get_ip_hostname", "success",),
        }),
        (_("Meta"), {
            "fields": ("id", "created",),
        }),
    )

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False


class AnonAuthAttemptAdmin(admin.ModelAdmin):

    @admin.display(description=_("Hostname"))
    def get_ip_hostname(self, obj: Union[OtpAttempt, LoginAttempt]):
        return get_hostname(obj.ip)

    search_fields = ("ip",)

    list_display = ("ip", "created",)
    list_filter = ("ip",)

    ordering = ("-created", "ip",)

    fieldsets = (
        (_("Versuch"), {
            "fields": ("ip", "get_ip_hostname",),
        }),
        (_("Meta"), {
            "fields": ("id", "created",),
        }),
    )

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False


class BurnedOtpAdmin(admin.ModelAdmin):

    @admin.display(description=_("Für Benutzer"), ordering="totp_setup_fk__user_fk__email")
    def get_owner(self, obj: BurnedOtp):
        return obj.totp_setup_fk.user_fk.email

    search_fields = ("totp_setup_fk__user_fk__email",)
    list_filter = ("totp_setup_fk__user_fk__email",)

    ordering = ("-created", "totp_setup_fk__user_fk")

    list_display = (
        "get_owner",
        "otp",
        "created",
    )

    fieldsets = (
        (_("OTP"), {
            "fields": ("otp", "totp_setup_fk",),
        }),
        (_("Besitzer"), {
            "fields": ("get_owner",),
        }),
        (_("Meta"), {
            "fields": ("id", "created",),
        }),
    )

    def has_delete_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False


class UserSessionAdmin(admin.ModelAdmin):

    @admin.display(description=_("Session"), ordering="session_fk__session_key")
    def get_session_key(self, obj: UserSession):
        return obj.session_fk.session_key

    @admin.display(description=_("Für Benutzer"), ordering="user_fk__email")
    def get_owner(self, obj: UserSession):
        return obj.user_fk.email

    list_display = ("get_session_key", "get_owner", "created",)
    search_fields = ("user_fk__email", "session_fk__session_key",)
    list_filter = ("user_fk__email",)

    fieldsets = (
        (_("Session"), {
            "fields": ("session_fk", "user_fk", "get_owner",),
        }),
        (_("Meta"), {
            "fields": ("id", "created", "last_modified",),
        }),
    )

    def has_change_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


class ForgotPasswordAdmin(admin.ModelAdmin):

    @admin.display(description=_("Für Benutzer"), ordering="user_fk__email")
    def get_owner(self, obj: ForgotPassword):
        return obj.user_fk.email

    @admin.display(description=_("Hostname"))
    def get_ip_hostname(self, obj: Union[OtpAttempt, LoginAttempt]):
        return get_hostname(obj.ip)

    list_display = ("get_owner", "created", "ip",)
    search_fields = ("user_fk__email", "ip",)
    list_filter = ("user_fk__email",)

    fieldsets = (
        (_("Benutzer"), {
            "fields": ("user_fk", "get_owner", "ip", "get_ip_hostname",),
        }),
        (_("Meta"), {
            "fields": ("id", "created",),
        }),
    )

    def has_change_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request):
        return False
