import dataclasses
import secrets
from base64 import urlsafe_b64encode
from typing import Optional

from cryptography.hazmat.primitives.hashes import SHA256
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from django.conf import settings


@dataclasses.dataclass
class SiteKey:
    public_salt: str
    key_purpose: Optional[str]
    fernet_key: bytes


def get_site_key(salt: Optional[str] = None, purpose: Optional[str] = None) -> SiteKey:
    """
    We only create a salted key by an already secret key.

    This is why 100000 iterations are enough for this purpose.

    :param salt: An optional salt for the key, one is generated if it is not provided.
    :param purpose: A purpose to make sure the key works only for that.
    :return: A SiteKey object.
    """
    if salt is None:
        salt = secrets.token_hex(16)
    key_salt = salt
    if purpose is not None:
        key_salt = f"{key_salt}#{purpose}"
    kdf = PBKDF2HMAC(
        algorithm=SHA256(),
        length=32,
        salt=key_salt.encode("utf-8"),
        iterations=100_000,
    )
    key = kdf.derive(settings.SECRET_KEY.encode("utf-8"))
    return SiteKey(
        public_salt=salt,
        key_purpose=purpose,
        fernet_key=urlsafe_b64encode(key),
    )
