import binascii
import dataclasses
from base64 import urlsafe_b64encode, urlsafe_b64decode
from datetime import timedelta
from typing import Optional, Union

from cryptography.fernet import Fernet, InvalidToken

from solawi_apps.url.key import get_site_key


@dataclasses.dataclass
class EncryptedParameter:
    public_salt: str
    purpose: Optional[str]
    payload: str


def encrypt_parameter(payload: bytes, purpose: Optional[str] = None) -> EncryptedParameter:
    key = get_site_key(purpose=purpose)
    fernet = Fernet(key=key.fernet_key)
    encrypted_payload = fernet.encrypt(payload)
    return EncryptedParameter(
        public_salt=key.public_salt,
        purpose=purpose,
        payload=urlsafe_b64encode(encrypted_payload).decode('utf-8'),
    )


def decrypt_parameter(payload: EncryptedParameter, ttl: Optional[Union[int, timedelta]] = None) -> Optional[bytes]:
    key = get_site_key(purpose=payload.purpose, salt=payload.public_salt)
    if ttl is not None:
        if isinstance(ttl, timedelta):
            ttl = ttl.total_seconds()
    fernet = Fernet(key=key.fernet_key)
    try:
        encrypted_payload = urlsafe_b64decode(payload.payload)
        decrypted = fernet.decrypt(encrypted_payload, ttl)
    except (InvalidToken, binascii.Error):
        return None
    return decrypted


def encrypt_str_parameter(payload: str, purpose: Optional[str] = None) -> EncryptedParameter:
    payload = payload.encode('utf-8')
    return encrypt_parameter(payload, purpose)


def decrypt_str_parameter(payload: EncryptedParameter, ttl: Optional[Union[int, timedelta]] = None) -> Optional[str]:
    decrypted = decrypt_parameter(payload, ttl)
    return decrypted.decode('utf-8')
