from argparse import ArgumentParser

from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.mail import EmailMessage
from django.core.management import BaseCommand, CommandError
from django.core.validators import validate_email
from django.db import transaction, IntegrityError
from django.template.loader import get_template
from django.urls import reverse
from django.utils.translation import gettext as _

from solawi_apps.app_permission.integration import get_all_installed_app_labels
from solawi_apps.app_permission.models import AccessPermission, DefaultPermission
from solawi_apps.user.registration import register_user


class Command(BaseCommand):

    help = _("Lade einen Benutzer ein")

    def add_arguments(self, parser: ArgumentParser):
        parser.add_argument("email", type=str, nargs="?", help=_("E-Mail-Adresse"))
        parser.add_argument(
            "--module", "-m", nargs="*", help=_("Apps, für die der Benutzer freigeschaltet sein soll")
        )

    def handle(self, *args, **options):
        if (emails := options.get("email")) is None:
            email = input(f"{_('E-Mail-Adresse')}: ").strip().lower()
        else:
            email = emails.strip().lower()
        try:
            validate_email(email)
        except ValidationError as err:
            raise CommandError(_("email ungültig")) from err
        with transaction.atomic():
            tx_id = transaction.savepoint()
            try:
                registration = register_user(email, as_superuser=False)
            except ValueError as err:
                raise CommandError(' / '.join([str(e) for e in err.args])) from err
            installed_modules = [t[0] for t in get_all_installed_app_labels()]
            if (modules := options.get("module")) is not None:
                if len(modules) > 0:
                    permissions = []
                    for m in modules:
                        module_name = f"solawi_apps.{m}"
                        if module_name in installed_modules:
                            permissions.append(AccessPermission(
                                user_fk_id=registration.user_id,
                                app_name=module_name,
                                enabled=True,
                            ))
                        else:
                            raise CommandError(_("Unbekanntes Modul: %s") % module_name)
                    AccessPermission.objects.bulk_create(permissions, ignore_conflicts=True)
            permissions = []
            for app_perm in DefaultPermission.objects.all():
                permissions.append(AccessPermission(
                    user_fk_id=registration.user_id,
                    app_name=app_perm.app_name,
                    enabled=app_perm.enabled,
                    active_from=app_perm.active_from,
                    active_to=app_perm.active_to,
                ))
            AccessPermission.objects.bulk_create(permissions, ignore_conflicts=True)
            reg_url = reverse(
                "user:register.register", kwargs={
                    "internal_user": registration.internal_username,
                    "registration_id": registration.registration_id,
                    "registration_check": registration.registration_check,
                }
            )
            mail_context = {
                "reg_url": f"{settings.URL_BASE}{reg_url}",
                "contact_url": f'{settings.URL_BASE}{reverse("solawi_apps.ui:contact")}',
                "email": registration.email_address,
            }
            mail_template = get_template("solawi_apps/user/register/mail-register.txt.j2")
            mail = EmailMessage(
                to=[registration.email_address],
                subject=_("[solawi-suite] Deine Anmeldung"),
                body=mail_template.render(mail_context),
            )
            try:
                mail.send(fail_silently=False)
                transaction.savepoint_commit(tx_id)
                self.stdout.write(self.style.SUCCESS("Erledigt"))
            except IOError:
                transaction.savepoint_rollback(tx_id)
                self.stdout.write(self.style.ERROR("Die Registierung kann nicht abgeschlossen werden."))
