from django.conf import settings
from django.contrib import messages
from django.urls import reverse_lazy
from django.utils.html import escape
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _, ngettext

from solawi_apps.announce.context import add_announcement_to_context
from solawi_apps.user.models import EmergencyCode


class EmergencyCodeMiddleware:

    codes_url = reverse_lazy("solawi_apps.user:manage.emergency-codes")
    mark_marked = "__emergency_codes__warning_issued__"
    announcement_id = "__emergency_codes__announce_id__"

    def __init__(self, get_response):
        self.get_response = get_response
        self.id_length = 48
        self.warn_threshold = settings.EMERGENCY_CODE_WARNING_THRESHOLD
        self.alert_threshold = settings.EMERGENCY_CODE_ALERT_THRESHOLD

    def create_message(self, request, level: int, message) -> None:
        codes_url = str(self.codes_url)
        code_message = mark_safe(
            '%(message)s<br>'
            '<a href="%(url)s" class="btn btn-sm mt-3 btn-secondary">'
            '<i class="fa fa-fw fa-solid fa-xmarks-lines"></i> %(text)s'
            '</a>' % {
                "message": escape(message),
                "url": escape(f"{str(self.codes_url)}"),
                "text": escape(_("Notfallcodes erneuern")),
            }
        )
        add_announcement_to_context(request, level, code_message, self.announcement_id, [codes_url])
        setattr(request, self.mark_marked, True)

    def __call__(self, request):
        if getattr(request, self.mark_marked, None) is None:
            if request.user.is_authenticated:
                count = EmergencyCode.objects.filter(owner_fk=request.user).count()
                if count == 0:
                    self.create_message(
                        request,
                        messages.ERROR,
                        _("Dir stehen keine keine Notfallcodes mehr zur Verfügung. Bitte erzeuge umgehend neue Codes!")
                    )
                elif count <= self.alert_threshold:
                    self.create_message(
                        request,
                        messages.ERROR,
                        ngettext(
                            "Dir steht nur noch ein einziger Notfallcode zur Verfügung. "
                            "Bitte erzeuge umgehend neue Codes.",
                            "Dir stehen nur noch %(count)d Notfallcodes zur Verfügung. "
                            "Bitte erzeuge umgehend neue Codes.",
                            count,
                        ) % {
                            "count": count,
                        }
                    )
                elif count <= self.warn_threshold:
                    self.create_message(
                        request,
                        messages.WARNING,
                        _(
                            "Dir stehen nur noch %(count)d Notfallcodes zur Verfügung. "
                            "Vielleicht ist jetzt ein guter Zeitpunkt, neue Codes zu erzeugen.",
                        ) % {
                            "count": count,
                        }
                    )
        return self.get_response(request)
