from django.conf import settings
from django.db import models
from django.db.models import OneToOneField
from django.utils.translation import gettext_lazy as _

from solawi_apps.app_permission.integration import get_all_installed_app_labels
from solawi_apps.db_lib.dbid import generate_id, ID_LENGTH
from solawi_apps.db_lib.models import CreatedMixin, LastModifiedMixin, ActiveFromMixin, ActiveToMixin, EnabledMixin, \
    EncryptedTextField, NanoIdPkMixin


class Registration(
    NanoIdPkMixin, EnabledMixin, ActiveFromMixin, ActiveToMixin, CreatedMixin, LastModifiedMixin, models.Model
):

    class Meta:
        verbose_name = _("Registrierung")
        verbose_name_plural = _("Registrierungen")
        default_permissions = ()

    user_fk = OneToOneField(
        settings.AUTH_USER_MODEL,
        verbose_name=_("User"),
        on_delete=models.CASCADE,
        null=False,
        blank=False,
        editable=False,
    )
    as_superuser = models.BooleanField(
        verbose_name=_("Registrierung als Superuser"), default=False, editable=False, null=False, blank=False
    )
    registration_id = models.CharField(
        verbose_name=_("Registrierungs-ID"),
        editable=False,
        default=generate_id,
        unique=True,
        max_length=ID_LENGTH,
        null=False,
        blank=False,
    )
    registration_check = EncryptedTextField(
        verbose_name=_("Registrierungs-Check"),
        null=False,
        blank=False,
        default=generate_id,
        editable=False
    )
    registration_finished = models.BooleanField(
        verbose_name=_("Registrierung abgeschlossen"), null=False, blank=False, default=False
    )

    def __str__(self):
        return _("Registrierung für %s") % self.user_fk.email


class AllowedEmailDomain(
    NanoIdPkMixin, CreatedMixin, LastModifiedMixin, EnabledMixin, ActiveFromMixin, ActiveToMixin, models.Model
):

    class Meta:
        verbose_name = _("Zugelassene Domain")
        verbose_name_plural = _("Zugelassene Domains")

    domain = models.CharField(verbose_name=_("Domain"), max_length=255, null=False, blank=False)

    def __str__(self):
        return self.domain


class DomainBasedAppAccess(
    NanoIdPkMixin, CreatedMixin, LastModifiedMixin, EnabledMixin, ActiveFromMixin, ActiveToMixin, models.Model
):

    class Meta:
        verbose_name = _("Domain-basierter App-Zugang")
        verbose_name_plural = _("Domain-basierte App-Zugänge")
        unique_together = (
            ("allowed_domain_fk", "app_name"),
        )

    allowed_domain_fk = models.ForeignKey(
        AllowedEmailDomain, verbose_name=_("Zugelassene Domain"), null=False, blank=False, on_delete=models.CASCADE
    )
    app_name = models.CharField(
        verbose_name=_("App Name"), choices=get_all_installed_app_labels, null=False, blank=False, max_length=200
    )

    def __str__(self):
        return f"{self.app_name} für {self.allowed_domain_fk}"
