from django.conf import settings
from django.contrib.sessions.models import Session
from django.db import models
from django.db.models import OneToOneField, ForeignKey, CharField, BooleanField, GenericIPAddressField
from django.utils.translation import gettext_lazy as _

from solawi_apps.db_lib.models import NanoIdPkMixin, CreatedMixin, LastModifiedMixin, EncryptedTextField


class TotpSetup(NanoIdPkMixin, CreatedMixin, LastModifiedMixin, models.Model):

    class Meta:
        verbose_name = _('TOTP Setup')
        verbose_name_plural = _('TOTP Setups')
        default_permissions = ()

    user_fk = OneToOneField(
        settings.AUTH_USER_MODEL,
        verbose_name=_("User"),
        on_delete=models.CASCADE,
        null=False,
        blank=False,
        editable=False,
    )
    totp_provisioning = EncryptedTextField(
        verbose_name=_("TOTP Provisionierung"),
        max_length=1024,
        null=False,
        blank=False,
        editable=False,
    )

    def __str__(self):
        return self.user_fk.email


class BurnedOtp(NanoIdPkMixin, CreatedMixin, models.Model):

    class Meta:
        verbose_name = _("Verbranntes Einmalpasswort")
        verbose_name_plural = _("Verbrannte Einmalpasswörter")
        unique_together = (
            ("totp_setup_fk", "otp"),
        )
        default_permissions = ()

    totp_setup_fk = ForeignKey(
        TotpSetup, verbose_name=_("TOTP Setup"), null=False, blank=False, on_delete=models.CASCADE, editable=False
    )
    otp = CharField(verbose_name=_("Einmalpasswort"), max_length=6, null=False, blank=False, editable=False)

    def __str__(self):
        return self.otp


class OtpAttempt(NanoIdPkMixin, CreatedMixin, models.Model):

    class Meta:
        verbose_name = _("OTP-Versuch")
        verbose_name_plural = _("OTP-Versuche")
        default_permissions = ()

    user_fk = ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_("User"),
        on_delete=models.CASCADE,
        null=False,
        blank=False,
        editable=False,
    )
    success = BooleanField(
        verbose_name=_("Erfolgreicher Versuch"), null=False, blank=False, editable=False, default=False
    )
    ip = GenericIPAddressField(verbose_name=_("IP Adresse"), null=True, blank=False, default=None)

    def __str__(self):
        return _("Versuch für %s") % self.user_fk.email


class LoginAttempt(NanoIdPkMixin, CreatedMixin, models.Model):

    class Meta:
        verbose_name = _("Login-Versuch")
        verbose_name_plural = _("Login-Versuche")
        default_permissions = ()

    user_fk = ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_("User"),
        on_delete=models.CASCADE,
        null=False,
        blank=False,
        editable=False,
    )
    success = BooleanField(
        verbose_name=_("Erfolgreicher Versuch"), null=False, blank=False, editable=False, default=False
    )
    ip = GenericIPAddressField(verbose_name=_("IP Adresse"), null=True, blank=False, default=None)

    def __str__(self):
        return _("Versuch für %s") % self.user_fk.email


class AuthenticationAttempt(NanoIdPkMixin, CreatedMixin, models.Model):

    class Meta:
        verbose_name = _("Authentifizierungs-Versuch")
        verbose_name_plural = _("Authentifizierungs-Versuche")
        default_permissions = ()

    ip = GenericIPAddressField(verbose_name=_("IP Adresse"), null=False, blank=False)

    def __str__(self):
        return _("Versuch von %s") % self.ip


class UserSession(NanoIdPkMixin, CreatedMixin, LastModifiedMixin, models.Model):

    class Meta:
        verbose_name = _("Benutzer-Sitzung")
        verbose_name_plural = _("Benutzer-Sitzungen")
        default_permissions = ()

    session_fk = OneToOneField(
        Session, verbose_name=_("Sitzung"), null=False, blank=False, on_delete=models.CASCADE, editable=False
    )
    user_fk = ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_("User"),
        null=False, blank=False,
        editable=False,
        on_delete=models.CASCADE,
    )

    def __str__(self):
        return self.session_fk.session_key


class ForgotPassword(NanoIdPkMixin, CreatedMixin, models.Model):

    class Meta:
        verbose_name = _("Vergessenes Passwort")
        verbose_name_plural = _("Vergessene Passwörter")
        default_permissions = ()

    user_fk = ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_("User"),
        null=False,
        blank=False,
        on_delete=models.CASCADE,
        editable=False,
    )
    ip = GenericIPAddressField(verbose_name=_("IP Adresse"), null=False, blank=False)

    def __str__(self):
        return self.user_fk.email


class EmergencyCode(NanoIdPkMixin, CreatedMixin, models.Model):

    class Meta:
        verbose_name = _("Notfall-Code")
        verbose_name_plural = _("Notfall-Codes")
        default_permissions = ()

    owner_fk = ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_("Besitzer"),
        editable=False,
        null=False,
        blank=False,
        on_delete=models.CASCADE,
    )
    code_hash = models.CharField(verbose_name=_("Code Hash"), editable=False, null=False, blank=False, max_length=255)


class EmergencyCodeAttempt(NanoIdPkMixin, CreatedMixin, models.Model):

    class Meta:
        verbose_name = _("Notfall-Code-Versuch")
        verbose_name_plural = _("Notfall-Code-Versuche")
        default_permissions = ()

    user_fk = ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_("User"),
        on_delete=models.CASCADE,
        null=False,
        blank=False,
        editable=False,
    )
    success = BooleanField(
        verbose_name=_("Erfolgreicher Versuch"), null=False, blank=False, editable=False, default=False
    )
    ip = GenericIPAddressField(verbose_name=_("IP Adresse"), null=True, blank=False, default=None)

    def __str__(self):
        return _("Versuch für %s") % self.user_fk.email
