import dataclasses
from datetime import timedelta
from typing import Optional

from django.contrib.auth import get_user_model
from django.core.validators import RegexValidator
from django.db import DatabaseError
from django.forms import CharField
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from pyotp import random_base32, TOTP, parse_uri
from qrcode import ERROR_CORRECT_L
from qrcode.main import QRCode
from qrcode.image.svg import SvgFillImage

from solawi_apps.user.models import TotpSetup, BurnedOtp, OtpAttempt


@dataclasses.dataclass
class OtpCheckResult:

    success: bool
    rate_limited: bool

    def __bool__(self):
        return self.success


def gen_secret() -> str:
    return random_base32()


def get_secret(provisioning_url: str) -> str:
    otp = parse_uri(provisioning_url)
    return otp.secret


def get_totp_object(secret: Optional[str] = None, username: Optional[str] = None, issuer: Optional[str] = None) -> TOTP:
    if secret is None:
        secret = gen_secret()
    if issuer is None:
        issuer = "[solawi-suite]"
    return TOTP(issuer=issuer, s=secret, digits=6, interval=30, name=username)


def get_svg_qr_code(totp_url: str) -> str:
    qr = QRCode(
        error_correction=ERROR_CORRECT_L,
        box_size=32,
        border=4,
        image_factory=SvgFillImage,
    )
    qr.add_data(totp_url)
    qr.make(fit=True)
    img = qr.make_image()
    return img.to_string(encoding="utf-8")


def make_otp_input_field(
    label: Optional = None, required: bool = True, **options
) -> CharField:
    if label is None:
        label = _("Einmal-Passwort")
    field_options = {
        "label": label,
        "required": required,
        "min_length": 6,
        "max_length": 6,
        "validators": [
            RegexValidator(regex=r"^\d{6}$"),
        ]
    }
    if options is not None:
        field_options.update(options)
    field = CharField(**field_options)
    return field


def check_against_provisioning(otp: Optional[str], provisioning_url: Optional[str]) -> bool:
    if otp is None or provisioning_url is None:
        return False
    return get_totp_object(get_secret(provisioning_url)).verify(otp)


def _check_otp(otp: str, for_user: get_user_model()) -> bool:
    otp = otp.strip()
    try:
        totp_setup = TotpSetup.objects.get(user_fk=for_user)
    except TotpSetup.DoesNotExist:
        return False
    if BurnedOtp.objects.filter(
        totp_setup_fk=totp_setup, created__gt=now() - timedelta(hours=12), otp__iexact=otp
    ).exists():
        return False
    if get_totp_object(get_secret(totp_setup.totp_provisioning)).verify(otp, valid_window=1):
        try:
            BurnedOtp.objects.create(totp_setup_fk=totp_setup, otp=otp)
            return True
        except DatabaseError:
            return False
    return False


def check_otp(otp: str, for_user: get_user_model(), ip: Optional[str]) -> OtpCheckResult:
    success = _check_otp(otp, for_user)
    rate_limit_hit = False
    OtpAttempt.objects.create(user_fk=for_user, success=success, ip=ip)
    if OtpAttempt.objects.filter(user_fk=for_user, created__gte=now() - timedelta(minutes=5)).count() > 10:
        rate_limit_hit = True
        success = False
    return OtpCheckResult(success=success, rate_limited=rate_limit_hit)
