import dataclasses
from uuid import uuid4

from django.contrib.auth import get_user_model
from django.db import transaction
from django.utils.translation import gettext_lazy as _

from solawi_apps.user.models import Registration, TotpSetup
from solawi_apps.user.otp import get_totp_object


@dataclasses.dataclass
class UserRegistration:
    internal_username: str
    email_address: str
    registers_as_superuser: bool
    registration_id: str
    registration_check: str
    user_id: int


def register_user(email: str, as_superuser: bool = False) -> UserRegistration:
    """
    Creates a new Registration attempt.

    :param email: E-Mail address of the the new user.
    :param as_superuser: Whether the user registers as superuser.
    :return: A UserRegistration object.
    :raises ValueError: If the e-mail address is already registered.
    """
    email = email.lower()
    UserModel = get_user_model()
    if UserModel.objects.filter(email__iexact=email).exists():
        raise ValueError(_("Benutzer existiert bereits"))
    with transaction.atomic():
        tx_id = transaction.savepoint()
        user = UserModel.objects.create_user(
            is_active=False,
            is_staff=False,
            is_superuser=False,
            email=email,
            username=str(uuid4()),
            password=None,
        )
        registration = Registration.objects.create(
            user_fk=user,
            as_superuser=as_superuser,
            enabled=True,
        )
        TotpSetup.objects.create(
            user_fk=user,
            totp_provisioning=get_totp_object(username=user.email).provisioning_uri(),
        )
        transaction.savepoint_commit(tx_id)
    return UserRegistration(
        user_id=user.id,
        internal_username=user.username,
        email_address=user.email,
        registers_as_superuser=as_superuser,
        registration_id=registration.registration_id,
        registration_check=registration.registration_check,
    )
