from datetime import timedelta
from urllib.parse import quote

from django.contrib import messages
from django.contrib.auth import logout, get_user_model, login, authenticate
from django.contrib.auth.hashers import verify_password
from django.contrib.auth.mixins import UserPassesTestMixin, LoginRequiredMixin
from django.contrib.messages.views import SuccessMessageMixin
from django.http import HttpRequest
from django.shortcuts import redirect
from django.urls import reverse_lazy, reverse
from django.utils.decorators import method_decorator
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from django.views.decorators.cache import never_cache
from django.views.generic import FormView

from solawi_apps.ui.forms.dummy import DummyForm
from solawi_apps.ui.forms.form import ErrorSensitiveFormViewMixin
from solawi_apps.ui.next_url import get_next_url, get_from_url
from solawi_apps.user.forms.auth import LoginForm, OtpForm, EmergencyCodeForm
from solawi_apps.user.ip import get_ip
from solawi_apps.user.models import LoginAttempt, AuthenticationAttempt, EmergencyCodeAttempt, EmergencyCode
from solawi_apps.user.otp import check_otp
from solawi_apps.user.session import logout_all_sessions
from solawi_platform import settings


INTERNAL_USERNAME_SESSION_KEY = "__internal_username_pre_otp__"
RATE_LIMIT_MSG = _("Hoppla! Das waren zu viele Fehlversuche. Bitte warte einige Zeit, bevor Du es erneut versuchst.")


def rate_limit_hit(request: HttpRequest):
    logout(request)
    messages.error(request, RATE_LIMIT_MSG)
    next_url = get_next_url(request)
    if next_url:
        return redirect(next_url, permanent=False)
    return redirect(settings.LOGIN_URL, permanent=False)


class LoginView(UserPassesTestMixin, SuccessMessageMixin, ErrorSensitiveFormViewMixin, FormView):

    raise_exception = True
    form_class = LoginForm
    template_name = 'solawi_apps/user/auth/login.html'

    user_object_key = "__user_object__"

    def test_func(self):
        return not self.request.user.is_authenticated

    @method_decorator(never_cache)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    def form_valid(self, form):
        logout(self.request)
        ip = get_ip(self.request)
        if ip is not None:
            AuthenticationAttempt.objects.create(ip=ip)
            if AuthenticationAttempt.objects.filter(ip=ip, created__gte=now() - timedelta(minutes=1)).count() > 10:
                form.add_error(None, RATE_LIMIT_MSG)
                return self.form_invalid(form)
        username = form.cleaned_data.get('username')
        password = form.cleaned_data.get('password')
        UserModel = get_user_model()
        try:
            t_user = UserModel.objects.filter(is_active=True).get(email=username)
        except (UserModel.DoesNotExist, UserModel.MultipleObjectsReturned):
            form.add_error(None, _("Benutzername oder Passwort falsch."))
            return self.form_invalid(form)
        if LoginAttempt.objects.filter(user_fk=t_user, created__gte=now()-timedelta(minutes=10)).count() > 5:
            form.add_error(None, RATE_LIMIT_MSG)
            return self.form_invalid(form)
        if (user := authenticate(username=t_user.username, password=password, request=self.request)) is not None:
            LoginAttempt.objects.create(user_fk=user, success=True, ip=ip)
            setattr(self, self.user_object_key, user)
            self.request.session[INTERNAL_USERNAME_SESSION_KEY] = user.username
            return super().form_valid(form)
        else:
            LoginAttempt.objects.create(user_fk=user, success=False, ip=ip)
        form.add_error(None, _("Benutzername oder Passwort falsch."))
        return self.form_invalid(form)

    def get_success_url(self):
        success_url = reverse("solawi_apps.user:auth.otp")
        if (next_url := get_next_url(self.request)) is not None:
            success_url = f"{success_url}?next={quote(next_url)}"
        return success_url

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["next_url"] = get_next_url(self.request)
        return context

    def get_success_message(self, cleaned_data = None):
        return _("Hallo %s! Nur noch ein weiterer Schritt!") % getattr(self, self.user_object_key).first_name


class OtpView(UserPassesTestMixin, SuccessMessageMixin, ErrorSensitiveFormViewMixin, FormView):

    raise_exception = True
    form_class = OtpForm
    template_name = "solawi_apps/user/auth/otp.html"

    def test_func(self):
        if self.request.user.is_authenticated:
            return False
        return self.request.session.get(INTERNAL_USERNAME_SESSION_KEY, None) is not None

    @method_decorator(never_cache)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["next_url"] = get_next_url(self.request)
        return context

    def form_valid(self, form):
        UserModel = get_user_model()
        try:
            user = UserModel.objects.filter(
                is_active=True,
            ).get(
                username=self.request.session.get(INTERNAL_USERNAME_SESSION_KEY),
            )
        except (UserModel.DoesNotExist, UserModel.MultipleObjectsReturned):
            form.add_error(None, _("Leider kannst du zur Zeit nicht angemeldet werden."))
            return self.form_invalid(form)
        otp_result = check_otp(form.cleaned_data["otp"], user, get_ip(self.request))
        if not otp_result.success:
            if otp_result.rate_limited:
                return rate_limit_hit(self.request)
            form.add_error("otp", _("Das Einmalpasswort ist ungültig."))
            return self.form_invalid(form)
        logout(self.request)
        login(self.request, user)
        return super().form_valid(form)

    def get_success_url(self):
        if (next_url := get_next_url(self.request)) is not None:
            return next_url
        return settings.LOGIN_REDIRECT_URL

    def get_success_message(self, cleaned_data):
        return _("Herzlich willkommen, %(name)s!") % {
            "name": self.request.user.first_name,
        }


class EmergencyCodeView(UserPassesTestMixin, SuccessMessageMixin, ErrorSensitiveFormViewMixin, FormView):

    raise_exception = True
    form_class = EmergencyCodeForm
    template_name = "solawi_apps/user/auth/emergency-code.html"

    def test_func(self):
        if self.request.user.is_authenticated:
            return False
        return self.request.session.get(INTERNAL_USERNAME_SESSION_KEY, None) is not None

    @method_decorator(never_cache)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    def get_success_url(self):
        if (next_url := get_next_url(self.request)) is not None:
            return next_url
        return settings.LOGIN_REDIRECT_URL

    def get_success_message(self, cleaned_data):
        return _("Herzlich willkommen, %(name)s! Bitte streiche den Notfallcode '%(code)s' von deiner Liste.") % {
            "name": self.request.user.first_name,
            "code": cleaned_data["emergency_code"]
        }

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["next_url"] = get_next_url(self.request)
        return context

    def form_valid(self, form):
        UserModel = get_user_model()
        try:
            user = UserModel.objects.filter(
                is_active=True,
            ).get(
                username=self.request.session.get(INTERNAL_USERNAME_SESSION_KEY),
            )
        except (UserModel.DoesNotExist, UserModel.MultipleObjectsReturned):
            form.add_error(None, _("Leider kannst du zur Zeit nicht angemeldet werden."))
            return self.form_invalid(form)
        hit_count = EmergencyCodeAttempt.objects.filter(user_fk=user, created__gte=now()-timedelta(minutes=10)).count()
        if hit_count >= 5:
            return rate_limit_hit(self.request)
        identified_code = None
        for k_obj in EmergencyCode.objects.filter(owner_fk=user).all():
            correct, needs_update = verify_password(form.cleaned_data["emergency_code"], k_obj.code_hash)
            if correct:
                identified_code = k_obj
                break
        EmergencyCodeAttempt.objects.create(
            user_fk=user,
            success=identified_code is not None,
            ip=get_ip(self.request),
        )
        if identified_code is not None:
            identified_code.delete()
            logout(self.request)
            login(self.request, user)
            return super().form_valid(form)
        form.add_error("emergency_code", _("Der Notfallcode ist ungültig."))
        return self.form_invalid(form)


class LogoutView(LoginRequiredMixin, SuccessMessageMixin, FormView):

    raise_exception = True
    form_class = DummyForm
    template_name = "solawi_apps/user/auth/logout.html"
    success_message = _("Logout erfolgreich!")

    def form_valid(self, form):
        logout(self.request)
        self.request.session.clear()
        return super().form_valid(form)

    def get_success_url(self):
        if (next_url := get_next_url(self.request)) is not None:
            return next_url
        if (logout_url := settings.LOGOUT_REDIRECT_URL) is not None:
            return logout_url
        return reverse_lazy("solawi_apps.ui:welcome")

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["back_link"] = get_from_url(self.request)
        return context

    @method_decorator(never_cache)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)


class LogoutAllSessionsView(LoginRequiredMixin, SuccessMessageMixin, FormView):

    raise_exception = True
    form_class = DummyForm
    template_name = "solawi_apps/user/auth/logout-all.html"
    success_message = _("Logout aller Sitzungen erfolgreich!")

    def form_valid(self, form):
        logout_all_sessions(self.request.user)
        logout(self.request)
        return super().form_valid(form)

    def get_success_url(self):
        if (next_url := get_next_url(self.request)) is not None:
            return next_url
        if (logout_url := settings.LOGOUT_REDIRECT_URL) is not None:
            return logout_url
        return reverse_lazy("solawi_apps.ui:welcome")

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["back_link"] = get_from_url(self.request)
        return context

    @method_decorator(never_cache)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)
