import string
from datetime import timedelta

from django.contrib import messages
from django.contrib.auth import logout, get_user_model
from django.contrib.auth.hashers import make_password
from django.contrib.auth.mixins import LoginRequiredMixin, UserPassesTestMixin
from django.contrib.auth.password_validation import validate_password
from django.contrib.messages.views import SuccessMessageMixin
from django.core.exceptions import ValidationError
from django.core.mail import EmailMessage
from django.db import transaction
from django.http import Http404, HttpResponse
from django.shortcuts import render
from django.template.loader import get_template
from django.urls import reverse_lazy, reverse
from django.utils.decorators import method_decorator
from django.utils.timezone import now
from django.views import View
from django.views.decorators.cache import never_cache
from django.views.generic import TemplateView, FormView
from django.utils.translation import gettext_lazy as _
from nanoid import generate

from solawi_apps.ui.forms.form import ErrorSensitiveFormViewMixin
from solawi_apps.ui.next_url import get_from_url
from solawi_apps.url.params import encrypt_str_parameter, decrypt_str_parameter, EncryptedParameter
from solawi_apps.user.forms.manage import ChangePasswordForm, OtpChangeForm, ForgotPasswordForm, NewPasswordForm
from solawi_apps.user.ip import get_ip
from solawi_apps.user.models import TotpSetup, ForgotPassword, EmergencyCode
from solawi_apps.user.otp import get_totp_object, get_secret, check_against_provisioning, get_svg_qr_code
from solawi_apps.user.session import logout_all_sessions


OTP_PROVISIONING_SESSION_KEY = "__change_otp_provisioning_session_key__"


class AccountManageView(LoginRequiredMixin, TemplateView):

    template_name = "solawi_apps/user/manage/index.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["self_service_apps"] = (
            (
                _("Passkeys verwalten"),
                "fa-solid fa-unlock-keyhole",
                "light",
                reverse_lazy("solawi_apps.passkeys:passkeys"),
            ),
            (
                _("Passwort ändern"),
                "fa-solid fa-key",
                "light",
                reverse_lazy("solawi_apps.user:manage.password"),
            ),
            (
                _("Einmal-Passwort-Geheimnis ändern"),
                "fa-solid fa-qrcode",
                "light",
                reverse_lazy("solawi_apps.user:manage.otp-secret"),
            ),
            (
                _("Notfallcodes erneuern"),
                "fa-solid fa-xmarks-lines",
                "light",
                reverse_lazy("solawi_apps.user:manage.emergency-codes"),
            ),
            (
                _("Logout"),
                "fa-solid fa-xmark",
                "warning",
                reverse_lazy("solawi_apps.user:auth.logout"),
            ),
            (
                _("Von allen Sitzungen abmelden"),
                "fa-solid fa-right-from-bracket",
                "danger",
                reverse_lazy("solawi_apps.user:auth.logout-all"),
            ),
        )
        return context


class ChangePasswordView(LoginRequiredMixin, SuccessMessageMixin, ErrorSensitiveFormViewMixin, FormView):

    success_message = _("Dein Passwort wurde erfolgreich geändert.")
    success_url = reverse_lazy("solawi_apps.user:manage")

    template_name = "solawi_apps/user/manage/change_password.html"

    form_class = ChangePasswordForm

    def form_valid(self, form):
        has_error = False
        user = self.request.user
        old_password = form.cleaned_data.get("old_password")
        if not user.check_password(old_password):
            has_error = True
            form.add_error("old_password", _("Das Passwort ist ungültig."))
        new_password1 = form.cleaned_data.get("new_password1")
        try:
            validate_password(new_password1, user=user)
        except ValidationError as err:
            has_error = True
            for error in err.messages:
                form.add_error("new_password1", error)
        new_password2 = form.cleaned_data.get("new_password2")
        if new_password1 != new_password2:
            has_error = True
            form.add_error("new_password2", _("Die neuen Passwörter stimmen nicht überein."))
        if has_error:
            return self.form_invalid(form)
        user.set_password(new_password1)
        user.save()
        logout(self.request)
        logout_all_sessions(user)
        return super().form_valid(form)

    @method_decorator(never_cache)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)


class ChangeOtpView(LoginRequiredMixin, SuccessMessageMixin, ErrorSensitiveFormViewMixin, FormView):

    success_message = _("Dein Einmal-Passwort-Geheimnis wurde erfolgreich geändert.")
    success_url = reverse_lazy("solawi_apps.user:manage")

    template_name = "solawi_apps/user/manage/change_otp.html"

    form_class = OtpChangeForm

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        if (totp := self.request.session.get(OTP_PROVISIONING_SESSION_KEY)) is None:
            totp = get_totp_object(username=self.request.user.email).provisioning_uri()
            self.request.session[OTP_PROVISIONING_SESSION_KEY] = totp
        context["otp_secret"] = get_secret(totp)
        return context

    def form_valid(self, form):
        has_error = False
        user = self.request.user
        old_password = form.cleaned_data.get("old_password")
        if not user.check_password(old_password):
            has_error = True
            form.add_error("old_password", _("Das Passwort ist ungültig."))
        provisioning = self.request.session.get(OTP_PROVISIONING_SESSION_KEY, None)
        if provisioning is None:
            form.add_error(None, _("Diese Aktion kann gerade nicht durchgeführt werden."))
            return self.form_invalid(form)
        if not check_against_provisioning(
            form.cleaned_data.get("otp"),
            provisioning,
        ):
            has_error = True
            form.add_error("otp", _("Das Einmal-Passwort ist ungültig."))
        if has_error:
            return self.form_invalid(form)
        try:
            totp = TotpSetup.objects.get(user_fk=self.request.user)
        except (TotpSetup.DoesNotExist, TotpSetup.MultipleObjectsReturned):
            form.add_error(None, _("Diese Aktion kann gerade nicht durchgeführt werden."))
            return self.form_invalid(form)
        totp.totp_provisioning = provisioning
        totp.save()
        logout(self.request)
        logout_all_sessions(user)
        return super().form_valid(form)

    @method_decorator(never_cache)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)


class ChangeOtpQrCodeView(LoginRequiredMixin, View):

    http_method_names = ["get"]

    content_type = "image/svg+xml"

    @method_decorator(never_cache)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        if (provisioning := request.session.get(OTP_PROVISIONING_SESSION_KEY, None)) is None:
            raise Http404()
        return HttpResponse(
            status=200,
            content_type=self.content_type,
            content=get_svg_qr_code(provisioning),
        )


class ForgotPasswordView(UserPassesTestMixin, SuccessMessageMixin, ErrorSensitiveFormViewMixin, FormView):

    raise_exception = True
    template_name = "solawi_apps/user/manage/forgot-password.html"
    form_class = ForgotPasswordForm

    success_url = reverse_lazy("solawi_apps.ui:welcome")
    success_message = _("Wir haben dir Informationen zur Wiederherstellung deines Zugangs geschickt.")

    def test_func(self):
        return not self.request.user.is_authenticated

    def _form_valid(self, form) -> bool:
        UserModel = get_user_model()
        try:
            user = UserModel.objects.filter(is_active=True).get(email__iexact=form.cleaned_data["email"])
        except (UserModel.DoesNotExist, UserModel.MultipleObjectsReturned):
            return True
        pending_requests = ForgotPassword.objects.filter(user_fk=user, created__gt=now() - timedelta(days=2)).count()
        if pending_requests > 3:
            return False
        with transaction.atomic():
            tx_id = transaction.savepoint()
            fp_obj = ForgotPassword.objects.create(user_fk=user, ip=get_ip(self.request))
            data = encrypt_str_parameter(
                payload=fp_obj.id,
                purpose="forgot-password",
            )
            url = self.request.build_absolute_uri(
                reverse("user:manage.forgot-password.reset", kwargs={
                    "salt": data.public_salt,
                    "data": data.payload,
                })
            )
            mail_template = get_template("solawi_apps/user/manage/mail-forgot-password.txt.j2")
            mail_body = mail_template.render({
                "reset_url": url,
                "contact_url": self.request.build_absolute_uri(reverse("solawi_apps.ui:contact")),
                "email": user.email,
                "first_name": user.first_name,
            })
            mail = EmailMessage(
                to=[user.email],
                subject=_("[solawi-suite] Passwort-Wiederherstellung"),
                body=mail_body,
            )
            try:
                mail.send(fail_silently=False)
                transaction.savepoint_commit(tx_id)
            except IOError:
                transaction.savepoint_rollback(tx_id)
                return False
        return True

    def form_valid(self, form):
        if not self._form_valid(form):
            form.add_error(
                None, _("Aktuell kann deine Anfrage nicht bearbeitet werden. Bitte versuche es später erneut."),
            )
            return self.form_invalid(form)
        return super().form_valid(form)

    @method_decorator(never_cache)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)


class ResetPasswordView(UserPassesTestMixin, SuccessMessageMixin, ErrorSensitiveFormViewMixin, FormView):

    form_class = NewPasswordForm
    template_name = "solawi_apps/user/manage/reset-password.html"
    raise_exception = True
    p_object_key = "__forgot_password_object__"

    success_message = _("Dein neues Passwort wurde festgelegt.")
    success_url = reverse_lazy("solawi_apps.user:manage")

    def test_func(self):
        if self.request.user.is_authenticated:
            return False
        public_salt = self.kwargs.get("salt")
        enc_data = self.kwargs.get("data")
        p_id = decrypt_str_parameter(EncryptedParameter(
            purpose="forgot-password",
            public_salt=public_salt,
            payload=enc_data,
        ), ttl=timedelta(hours=1))
        if p_id is None:
            return False
        try:
            recovery = ForgotPassword.objects.filter(created__gte=now() - timedelta(hours=1)).get(id=p_id)
        except (ForgotPassword.DoesNotExist, ForgotPassword.MultipleObjectsReturned):
            return False
        setattr(self, self.p_object_key, recovery)
        return True

    @method_decorator(never_cache)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    def form_valid(self, form):
        if (forgot_password := getattr(self, self.p_object_key, None)) is None:
            form.add_error(None, _("Diese Aktion kann zur Zeit nicht durchgeführt werden."))
            return self.form_invalid(form)
        if (user := forgot_password.user_fk) is None:
            form.add_error(None, _("Diese Aktion kann zur Zeit nicht durchgeführt werden."))
            return self.form_invalid(form)
        has_error = False
        new_password1 = form.cleaned_data["new_password1"]
        try:
            validate_password(new_password1, user=user)
        except ValidationError as err:
            has_error = True
            for error in err.messages:
                form.add_error("new_password1", error)
        new_password2 = form.cleaned_data["new_password2"]
        if new_password1 != new_password2:
            has_error = True
            form.add_error("new_password2", _("Die neuen Passwörter stimmen nicht überein."))
        if has_error:
            return self.form_invalid(form)
        with transaction.atomic():
            forgot_password.delete()
            user.set_password(new_password1)
            user.save()
        logout(self.request)
        logout_all_sessions(user)
        return super().form_valid(form)


class EmergencyCodeRegenView(LoginRequiredMixin, View):

    http_method_names = ["get", "post"]
    template_name = "solawi_apps/user/manage/renew-emergency-codes.html"

    code_size = 16
    code_alphabet = string.ascii_uppercase + string.digits

    @method_decorator(never_cache)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    def get(self, request):
        if (return_url := get_from_url(request)) is None:
            return_url = reverse("solawi_apps.user:manage")
        return render(request, self.template_name, {
            "return_url": return_url,
        })

    def post(self, request):
        if (return_url := get_from_url(request)) is None:
            return_url = reverse("solawi_apps.user:manage")
        codes = []
        while len(codes) < 10:
            codes.append(generate(self.code_alphabet, self.code_size))
        with transaction.atomic():
            tx_id = transaction.savepoint()
            EmergencyCode.objects.filter(owner_fk=request.user).delete()
            for code in codes:
                EmergencyCode.objects.create(
                    owner_fk=request.user,
                    code_hash=make_password(code),
                )
            transaction.savepoint_commit(tx_id)
        messages.success(request, _("Neue Notfallcodes erzeugt."))
        return render(request, self.template_name, {
            "return_url": return_url,
            "codes": codes,
        })
