from datetime import timedelta
from typing import Optional
from uuid import UUID

from django.conf import settings
from django.contrib.auth.mixins import UserPassesTestMixin
from django.contrib.auth.password_validation import validate_password
from django.contrib.messages.views import SuccessMessageMixin
from django.core.exceptions import ValidationError
from django.core.mail import EmailMessage
from django.db import transaction, DatabaseError, IntegrityError
from django.http import HttpResponse, HttpRequest, Http404
from django.template.loader import get_template
from django.urls import reverse_lazy, reverse
from django.utils.decorators import method_decorator
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from django.views import View
from django.views.decorators.cache import never_cache
from django.views.generic import TemplateView, FormView

from solawi_apps.app_permission.models import AccessPermission, DefaultPermission
from solawi_apps.ui.forms.form import ErrorSensitiveFormViewMixin
from solawi_apps.user.forms.register import Step1Form, Step2Form
from solawi_apps.user.ip import get_ip
from solawi_apps.user.models import AllowedEmailDomain, TotpSetup, Registration, DomainBasedAppAccess
from solawi_apps.user.otp import get_svg_qr_code, get_secret, check_otp
from solawi_apps.user.registration import register_user


class StartRegisterMeView(UserPassesTestMixin, ErrorSensitiveFormViewMixin, FormView):

    template_name = "solawi_apps/user/register/step1.html"
    form_class = Step1Form
    success_url = reverse_lazy("solawi_apps.user:register.start.success")
    raise_exception = True

    @method_decorator(never_cache)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    def test_func(self):
        return not self.request.user.is_authenticated

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["allowed_domains_only"] = settings.ONLY_ALLOWED_DOMAINS_FOR_REGISTRATION
        return context

    def form_valid(self, form):
        if settings.ONLY_ALLOWED_DOMAINS_FOR_REGISTRATION:
            __, domain = form.cleaned_data.get("email").rsplit("@", 1)
            t_now = now()
            if not AllowedEmailDomain.objects.filter(
                domain__iexact=domain,
                enabled=True,
                active_from__lte=t_now,
                active_to__gte=t_now,
            ).exists():
                form.add_error(
                    "email",
                    _("Die Domain '%s' ist derzeit nicht für Registrierungen zugelassen.") % domain,
                )
                return self.form_invalid(form)
        with transaction.atomic():
            tx_id = transaction.savepoint()
            try:
                registration_info = register_user(form.cleaned_data.get("email"), as_superuser=False)
            except ValueError:
                transaction.savepoint_rollback(tx_id)
                form.add_error(
                    "email", _("Mit dieser E-Mail-Adresse kann aktuell keine Registrierung durchgeführt werden.")
                )
                return self.form_invalid(form)
            reg_url = reverse(
                "user:register.register", kwargs={
                    "internal_user": registration_info.internal_username,
                    "registration_id": registration_info.registration_id,
                    "registration_check": registration_info.registration_check,
                }
            )
            mail_context = {
                "reg_url": self.request.build_absolute_uri(reg_url),
                "contact_url": self.request.build_absolute_uri(reverse("solawi_apps.ui:contact")),
                "email": registration_info.email_address,
            }
            mail_template = get_template("solawi_apps/user/register/mail-register.txt.j2")
            mail = EmailMessage(
                to=[registration_info.email_address],
                subject=_("[solawi-suite] Deine Anmeldung"),
                body=mail_template.render(mail_context),
            )
            try:
                mail.send(fail_silently=False)
                transaction.savepoint_commit(tx_id)
            except IOError:
                transaction.savepoint_rollback(tx_id)
                form.add_error(
                    None, _("Aktuell kann deine Anfrage nicht bearbeitet werden. Bitte versuche es später erneut.")
                )
                return self.form_invalid(form)
        return super().form_valid(form)


class StartRegisterMeSuccessView(TemplateView):

    template_name = "solawi_apps/user/register/step1.success.html"


def get_registration(
        internal_user: str, registration_id: str, registration_check: str
) -> Optional[Registration]:
    t_now = now()
    try:
        registration = Registration.objects.filter(
            enabled=True,
            active_from__lte=t_now,
            active_to__gte=t_now,
            created__gte=t_now - timedelta(days=2),
            registration_finished=False,
            registration_id=registration_id,
            user_fk__username=internal_user,
        ).get()
    except Registration.DoesNotExist:
        return None
    return registration if registration.registration_check == registration_check else None


class RegisterTotpQrCodeView(View):

    http_method_names = ["get"]
    content_type = "image/svg+xml"

    @method_decorator(never_cache)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    def get(
        self, request: HttpRequest, internal_user: UUID, registration_id: str, registration_check: str, otp: str
    ) -> HttpResponse:
        if (reg := get_registration(str(internal_user), registration_id, registration_check)) is None:
            raise Http404()
        try:
            totp = TotpSetup.objects.filter(
                user_fk=reg.user_fk,
            ).get(id=otp)
        except TotpSetup.DoesNotExist:
            raise Http404()
        return HttpResponse(
            status=200,
            content_type=self.content_type,
            content=get_svg_qr_code(totp.totp_provisioning),
        )


class RegisterMeView(SuccessMessageMixin, ErrorSensitiveFormViewMixin, FormView):

    form_class = Step2Form
    template_name = "solawi_apps/user/register/step2.html"

    current_user_key = "__current_user__"
    current_registration_key = "__current_registration__"

    success_url = settings.LOGIN_REDIRECT_URL
    success_message = _("Registrierung erfolgreich! Du kannst dich jetzt anmelden.")

    @method_decorator(never_cache)
    def dispatch(self, request, *args, **kwargs):
        if (reg := get_registration(
            str(self.kwargs.get("internal_user")),
            self.kwargs.get("registration_id"),
            self.kwargs.get("registration_check"),
        )) is None:
            raise Http404()
        setattr(self, self.current_registration_key, reg)
        setattr(self, self.current_user_key, reg.user_fk)
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        if (reg := getattr(self, self.current_registration_key, None)) is None:
            raise Http404()
        try:
            totp = TotpSetup.objects.get(user_fk=reg.user_fk)
        except TotpSetup.DoesNotExist:
            raise Http404()
        context_update = {
            "qr_url": reverse("solawi_apps.user:register.register.qr", kwargs={
                "internal_user": str(self.kwargs.get("internal_user")),
                "registration_id": self.kwargs.get("registration_id"),
                "registration_check": self.kwargs.get("registration_check"),
                "otp": totp.id,
            }),
            "otp_secret": get_secret(totp.totp_provisioning),
        }
        context = super().get_context_data(**kwargs)
        context.update(context_update)
        return context

    def form_valid(self, form):
        has_still_errors = False
        password_1 = form.cleaned_data.get("password1")
        password_2 = form.cleaned_data.get("password2")
        if password_1 != password_2:
            has_still_errors = True
            form.add_error("password2", _("Die Passwörter stimmen nicht überein."))
        user = getattr(self, self.current_user_key, None)
        try:
            validate_password(password_1, user=user)
        except ValidationError as e:
            has_still_errors = True
            for error in e.messages:
                form.add_error("password1", error)
        registration = getattr(self, self.current_registration_key, None)
        if user is None or registration is None:
            raise Http404()
        otp_check = check_otp(form.cleaned_data["otp"], for_user=user, ip=get_ip(self.request))
        if not otp_check.success:
            has_still_errors = True
            if otp_check.rate_limited:
                form.add_error(
                    None,
                    _("Das waren zu viele Fehlversuche. Bitte warte ein paar Minuten, bevor du es erneut versuchst."),
                )
            else:
                form.add_error("otp", _("Das Einmal-Passwort ist ungültig."))
        if has_still_errors:
            return self.form_invalid(form)

        __, domain = user.email.rsplit("@", 1)
        domain_permissions = [{
            "app_name": a.app_name,
            "enabled": a.enabled,
            "active_from": a.active_from,
            "active_to": a.active_to,
        } for a in DomainBasedAppAccess.objects.filter(allowed_domain_fk__domain__iexact=domain).all()]
        default_permissions = [{
            "app_name": a.app_name,
            "enabled": a.enabled,
            "active_from": a.active_from,
            "active_to": a.active_to,
        } for a in DefaultPermission.objects.all()]

        with transaction.atomic():
            tx_id = transaction.savepoint()
            try:
                user.set_password(password_1)
                user.is_active = True
                user.is_superuser = registration.as_superuser
                user.is_staff = registration.as_superuser
                user.first_name = form.cleaned_data["first_name"]
                user.last_name = form.cleaned_data["last_name"]
                user.save()
                registration.delete()
                AccessPermission.objects.bulk_create(
                    [AccessPermission(
                        app_name=app_perm["app_name"],
                        user_fk=user,
                        enabled=app_perm["enabled"],
                        active_from=app_perm["active_from"],
                        active_to=app_perm["active_to"],
                    ) for app_perm in domain_permissions],
                    ignore_conflicts=True,
                )
                AccessPermission.objects.bulk_create(
                    [AccessPermission(
                        app_name=app_perm["app_name"],
                        user_fk=user,
                        enabled=app_perm["enabled"],
                        active_from=app_perm["active_from"],
                        active_to=app_perm["active_to"],
                    ) for app_perm in default_permissions],
                    ignore_conflicts=True,
                )
            except DatabaseError:
                transaction.savepoint_rollback(tx_id)
                form.add_error(
                    None,
                    _("Die Registrierung kann zur Zeit nicht abgeschlossen werden. Bitte versuche es später erneut."),
                )
                return self.form_invalid(form)
            transaction.savepoint_commit(tx_id)
        return super().form_valid(form)
