import os
import re
import secrets
import subprocess
from datetime import datetime, timezone
from pathlib import Path
from shutil import rmtree
from tempfile import mkdtemp
from time import sleep

import requests
from robot.api.deco import keyword
from robot.errors import KeywordError, FrameworkError
from robot.libraries.BuiltIn import BuiltIn

from solawi_uat.environment.port import find_free_port


class SolawiSuiteInstanceLibrary:

    ROBOT_LIBRARY_SCOPE = 'GLOBAL'
    ROBOT_LIBRARY_VERSION = '0.1.0'

    TOTP_EXTRACTION_REGEX = re.compile(r'^.*(?P<otp>\d{6}).*$', re.DOTALL)

    def __init__(self):
        self.host = None
        self.seed = secrets.token_hex(8)
        self.temp_folder = mkdtemp(prefix="solawi-suite-uat-temp", suffix=self.seed)
        self.temp_path = Path(self.temp_folder)
        self.database_file = self.temp_path / "db.sqlite3"
        os.environ["DJANGO_DATABASE_FILE"] = str(self.database_file.absolute())
        os.environ["SOLAWI_SUITE_DEBUG"] = "0"
        os.environ["SOLAWI_SUITE_SECRET_KEY"] = "0x3739d20117be06f317d31bf30296b563e37a1a5ef6299a0e920675288f920627"
        self.mail_dir = self.temp_path / "_email"
        self.mail_dir.mkdir(exist_ok=True)
        os.environ["DJANGO_EMAIL_DIR"] = str(self.mail_dir.absolute())
        base_dir = Path(__file__)
        while not (base_dir / "manage.py").exists():
            base_dir = base_dir.parent
            if base_dir == base_dir.parent:
                raise FrameworkError("Cannot find manage.py")
        self.django_root = base_dir
        self.django_process = None

    @keyword(name="Get Test Host")
    def get_test_host(self):
        if self.host is None:
            raise KeywordError("Host not defined; try to start the instance first!")
        return self.host

    @keyword(name="Clean Temporary Folder")
    def clean_temporary_folder(self):
        rmtree(self.temp_folder, ignore_errors=True)

    @keyword(name="Run Manage.Py Command")
    def run_manage_py_command(self, *args):
        completed = subprocess.run(
            ["poetry", "run", "manage.py", *args],
            capture_output=True,
        )
        if completed.returncode != 0:
            BuiltIn().log(completed.stdout)
            BuiltIn().log(completed.stderr)
            raise KeywordError("Failed to execute command")
        return completed

    @keyword(name="Create Test Database")
    def create_test_database(self):
        self.run_manage_py_command("migrate")

    @keyword(name="Start Daphne Server")
    def start_daphne_server(self):
        port = find_free_port()
        self.host = f"http://localhost:{port:d}"
        os.environ["SOLAWI_SUITE_TEST_ORIGIN"] = self.host
        self.django_process = subprocess.Popen(
            ["poetry", "run", "python", "manage.py", "runserver", f"127.0.0.1:{port:d}"],
            cwd=str(self.django_root.absolute()),
        )

    @keyword(name="Shutdown Daphne Server")
    def shutdown_daphne_server(self):
        if self.django_process is None:
            return
        self.django_process.terminate()

    @keyword(name="Wait For Application Startup")
    def wait_for_application_startup(
        self, retry_interval_seconds: int = 1, max_retries: int = 15, request_timeout: float = 2.5
    ):
        if self.host is None:
            raise KeywordError("Host not defined; try to start the instance first!")
        try_no = 0
        url = f"{self.host}/"
        while try_no < max_retries:
            sleep(retry_interval_seconds)
            try_no += 1
            try:
                requests.get(url, timeout=request_timeout)
            except (requests.exceptions.Timeout, requests.exceptions.ConnectionError):
                continue
            except Exception as r:
                raise KeywordError(r)
            return
        raise KeywordError("Application failed to start")

    @keyword(name="Get Latest Mail")
    def get_latest_mail(self, newer_than: datetime = None) -> str:
        if not self.mail_dir.exists():
            raise KeywordError("No email folder found")
        mails = [*self.mail_dir.glob("*.log")]
        if len(mails) == 0:
            raise KeywordError("No mail found")
        sorted_mails = sorted(mails, key=lambda m: m, reverse=True)
        mail = sorted_mails[0]
        mail_obj = self.mail_dir / mail
        if newer_than is not None:
            mail_dt = datetime.fromtimestamp(mail_obj.stat().st_mtime, tz=timezone.utc)
            if mail_dt <= newer_than:
                raise KeywordError("No mail found that is new enough")
        with open(mail_obj, "rt", encoding="utf-8") as mail_data:
            return mail_data.read()

    @keyword(name="Get TOTP For User")
    def get_totp_for_user(self, user: str, differing_from: str = "") -> str:
        while True:
            result = self.run_manage_py_command("gen_otp", user)
            matcher = self.TOTP_EXTRACTION_REGEX.match(result.stdout.decode("utf-8"))
            if matcher is None:
                raise KeywordError("No TOTP found")
            new_otp = matcher.group("otp")
            if new_otp != differing_from:
                break
            sleep(1.0)
        return new_otp
